#ifndef CGUL_BDF_GLYPH_CXX_H
#define CGUL_BDF_GLYPH_CXX_H

/**
 * C++ wrapper for <code>cgul_bdf_glyph</code>.
 *
 * @file
 * @author Paul Serice
 * @brief C++ wrapper for <code>cgul_bdf_glyph</code>
 */

#include "cgul_bdf_glyph.h"
#include "cgul_exception_cxx.h"
#include "cgul_string_cxx.h"

/**
 * This class provides the C++ bindings for C
 * <code>cgul_bdf_glyph</code> objects.  The main purpose of this
 * class is to convert the C-style function calls and exception
 * handling in <code>cgul_bdf_glyph</code> into C++-style function
 * calls and exception handling.  This class must not be extended.
 *
 * @brief C++ bindings for <code>cgul_bdf_glyph</code>
 */
class cgul_bdf_glyph_cxx {

public:

    /*
     * IMPLEMENTATION NOTE:
     * ===================
     *
     * There are no real cgul_bdf_glyph_cxx objects.  Instead, this
     * class is carefully implemented so that "cgul_bdf_glyph"
     * objects (not "cgul_bdf_glyph_cxx" objects) can be directly
     * used to invoke the methods of this class.
     *
     * It is done this way to avoid having to keep track of some sort
     * of C++ wrapper for cgul_bdf_glyph objects returned by
     * cgul_bdf_cxx::get_glyph().
     *
     * This works because this class has no associated data _and_ no
     * virtual methods which would cause a vtable lookup, and vtables
     * are typically embedded in the object itself.
     *
     * In summary, do _not_ add data or virtual functions to this
     * class.
     */

    /**
     * This method does a deep copy of <code>src</code> and stores it
     * in this object.  If an error occurs, an exception is thrown,
     * and this object will be in an undefined state.
     *
     * @param[in] src source glyph
     */
    void
    copy(cgul_bdf_glyph_cxx* src)
    {
        cgul_bdf_glyph_t glyph = (cgul_bdf_glyph_t)this;
        cgul_bdf_glyph_t rhs = (cgul_bdf_glyph_t)src;
        cgul_exception_t local = NULL;
        cgul_bdf_glyph__copy(&local, glyph, rhs);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return;
    }

    /**
     * This method returns the descriptive name for the character.  This
     * is the name that belongs on the STARTCHAR line.  If no descriptive
     * name is available, <code>NULL</code> is returned.  Thus, it is
     * important that you always check the return value.  The pointer
     * returned belongs to this class and should not be altered or freed.
     *
     * @return descriptive name or <code>NULL</code>
     */
    const char*
    get_descriptive_name() const
    {
        const char* rv = NULL;
        cgul_bdf_glyph_t glyph = (cgul_bdf_glyph_t)this;
        cgul_exception_t local = NULL;
        rv = cgul_bdf_glyph__get_descriptive_name(&local, glyph);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Get the encoding.
     *
     * @return encoding
     */
    cgul_wchar_t
    get_encoding() const
    {
        cgul_wchar_t rv = (cgul_wchar_t)0;
        cgul_bdf_glyph_t glyph = (cgul_bdf_glyph_t)this;
        cgul_exception_t local = NULL;
        rv = cgul_bdf_glyph__get_encoding(&local, glyph);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Get the ascent for the glyph.  This is a value derived from the
     * glyph's bounding box.
     *
     * @return ascent of glyph
     */
    unsigned int
    get_ascent() const
    {
        unsigned int rv = 0;
        cgul_bdf_glyph_t glyph = (cgul_bdf_glyph_t)this;
        cgul_exception_t local = NULL;
        rv = cgul_bdf_glyph__get_ascent(&local, glyph);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Get the descent for the glyph.  This is a value derived from the
     * glyph's bounding box.
     *
     * @return descent of glyph
     */
    unsigned int
    get_descent() const
    {
        unsigned int rv = 0;
        cgul_bdf_glyph_t glyph = (cgul_bdf_glyph_t)this;
        cgul_exception_t local = NULL;
        rv = cgul_bdf_glyph__get_descent(&local, glyph);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Get the scalable width in the x-direction.
     *
     * @return scalable width in the x-direction
     */
    unsigned int
    get_swidth_x() const
    {
        unsigned int rv = 0;
        cgul_bdf_glyph_t glyph = (cgul_bdf_glyph_t)this;
        cgul_exception_t local = NULL;
        rv = cgul_bdf_glyph__get_swidth_x(&local, glyph);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Get the scalable width in the y-direction.
     *
     * @return scalable width in the y-direction
     */
    unsigned int
    get_swidth_y() const
    {
        unsigned int rv = 0;
        cgul_bdf_glyph_t glyph = (cgul_bdf_glyph_t)this;
        cgul_exception_t local = NULL;
        rv = cgul_bdf_glyph__get_swidth_y(&local, glyph);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Get the device width in the x-direction.
     *
     * @return device width in the x-direction
     */
    unsigned int
    get_dwidth_x() const
    {
        unsigned int rv = 0;
        cgul_bdf_glyph_t glyph = (cgul_bdf_glyph_t)this;
        cgul_exception_t local = NULL;
        rv = cgul_bdf_glyph__get_dwidth_x(&local, glyph);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Get the device width in the y-direction.
     *
     * @return device width in the y-direction
     */
    unsigned int
    get_dwidth_y() const
    {
        unsigned int rv = 0;
        cgul_bdf_glyph_t glyph = (cgul_bdf_glyph_t)this;
        cgul_exception_t local = NULL;
        rv = cgul_bdf_glyph__get_dwidth_y(&local, glyph);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Get the width of the glyph's bounding box.
     *
     * @return width of the glyph's bounding box
     */
    unsigned int
    get_bb_width() const
    {
        unsigned int rv = 0;
        cgul_bdf_glyph_t glyph = (cgul_bdf_glyph_t)this;
        cgul_exception_t local = NULL;
        rv = cgul_bdf_glyph__get_bb_width(&local, glyph);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Get the height of the glyph's bounding box.
     *
     * @return height of the glyph's bounding box
     */
    unsigned int
    get_bb_height() const
    {
        unsigned int rv = 0;
        cgul_bdf_glyph_t glyph = (cgul_bdf_glyph_t)this;
        cgul_exception_t local = NULL;
        rv = cgul_bdf_glyph__get_bb_height(&local, glyph);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Get the x-offset of the lower-left corner of the glyph's bounding box.
     *
     * @return x-offset of the lower-left corner of the glyph's bounding box
     */
    int
    get_bb_x_offset() const
    {
        int rv = 0;
        cgul_bdf_glyph_t glyph = (cgul_bdf_glyph_t)this;
        cgul_exception_t local = NULL;
        rv = cgul_bdf_glyph__get_bb_x_offset(&local, glyph);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Get the y-offset of the lower-left corner of the glyph's bounding box.
     *
     * @return y-offset of the lower-left corner of the glyph's bounding box
     */
    int
    get_bb_y_offset() const
    {
        int rv = 0;
        cgul_bdf_glyph_t glyph = (cgul_bdf_glyph_t)this;
        cgul_exception_t local = NULL;
        rv = cgul_bdf_glyph__get_bb_y_offset(&local, glyph);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * This method returns <code>true</code> if the dot at
     * <code>(x,y)</code> in the glyph should be drawn.  The x- and
     * y-coordinates are relative to the top-left corner of the
     * glyph's bounding box, not the top-left corner of the glyph
     * itself.  Also note that <code>x</code> and <code>y</code>
     * values start with zero, not one.
     *
     * @param[in] x x-coordinate
     * @param[in] y y-coordinate
     * @return whether the dot should be drawn
     */
    int
    get_dot(unsigned int x,
            unsigned int y) const
    {
        int rv = 0;
        cgul_bdf_glyph_t glyph = (cgul_bdf_glyph_t)this;
        cgul_exception_t local = NULL;
        rv = cgul_bdf_glyph__get_dot(&local, glyph, x, y);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Get the width of the glyph in bytes.  Remember that the data is
     * a bitmap so it is packed with 8 pixels per byte.
     *
     * @return width of the glyph in bytes
     */
    size_t
    get_bytes_wide() const
    {
        size_t rv = 0;
        cgul_bdf_glyph_t glyph = (cgul_bdf_glyph_t)this;
        cgul_exception_t local = NULL;
        rv = cgul_bdf_glyph__get_bytes_wide(&local, glyph);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * This method appends the bitmap of the glyph as a character
     * sequence of hex digits to the string <code>s</code>.  The appended
     * character sequence can be used as the data in the BITMAP section of
     * a BDF file.  If an error occurs, an exception is thrown.
     *
     * @param[in] s string
     */
    void
    append_bitmap_to_string(const cgul_string_cxx& s)
    {
        cgul_bdf_glyph_t glyph = (cgul_bdf_glyph_t)this;
        cgul_exception_t local = NULL;
        cgul_bdf_glyph__append_bitmap_to_string(&local, glyph, s.get_obj());
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Write the string representation of the glyph to
     * <code>f</code>.  This is different from
     * <code>cgul_bdf_glyph__fdraw()</code> in that it produces a valid
     * snippet of BDF that can be used as the entire <code>STARTCHAR
     * ... ENDCHAR</code> section.
     *
     * @param[in] f file
     */
    void
    write_file(FILE* f)
    {
        cgul_bdf_glyph_t glyph = (cgul_bdf_glyph_t)this;
        cgul_exception_t local = NULL;
        cgul_bdf_glyph__write_file(&local, glyph, f);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * This method appends an ASCII image of the glyph to the file
     * <code>f</code>.  The image only shows the part of the glyph that is
     * in the bounding box.  Each active pixel is represented by the
     * character <code>c_active</code>.  Each inactive pixel is
     * represented by <code>c_inactive</code>.  If <code>prefix</code> is
     * not <code>NULL</code>, it will be prefixed to each scan line.  If
     * <code>suffix</code> is not <code>NULL</code>, it will be appended
     * to each scan line.  If an error occurs, an exception is thrown.
     *
     * @param[in] f file
     * @param[in] prefix prefix for each scan line
     * @param[in] suffix suffix for each scan line
     * @param[in] c_active active character
     * @param[in] c_inactive inactive character
     */
    void
    fdraw(FILE* f,
          const char* prefix = NULL,
          const char* suffix = NULL,
          char c_active = '*',
          char c_inactive = ' ')
    {
        cgul_bdf_glyph_t glyph = (cgul_bdf_glyph_t)this;
        cgul_exception_t local = NULL;
        cgul_bdf_glyph__fdraw(&local,
                              glyph,
                              f,
                              prefix,
                              suffix,
                              c_active,
                              c_inactive);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * This method appends an ASCII image of the glyph to the string
     * <code>s</code>.  The image only shows the part of the glyph that is
     * in the bounding box.  Each active pixel is represented by the
     * character <code>c_active</code>.  Each inactive pixel is
     * represented by <code>c_inactive</code>.  If <code>prefix</code> is
     * not <code>NULL</code>, it will be prefixed to each scan line.  If
     * <code>suffix</code> is not <code>NULL</code>, it will be appended
     * to each scan line.  If an error occurs, an exception is thrown.
     *
     * @param[in,out] s string
     * @param[in] prefix prefix for each scan line
     * @param[in] suffix suffix for each scan line
     * @param[in] c_active active character
     * @param[in] c_inactive inactive character
     */
    void
    sdraw(const cgul_string_cxx& s,
          const char* prefix = NULL,
          const char* suffix = NULL,
          char c_active = '*',
          char c_inactive = ' ')
    {
        cgul_bdf_glyph_t glyph = (cgul_bdf_glyph_t)this;
        cgul_exception_t local = NULL;
        cgul_bdf_glyph__sdraw(&local,
                              glyph,
                              s.get_obj(),
                              prefix,
                              suffix,
                              c_active,
                              c_inactive);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }


private:

    /* Hide these. */
    cgul_bdf_glyph_cxx();
    cgul_bdf_glyph_cxx(const cgul_bdf_glyph_cxx&);
    ~cgul_bdf_glyph_cxx();
    cgul_bdf_glyph_cxx& operator=(const cgul_bdf_glyph_cxx&);

};

#endif  /*  CGUL_BDF_GLYPH_CXX_H  */
