#ifndef CGUL_BLOCK_READER_H
#define CGUL_BLOCK_READER_H

/**
 * Interface that defines the ability to read blocks of data from an
 * external source.  The main reason for this interface is to abstract
 * reading blocks from a file or from memory. <p>
 *
 * Because this is an abstract base class, instances cannot be created
 * directly.  Instead, concrete instances can be created using the
 * following classes: <p>
 *
 * <pre>
 *     cgul_block_reader_file
 *     cgul_block_reader_memory
 * </pre>
 *
 * Perhaps a better solution would be to use the standard C
 * <code>FILE</code> class along with <code>fopen()</code> and
 * <code>fmemopen()</code>, but <code>fmemopen()</code> is not widely
 * available at the time of writing. <p>
 *
 * A similar class is <code>cgul_crlf_file</code> which returns lines
 * of text instead of blocks of data. <p>
 *
 * @file
 * @author Paul Serice
 * @brief block reader abstraction
 * @see cgul_crlf_file
 * @see cgul_stream_reader
 */

#include "cgul_common.h"
#include "cgul_exception.h"

CGUL_BEGIN_C

/** Opaque pointer to a <code>cgul_block_reader</code> instance. */
typedef struct cgul_block_reader* cgul_block_reader_t;

/**
 * Virtual destructor.  Derived classes should override to deallocate
 * resources as necessary.  Classes that implement this interface must
 * correctly chain their destructor with their parent class's
 * destructor.  Remember that <code>cgul_block_reader</code> is
 * conceptually an interface, not a class, so do not chain with it;
 * otherwise, you'll get infinite recursion because the methods
 * defined for the interface do nothing other than dispatch based on
 * the vtable. <p>
 *
 * @param[in] br <code>cgul_block_reader</code> instance
 */
extern CGUL_EXPORT void
cgul_block_reader__delete(cgul_block_reader_t br);

/**
 * Virtual method that reads one block of data <code>block</code> of
 * size <code>block_size</code> bytes from an external data source.
 * The user is responsible for making sure <code>block</code> points to
 * at least <code>block_size</code> contiguous bytes of memory.  This
 * method returns the number of bytes read or <code>0</code> if EOF is
 * reached without error.  If an error occurs, <code>0</code> is
 * returned, and an exception is thrown. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] br <code>cgul_block_reader</code> instance
 * @param[in] block block
 * @param[in] block_size block size in bytes
 * @return number of bytes read
 */
extern CGUL_EXPORT size_t
cgul_block_reader__read_next_block(cgul_exception_t* cex,
                                   cgul_block_reader_t br,
                                   void* block,
                                   size_t block_size);

/**
 * Virtual method that returns the presentation name.
 *
 * @param[in,out] cex c-style exception
 * @param[in] br <code>cgul_block_reader</code> instance
 * @return presentation name
 */
extern CGUL_EXPORT const char*
cgul_block_reader__get_presentation_name(cgul_exception_t* cex,
                                         cgul_block_reader_t br);

/**
 * Virtual method that sets the presentation name.  The presentation
 * name is primarily used for error reporting and can be set to
 * anything, but it does need to be set again if the input source
 * changes.  An exception is thrown if an attempt is made to set the
 * presentation name without first setting the input source.
 *
 * @param[in,out] cex c-style exception
 * @param[in] br <code>cgul_block_reader</code> instance
 * @param[in] presentation_name presentation name
 */
extern CGUL_EXPORT void
cgul_block_reader__set_presentation_name(cgul_exception_t* cex,
                                         cgul_block_reader_t br,
                                         const char* presentation_name);

CGUL_END_C

#endif  /*  CGUL_BLOCK_READER_H  */
