#ifndef CGUL_BLOCK_READER_CXX_H
#define CGUL_BLOCK_READER_CXX_H

/**
 * C++ wrapper for <code>cgul_block_reader</code>.
 *
 * @file
 * @author Paul Serice
 * @brief C++ wrapper for <code>cgul_block_reader</code>
 */

#include "cgul_block_reader.h"

/**
 * This class provides a proper C++ abstract base class for the
 * <code>cgul_block_reader</code> interface.
 *
 * @see cgul_block_reader_file_cxx
 * @see cgul_block_reader_memory_cxx
 */
class cgul_block_reader_cxx {

public:

    /**
     * Virtual destructor.
     */
    virtual
    ~cgul_block_reader_cxx()
    {
        /* empty */
    }

    /**
     * Virtual method that reads one block of data <code>block</code> of
     * size <code>block_size</code> bytes from an external data source.
     * The user is responsible for making sure <code>block</code> points to
     * at least <code>block_size</code> contiguous bytes of memory.  This
     * method returns the number of bytes read or <code>0</code> if EOF is
     * reached without error.  If an error occurs, <code>0</code> is
     * returned, and an exception is thrown. <p>
     *
     * @param[in] block block
     * @param[in] block_size block size in bytes
     * @return number of bytes read
     */
    virtual size_t
    read_next_block(void* block,
                    size_t block_size) = 0;

    /**
     * Virtual method that returns the presentation name.
     *
     * @return presentation name
     */
    virtual const char*
    get_presentation_name() const = 0;

    /**
     * Virtual method that sets the presentation name.  The presentation
     * name is primarily used for error reporting and can be set to
     * anything, but it does need to be set again if the input source
     * changes.  An exception is thrown if an attempt is made to set the
     * presentation name without first setting the input source.
     *
     * @param[in] presentation_name presentation name
     */
    virtual void
    set_presentation_name(const char* presentation_name) = 0;

    /**
     * Return the underlying <code>cgul_block_reader</code>
     * object. <p>
     *
     * @note Most of the C++ wrappers for cgul have a method called
     * <code>get_obj()</code> which is similar to this method except
     * it returns a pointer to the most derived class; whereas, this
     * method returns the same pointer but cast to the base class. <p>
     *
     * @return underlying <code>cgul_block_reader</code> object
     */
    virtual cgul_block_reader_t
    as_block_reader() const = 0;

};

#endif  /*  CGUL_BLOCK_READER_CXX_H  */
