#ifndef CGUL_BLOCK_READER_MEMORY_H
#define CGUL_BLOCK_READER_MEMORY_H

/**
 * Implementation of the <code>cgul_block_reader</code> interface for
 * data stored in memory.
 *
 * @file
 * @author Paul Serice
 * @brief memory block reader
 */

#include "cgul_common.h"
#include "cgul_exception.h"
#include "cgul_block_reader.h"

CGUL_BEGIN_C

/** Opaque pointer to a <code>cgul_block_reader_memory</code> instance. */
typedef struct cgul_block_reader_memory* cgul_block_reader_memory_t;

/**
 * Safely upcast <code>brmemory</code> to the
 * <code>cgul_block_reader</code> interface.
 *
 * @param[in] cex c-style exception
 * @param[in] brmemory <code>cgul_block_reader_memory</code> instance
 * @see cgul_block_reader
 */
extern CGUL_EXPORT cgul_block_reader_t
cgul_block_reader_memory__as_block_reader(cgul_exception_t* cex,
                                          cgul_block_reader_memory_t brmemory);

/**
 * Create a new <code>cgul_block_reader_memory</code> instance.  After
 * this method returns, <code>cgul_block_reader_memory__open()</code>
 * should be called before calling
 * <code>cgul_block_reader_memory__read_next_block()</code>.  The
 * client is responsible for calling
 * <code>cgul_block_reader_memory__delete()</code> on the object
 * returned.  If an error occurs, <code>NULL</code> is returned, and
 * an exception is thrown. <p>
 *
 * @note It is safe to cast the object returned directly to a
 * <code>cgul_block_reader</code> (i.e., to the base type). <p>
 *
 * @param[in,out] cex c-style exception
 * @return new <code>cgul_block_reader_memory</code> instance
 */
extern CGUL_EXPORT cgul_block_reader_memory_t
cgul_block_reader_memory__new(cgul_exception_t* cex);

/**
 * This method creates a new <code>cgul_block_reader_memory</code>
 * instance and calls <code>cgul_block_reader_memory__open()</code>
 * passing it <code>buffer</code> and <code>buffer_size</code>.
 * This class does not take ownership of <code>buffer</code> so the
 * client is still responsible for freeing <code>buffer</code> if
 * necessary.  The client is also responsible for calling
 * <code>cgul_block_reader_memory__delete()</code> on the object
 * returned.  If an error occurs, <code>NULL</code> is returned, and
 * an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] buffer memory buffer
 * @param[in] buffer_size size of <code>buffer</code> in bytes
 * @return new <code>cgul_block_reader_memory</code> instance
 */
extern CGUL_EXPORT cgul_block_reader_memory_t
cgul_block_reader_memory__new_from_memory(cgul_exception_t* cex,
                                          void* buffer,
                                          size_t buffer_size);

/**
 * This virtual method deletes the <code>brmemory</code> instance
 * freeing all internally allocated resources.  The client must not
 * use <code>brmemory</code> after calling this method.
 *
 * @param[in] brmemory <code>cgul_block_reader_memory</code> instance
 */
extern CGUL_EXPORT void
cgul_block_reader_memory__delete(cgul_block_reader_memory_t brmemory);

/**
 * Set the buffer <code>buffer</code> to use when reading blocks of
 * data from memory.  If a memory buffer is already open, it will be
 * closed before this method attempts to open the new memory buffer.
 * This class does not take ownership of <code>buffer</code>.  Thus,
 * the client is still responsible for freeing <code>buffer</code> if
 * necessary.  If an error occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] brmemory <code>cgul_block_reader_memory</code> instance
 * @param[in] buffer input memory buffer
 * @param[in] buffer_size size of <code>buffer</code> in bytes
 */
extern CGUL_EXPORT void
cgul_block_reader_memory__open(cgul_exception_t* cex,
                               cgul_block_reader_memory_t brmemory,
                               void* buffer,
                               size_t buffer_size);

/**
 * Close the memory.  After this method returns,
 * <code>cgul_block_reader_memory__open</code> should be called before
 * calling any other method.
 *
 * @param[in,out] cex c-style exception
 * @param[in] brmemory <code>cgul_block_reader_memory</code> instance
 */
extern CGUL_EXPORT void
cgul_block_reader_memory__close(cgul_exception_t* cex,
                                cgul_block_reader_memory_t brmemory);

/**
 * Virtual method that reads one block of data <code>block</code> of
 * size <code>block_size</code> bytes from the underlying buffer.  The
 * user is responsible for making sure <code>block</code> points to at
 * least <code>block_size</code> contiguous bytes of memory.  The user
 * is also responsible for making sure <code>block</code> does not
 * overlap with the open buffer from which blocks are read (as
 * required by <code>memcpy()</code>).  This method returns the number
 * of bytes read or <code>0</code> if EOF is reached without error.
 * If an error occurs, <code>0</code> is returned, and an exception is
 * thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] brmemory <code>cgul_block_reader_memory</code> instance
 * @param[in] block block
 * @param[in] block_size block size in bytes
 * @return number of bytes read
 */
extern CGUL_EXPORT size_t
cgul_block_reader_memory__read_next_block(cgul_exception_t* cex,
                                          cgul_block_reader_memory_t brmemory,
                                          void* block,
                                          size_t block_size);

/**
 * Virtual method that returns the presentation name.
 *
 * @param[in,out] cex c-style exception
 * @param[in] brmemory <code>cgul_block_reader_memory</code> instance
 * @return presentation name
 */
extern CGUL_EXPORT const char*
cgul_block_reader_memory__get_presentation_name(
    cgul_exception_t* cex,
    cgul_block_reader_memory_t brmemory);

/**
 * Virtual method that sets the presentation name.  The presentation
 * name is primarily used for error reporting and can be set to
 * anything, but it does need to be set again if the input buffer
 * changes.  If the presentation name is not set,
 * <code>"MEMORY"</code> is returned.  An exception is thrown if an
 * attempt is made to set the presentation name without first setting
 * the input buffer.
 *
 * @param[in,out] cex c-style exception
 * @param[in] brmemory <code>cgul_block_reader_memory</code> instance
 * @param[in] presentation_name presentation name
 */
extern CGUL_EXPORT void
cgul_block_reader_memory__set_presentation_name(
    cgul_exception_t* cex,
    cgul_block_reader_memory_t brmemory,
    const char* presentation_name);

/**
 * Return the currently opened buffer from which blocks are read.  The
 * client must not do anything that would invalidate the pointer or
 * corrupt the associated data stream.  This method throws an
 * exception only if a file is not currently open. <p>
 *
 * Because the client is responsible for freeing the buffer if
 * necessary, it can be convenient to call this method just before
 * deleting the object in order to free the pointer returned. <p>
 *
 * @param[in] cex c-style exception
 * @param[in] brmemory <code>cgul_block_reader_memory</code> instance
 * @return currently opened file or memory
 */
extern CGUL_EXPORT void*
cgul_block_reader_memory__get_buffer(cgul_exception_t* cex,
                                     cgul_block_reader_memory_t brmemory);

CGUL_END_C

#endif  /*  CGUL_BLOCK_READER_MEMORY_H  */
