#ifndef CGUL_CRLF_FILE_CXX_H
#define CGUL_CRLF_FILE_CXX_H

/**
 * C++ wrapper for <code>cgul_crlf_file</code>.
 *
 * @file
 * @author Paul Serice
 * @brief C++ wrapper for <code>cgul_crlf_file</code>
 */

#include "cgul_crlf_file.h"
#include "cgul_exception_cxx.h"

/**
 * This class provides the C++ bindings for C <code>cgul_crlf_file</code>
 * objects.  The main purpose of this class is to convert the C-style
 * function calls and exception handling in <code>cgul_crlf_file</code>
 * into C++-style function calls and exception handling.
 *
 * @brief C++ bindings for <code>cgul_crlf_file</code>
 */
class cgul_crlf_file_cxx {

public:

    /**
     * This method creates a new <code>cgul_crlf_file_cxx</code>
     * instance.  After this method returns, <code>open_fname</code>,
     * <code>open_file()</code>, or <code>open_memory()</code> should
     * be called before calling any other method.  If an error occurs,
     * an exception is thrown.
     */
    cgul_crlf_file_cxx()
    {
        cgul_exception_t local = NULL;
        obj = cgul_crlf_file__new(&local);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * This method creates a new <code>cgul_crlf_file_cxx</code>
     * instance and calls <code>open()</code> passing it
     * <code>fname</code>.  The file will be closed when this instance
     * is deleted.  If an error occurs, an exception is thrown. <p>
     *
     * If <code>is_block_buffered</code> is true, this method results
     * in a very fast <code>cgul_crlf_file_cxx</code> instance, but
     * the underlying file should be a regular file or any other file
     * that is capable of block buffering.  If the file is not capable
     * of block buffering, <code>is_block_buffered</code> should be
     * false which results in a much slower
     * <code>cgul_crlf_file_cxx</code> instance, but one that should
     * be able to read from any type of file.  As a general rule, if
     * you are reading from a regular file or are working in batch
     * mode, <code>is_block_buffered</code> should be true to get the
     * best performance.  If you are reading from a line-oriented or
     * character-oriented device like a terminal or from a named pipe
     * across which small messages are being passed,
     * <code>is_block_buffered</code> should be false. <p>
     *
     * @param[in] fname file name
     * @param[in] is_block_buffered whether the file is block buffered
     */
    cgul_crlf_file_cxx(const char* fname,
                       bool is_block_buffered = true)
    {
        cgul_exception_t local = NULL;
        obj = cgul_crlf_file__new_from_fname(&local,
                                             fname,
                                             is_block_buffered ? 1 : 0);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * This method creates a new <code>cgul_crlf_file_cxx</code>
     * instance.  This class does not take ownership of
     * <code>f</code>.  Thus, the client is still responsible for
     * calling <code>fclose()</code> on <code>f</code>.  If an error
     * occurs, an exception is thrown.
     *
     * If <code>is_block_buffered</code> is true, this method results
     * in a very fast <code>cgul_crlf_file_cxx</code> instance, but
     * the underlying file should be a regular file or any other file
     * that is capable of block buffering.  If the file is not capable
     * of block buffering, <code>is_block_buffered</code> should be
     * false which results in a much slower
     * <code>cgul_crlf_file_cxx</code> instance, but one that should
     * be able to read from any type of file.  As a general rule, if
     * you are reading from a regular file or are working in batch
     * mode, <code>is_block_buffered</code> should be true to get the
     * best performance.  If you are reading from a line-oriented or
     * character-oriented device like a terminal or from a named pipe
     * across which small messages are being passed,
     * <code>is_block_buffered</code> should be false. <p>
     *
     * @param[in] f underlying file
     * @param[in] is_block_buffered whether the file is block buffered
     */
    cgul_crlf_file_cxx(FILE* f,
                       bool is_block_buffered = true)
    {
        cgul_exception_t local = NULL;
        obj = cgul_crlf_file__new_from_file(&local,
                                            f,
                                            is_block_buffered ? 1 : 0);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * This method creates a new <code>cgul_crlf_file_cxx</code>
     * instance and calls <code>open()</code> passing it
     * <code>buf</code> and <code>buf_size</code>.  This class does
     * not take ownership of <code>buf</code> so the client is still
     * responsible for freeing <code>buf</code> if necessary.  If an
     * error occurs, an exception is thrown. <p>
     *
     * @param[in] buf memory buffer
     * @param[in] buf_size size of <code>buf</code> in bytes
     */
    cgul_crlf_file_cxx(const char* buf,
                       size_t buf_size)
    {
        cgul_exception_t local = NULL;
        obj = cgul_crlf_file__new_from_memory(&local, buf, buf_size);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * This method deletes the instance freeing all internally
     * allocated resources.  This does not include closing the
     * underlying file if it was obtained from the client by a call to
     * <code>cgul_crlf_file_cxx(FILE*)</code>.
     */
    virtual
    ~cgul_crlf_file_cxx()
    {
        if (obj) {
            cgul_crlf_file__delete(obj);
            obj = NULL;
        }
    }

    /**
     * Open the file with name <code>fname</code> and use it when
     * getting lines.  If a file is already open, it will be closed
     * before this method attempts to open the new file.  The new file
     * will be closed when this instance is deleted.  If an error
     * occurs, an exception is thrown. <p>
     *
     * If <code>is_block_buffered</code> is true, this method results
     * in a very fast <code>cgul_crlf_file_cxx</code> instance, but
     * the underlying file should be a regular file or any other file
     * that is capable of block buffering.  If the file is not capable
     * of block buffering, <code>is_block_buffered</code> should be
     * false which results in a much slower
     * <code>cgul_crlf_file_cxx</code> instance, but one that should
     * be able to read from any type of file.  As a general rule, if
     * you are reading from a regular file or are working in batch
     * mode, <code>is_block_buffered</code> should be true to get the
     * best performance.  If you are reading from a line-oriented or
     * character-oriented device like a terminal or from a named pipe
     * across which small messages are being passed,
     * <code>is_block_buffered</code> should be false. <p>
     *
     * @param[in] fname file name
     * @param[in] is_block_buffered whether the file is block buffered
     */
    virtual void
    open_fname(const char* fname,
               bool is_block_buffered = true)
    {
        cgul_exception_t local = NULL;
        cgul_crlf_file__open_fname(&local,
                                   obj,
                                   fname,
                                   is_block_buffered ? 1 : 0);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Use the file <code>f</code> when getting lines.  If a file is
     * already open, it will be closed before this method attempts to
     * open the new file.  This class does not take ownership of
     * <code>f</code>.  Thus, the client is still responsible for
     * calling <code>fclose()</code> on <code>f</code>.  If an error
     * occurs, an exception is thrown. <p>
     *
     * If <code>is_block_buffered</code> is true, this method results
     * in a very fast <code>cgul_crlf_file_cxx</code> instance, but
     * the underlying file should be a regular file or any other file
     * that is capable of block buffering.  If the file is not capable
     * of block buffering, <code>is_block_buffered</code> should be
     * false which results in a much slower
     * <code>cgul_crlf_file_cxx</code> instance, but one that should
     * be able to read from any type of file.  As a general rule, if
     * you are reading from a regular file or are working in batch
     * mode, <code>is_block_buffered</code> should be true to get the
     * best performance.  If you are reading from a line-oriented or
     * character-oriented device like a terminal or from a named pipe
     * across which small messages are being passed,
     * <code>is_block_buffered</code> should be false. <p>
     *
     * @param[in] f file
     * @param[in] is_block_buffered whether the file is block buffered
     */
    virtual void
    open_file(FILE* f,
              bool is_block_buffered = true)
    {
        cgul_exception_t local = NULL;
        cgul_crlf_file__open_file(&local,
                                  obj,
                                  f,
                                  is_block_buffered ? 1 : 0);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Use the memory buffer <code>buf</code> when getting lines.  If
     * a data source is already open, it will be closed before this
     * method attempts to open the new data source.  This class does
     * not take ownership of <code>buf</code>.  Thus, the client is
     * still responsible for freeing <code>buf</code> if necessary.
     * If an error occurs, an exception is thrown. <p>
     *
     * @param[in] buf memory buffer
     * @param[in] buf_size size of <code>buf</code> in bytes
     */
    virtual void
    open_memory(const char* buf,
                size_t buf_size)
    {
        cgul_exception_t local = NULL;
        cgul_crlf_file__open_memory(&local, obj, buf, buf_size);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Close the file.  After this method returns, <code>open()</code>
     * should be called before calling any other method.
     */
    virtual void
    close() {
        cgul_exception_t local = NULL;
        cgul_crlf_file__close(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Return the name of the currently opened file.  If a
     * <code>FILE*</code> was opened instead of a file name,
     * <code>"FILE"</code> will be used.  If memory was opened instead
     * of a file name, <code>"MEMORY"</code> will be used.  The client
     * must not attempt to free the pointer returned.  This method
     * throws an exception only if a file is not currently open.
     *
     * @return name of the currently opened file
     */
    virtual const char*
    get_file_name() const
    {
        const char* rv = NULL;
        cgul_exception_t local = NULL;
        rv = cgul_crlf_file__get_fname(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Return the currently opened file.  If a file name or
     * <code>FILE*</code> was opened, a <code>FILE*</code> is returned.
     * If memory was opened, a <code>char*</code> is returned.  The client
     * must not do anything that would invalidate the pointer or corrupt
     * the associated data stream.  This method throws an exception only
     * if a file is not currently open.
     *
     * @return currently opened file or memory
     */
    virtual const void*
    get_file() const
    {
        const void* rv = NULL;
        cgul_exception_t local = NULL;
        rv = cgul_crlf_file__get_file(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * This method returns whether the leading UTF-8 byte-order mark (BOM)
     * should be removed from the first line if it is present.
     *
     * @return whether to strip a leading UTF-8 byte-order mark
     */
    virtual int
    get_strip_utf8_bom() const
    {
        int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_crlf_file__get_strip_utf8_bom(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * By default, this class detects the leading UTF-8 byte-order
     * mark (BOM) and strips it from the first line returned by
     * <code>get_line()</code> if it is present.  It then clears its
     * internal flag so that BOMs internal to the text file will be
     * returned.  This is generally what you want because the leading
     * BOM is not significant but the internal BOMs are. <p>
     *
     * You can alter the way this class handles the leading BOM by calling
     * this method with <code>strip_utf8_bom</code> set to <code>0</code>.
     * This will cause the leading BOM to be returned as part of the first
     * line.  This can be useful, for example, if you just want to convert
     * the text file and are not interested in its contents. <p>
     *
     * It should be noted that most operating systems do not save UTF-8
     * text files with a leading BOM because UTF-8 is a character stream
     * and, as such, does not have byte-order problems; however, Microsoft
     * Windows adds the BOM to its UTF-8 text files presumably to help
     * distinguish UTF-8 text files from text files with different
     * encodings. <p>
     *
     * @param[in] strip_utf8_bom whether to strip leading UTF-8 byte-order mark
     */
    virtual void
    set_strip_utf8_bom(int strip_utf8_bom)
    {
        cgul_exception_t local = NULL;
        cgul_crlf_file__set_strip_utf8_bom(&local, obj, strip_utf8_bom);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Return the array of lines from the text file <code>fname</code> in
     * <code>*lines</code> and the number of lines in
     * <code>*line_count</code>.  Each line is stored in the array as a
     * C-style string.  The client is responsible for calling
     * <code>free()</code> on each line.  If an error occurs, an exception
     * is thrown. <p>
     *
     * To use this function, do something like the following: <p>
     *
     * <pre>
     *
     *     char** lines = NULL;
     *     unsigned long int line_count = 0;
     *     crlf.get_lines(&lines, &line_count);
     *
     * </pre>
     *
     * @param[out] lines ordered list of lines from the file
     * @param[out] line_count number of lines
     * @return list of lines from the file
     */
    void
    get_lines(char*** lines,
              unsigned long int* line_count)
    {
        cgul_exception_t local = NULL;
        cgul_crlf_file__get_lines(&local, obj, lines, line_count);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * This method returns the next line of text from the underlying
     * DOS, Mac, or Unix text file.  The caller must not call
     * <code>free()</code> on the pointer returned as it typically
     * points into a larger block owned by this class.  If EOF is
     * reached, <code>NULL</code> is returned.  If an error occurs, an
     * exception is thrown.
     *
     * @return next line of text
     */
    virtual const char*
    get_line()
    {
        const char* rv = NULL;
        cgul_exception_t local = NULL;
        rv = cgul_crlf_file__get_line(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Get the line count for the last line returned by
     * <code>get_line</code>.  The line count is one-based.  No
     * attempt is made to prevent the return value from overflowing.
     * So, the caller is responsible for verifying the return value.
     *
     * @return line count
     */
    virtual unsigned long
    get_line_count() const
    {
        unsigned long rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_crlf_file__get_line_count(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Get the line offset for the last line returned by
     * <code>cgul_crlf_file__get_line</code>.  The offset is zero-based.
     * If the underlying file is binary and random access, you can use the
     * return value to directly seek to the line as follows:
     *
     * <pre>
     *     f.fseek(offset, SEEK_SET);
     * </pre>
     *
     * Because the prototype for <code>fseek()</code> requires a
     * <code>long</code> for the offset parameter, no attempt is made to
     * prevent the return value from overflowing.  So, the caller is
     * responsible for verifying the return value. <p>
     *
     * Note that the offset returned is basically the number of bytes from
     * the start of the file to the current line.  This is not necessarily
     * the same as the number of characters which depends on how the file
     * is encoded. <p>
     *
     * @return file offset
     */
    virtual unsigned long
    get_line_offset() const
    {
        unsigned long rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_crlf_file__get_line_offset(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * This method calls <code>cgul_libc__fseek()</code> on the
     * underlying <code>FILE*</code> object and resets the underlying
     * <code>cgul_crlf</code> object so that reading new lines can
     * continue at <code>offset</code> which is relative to
     * <code>whence</code>.  If an error occurs, an exception is
     * thrown. <p>
     *
     * You can call <code>get_line_offset()</code> immediately after
     * calling <code>get_line()</code> to determine the offset for the
     * beginning of the last line returned. <p>
     *
     * @param[in] offset offset
     * @param[in] whence position to which offset is relative
     */
    virtual void
    fseek(unsigned long offset,
          int whence)
    {
        cgul_exception_t local = NULL;
        cgul_crlf_file__fseek(&local, obj, offset, whence);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * This method calls <code>cgul_libc__rewind()</code> on the
     * underlying <code>FILE*</code> object and resets the underlying
     * <code>cgul_crlf</code> object so that reading new lines can
     * continue at the beginning of the file.  This method throws an
     * exception if either <code>open</code> or
     * <code>open_file()</code> has not been called.  This method also
     * throws an exception if the underlying file is not seekable.
     */
    virtual void
    rewind()
    {
        cgul_exception_t local = NULL;
        cgul_crlf_file__rewind(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * This method returns the size of the buffer used to read blocks out
     * of the underlying file.
     *
     * @return buffer size
     */
    virtual unsigned long
    get_buffer_size() const
    {
        unsigned long rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_crlf_file__get_buffer_size(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * This method sets the size of the buffer used to read blocks out of
     * the underlying file.  By default, the buffer size is 16K which is
     * efficient for processing large files en masse but may not be
     * efficient when randomly accessing small parts of the same file.  If
     * an error occurs allocating the new buffer, an exception is thrown,
     * and the original buffer will continue to be used.
     *
     * @param[in] bsize new buffer size
     */
    virtual void
    set_buffer_size(unsigned long bsize)
    {
        cgul_exception_t local = NULL;
        cgul_crlf_file__set_buffer_size(&local, obj, bsize);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Get the underlying <code>cgul_crlf_file</code> object.
     *
     * @return underlying object
     */
    virtual cgul_crlf_file_t
    get_obj() const
    {
        return obj;
    }

    /**
     * Take the underlying <code>cgul_crlf_file</code> object.  This
     * means the underlying object will not be deleted when the
     * wrapper goes out of scope.  Also, because you have taken the
     * underlying object, no other methods should be called on this
     * wrapper's instance.  Lastly, after taking the underlying
     * object, it is the caller's responsibility to delete the
     * underlying object by calling
     * <code>cgul_crlf_file__delete()</code>.
     *
     * @return underlying object
     */
    virtual cgul_crlf_file_t
    take_obj()
    {
        cgul_crlf_file_t rv = obj;
        obj = NULL;
        return rv;
    }

    /**
     * Set the new underlying object to <code>rhs</code>.  This causes
     * the old underlying object to be deleted which invalidates any
     * outstanding pointers to or iterators for the old underlying
     * object. <p>
     *
     * This instance takes ownership of <code>rhs</code> which means
     * <code>rhs</code> will be automatically deleted when the C++
     * wrapper is deleted.  To prevent automatic deletion of
     * <code>rhs</code>, call <code>take_obj()</code> when the C++
     * wrapper is no longer needed. <p>
     *
     * @param[in] rhs right-hand side
     */
    virtual void
    set_obj(cgul_crlf_file_t rhs)
    {
        if (rhs && (rhs != obj)) {
            if (obj) {
                cgul_crlf_file__delete(obj);
            }
            obj = rhs;
        }
    }

private:

    /* Hide these. */
    cgul_crlf_file_cxx(const cgul_crlf_file_cxx&);
    cgul_crlf_file_cxx& operator=(const cgul_crlf_file_cxx&);

    /* Underlying <code>cgul_crlf_file</code> object. */
    cgul_crlf_file_t obj;

};

#endif  /*  CGUL_CRLF_FILE_CXX_H  */
