#ifndef CGUL_EVENT_ROUTER_CXX_H
#define CGUL_EVENT_ROUTER_CXX_H

/**
 * C++ wrapper for <code>cgul_event_router</code>.
 *
 * @file
 * @author Paul Serice
 * @brief C++ wrapper for <code>cgul_event_router</code>
 */

#include "cgul_exception_cxx.h"
#include "cgul_event_router.h"

/**
 * This class provides the C++ bindings for C
 * <code>cgul_event_router</code> objects.  The main purpose of this class
 * is to convert the C-style function calls and exception handling in
 * <code>cgul_event_router</code> into C++-style function calls and
 * exception handling.
 *
 * @brief C++ bindings for <code>cgul_event_router</code>
 */
class cgul_event_router_cxx {

public:

    /*
     * IMPLEMENTATION NOTE:
     * ===================
     *
     * There are no real cgul_event_router_cxx objects.  Instead, this
     * class is carefully implemented so that "cgul_event_router"
     * objects (not "cgul_event_router_cxx" objects) can be directly
     * used to invoke the methods of this class.
     *
     * This works because this class has no associated data _and_ no
     * virtual methods which would cause a vtable lookup, and vtables
     * are typically embedded in the object itself.
     *
     * In summary, do _not_ add data or virtual functions to this
     * class.
     */

    /**
     * In order to register to listen to events, you must implement a
     * function that is of type <code>callback_t</code>. <p>
     *
     * @see cgul_event_router__callback_t
     */
    typedef cgul_event_router__callback_t callback_t;

    /**
     * Opaque pointer a <code>listener</code> instance.  An object of
     * this type is passed back to the user by
     * <code>add_listener()</code>.  It can be used to remove a
     * listener or change the data passed to a listener.
     */
    typedef cgul_event_router__listener_t listener_t;

    /**
     * Return the single instance of this class.  Because
     * <code>cgul_event_router_cxx</code> is a singleton, the caller
     * is generally <b>not</b> responsible for cleaning up by
     * arranging for the instance to be deleted.  If memory cannot be
     * allocated, an exception is thrown. <p>
     *
     * @return single instance of this class.
     */
    static cgul_event_router_cxx*
    get_instance()
    {
        cgul_event_router_cxx* rv = NULL;
        cgul_exception_t local = NULL;
        rv = (cgul_event_router_cxx*)cgul_event_router__get_instance(&local);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Destructor for the <code>cgul_event_router_cxx</code> instance.
     * In a typical program, the instance should only be deleted once
     * (if at all) immediately before the program exists. <p>
     */
    static void
    delete_instance()
    {
        cgul_event_router_t er = (cgul_event_router_t)get_instance();
        cgul_event_router__delete(er);
    }

    /**
     * This method adds <code>cb</code> to listen for the event specified
     * by the rendezvous string <code>rendezvous</code>.  A copy of
     * <code>rendezvous</code> is made so the client does not have to keep
     * <code>rendezvous</code> in scope. <p>
     *
     * Events will be sent to the listeners in the order in which the
     * listeners were added. <p>
     *
     * The "all" rendezvous string is special; if you use it, your
     * listener will receive all events without having to explicitly
     * register for each one.  Furthermore, the "all" event listeners will
     * receive the events before the normal event listeners.  This makes
     * the output of a logger that is attached to the "all" event sane
     * when your code recursively sends events. <p>
     *
     * The <code>cb</code> callback can be added more than once per
     * rendezvous string.  This makes it possible to use this class with
     * C++ code when using a single private "static" method to forward the
     * callback to a member method for more than one object of the C++
     * class.  When using this tactic, each C++ object has be to
     * registered individually, and this class makes it possible to do
     * so. <p>
     *
     * <code>cb</code> should return true if it wants the event to be sent
     * to all remaining listeners.  If <code>cb</code> returns false, the
     * event will not be sent to any function that was added after
     * <code>cb</code>. <p>
     *
     * This class is reentrant so it is safe to recursively add or remove
     * listeners or even send events from <code>cb</code>.  <p>
     *
     * If an error occurs, an exception will be thrown. <p>
     *
     * @param[in] rendezvous rendezvous string
     * @param[in] cb callback function
     * @param[in] data opaque client data to pass back to <code>cb</code>
     * @return listener
     */
    listener_t
    add_listener(const char* rendezvous,
                 callback_t cb,
                 const void* data)
    {
        listener_t rv = NULL;
        cgul_event_router_t er = (cgul_event_router_t)this;
        cgul_exception_t local = NULL;
        rv = cgul_event_router__add_listener(&local,
                                             er,
                                             rendezvous,
                                             cb,
                                             data);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Remove <code>listener</code>.  After calling this method,
     * <code>listener</code> is invalid and should not be used again by
     * the client. <p>
     *
     * This class is reentrant so it is safe to recursively add or remove
     * listeners or even send events from the callback associated with
     * <code>listener</code>.  <p>
     *
     * @param[in] listener listener
     */
    void
    remove_listener(listener_t listener)
    {
        cgul_event_router_t er = (cgul_event_router_t)this;
        cgul_exception_t local = NULL;
        cgul_event_router__remove_listener(&local, er, listener);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Send <code>event</code> to all listeners registered for events
     * associated with the <code>rendezvous</code> string.  Events will be
     * sent to the listeners in the order in which the listeners were
     * added.  This class is reentrant so it is safe to call this method
     * from the callbacks associated with listeners thereby recursively
     * sending events.  If an error occurs, an exception will be thrown.
     *
     * @param[in] rendezvous rendezvous string
     * @param[in] event event to send
     */
    void
    send_event(const char* rendezvous,
               const void* event)
    {
        cgul_event_router_t er = (cgul_event_router_t)this;
        cgul_exception_t local = NULL;
        cgul_event_router__send_event(&local,
                                      er,
                                      rendezvous,
                                      event);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * This method sets the client data that is passed back to
     * <code>listener</code> when it is sent an event.
     *
     * @param[in] listener listener
     * @param[in] data client data
     */
    void
    set_data(listener_t listener,
             const void* data)
    {
        cgul_event_router_t er = (cgul_event_router_t)this;
        cgul_exception_t local = NULL;
        cgul_event_router__set_data(&local, er, listener, data);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Get the count of listeners currently registered for the event
     * associated with the <code>rendezvous</code> string. <p>
     *
     * @param[in] rendezvous rendezvous string
     * @return listener count
     */
    size_t
    get_listener_count(const char* rendezvous) const
    {
        size_t rv = 0;
        cgul_event_router_t er = (cgul_event_router_t)this;
        cgul_exception_t local = NULL;
        rv = cgul_event_router__get_listener_count(&local,
                                                   er,
                                                   rendezvous);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

private:

    /* Hide these. */
    cgul_event_router_cxx();
    cgul_event_router_cxx(const cgul_event_router_cxx&);
    ~cgul_event_router_cxx();
    cgul_event_router_cxx& operator=(const cgul_event_router_cxx&);

};

#endif  /*  CGUL_EVENT_ROUTER_CXX_H  */
