#ifndef CGUL_EXCEPTION_CXX_H
#define CGUL_EXCEPTION_CXX_H

/**
 * C++ wrapper for <code>cgul_exception</code>.
 *
 * @file
 * @author Paul Serice
 * @brief C++ wrapper for <code>cgul_exception</code>
 */

#include "cgul_exception.h"

/**
 * This class provides the C++ bindings for C
 * <code>cgul_exception</code> objects.  The main purpose of this
 * class is to convert the C-style function calls into C++-style
 * function calls and to provide a C++ object that is suitable for
 * being used like any other C++ exception. <p>
 *
 * This class does not inherit from <code>std::exception</code>.
 * While that would be nice, this is just a major source of
 * portability problems because the header and namespace for
 * "exception" is almost never the same for older but still
 * commonly-used C++ compilers. <p>
 *
 * @brief C++ bindings for <code>cgul_exception</code>
 * @see cgul_exception_t
 */
class cgul_exception_cxx {

public:

    /**
     * Construct a new <code>cgul_exception_cxx</code> object using a
     * printf()-style format string.
     *
     * @param[in] format printf()-style format string
     * @param[in] ... variable number of arguments that satisfy the
     * format string
     */
    cgul_exception_cxx(const char* format, ...)
#if defined(__GNUC__) && !defined(__MINGW_H)
    /* The index values have to account for the implicit "this" pointer. */
    __attribute__ ((format (printf, 2, 3)))
#endif
    {
        va_list args;
        va_start(args, format);
        obj = cgul_exception__new_from_va_list(format, args);
        va_end(args);
    }

    /**
     * Construct a new <code>cgul_exception_cxx</code> object using
     * the vprintf()-style variable argument list <code>args</code>.
     *
     * @param[in] format printf()-style format string
     * @param args variable number of arguments that satisfy
     * <code>format</code>
     */
    cgul_exception_cxx(const char* format, va_list args)
    {
        obj = cgul_exception__new_from_va_list(format, args);
    }

    /**
     * Construct a new <code>cgul_exception_cxx</code> object by
     * wrapping and taking ownership of an existing
     * <code>cgul_exception</code>.
     *
     * @param[in] rhs right-hand side
     */
    cgul_exception_cxx(cgul_exception_t rhs)
    {
        /* Take ownership of the pointer. */
        obj = rhs;
    }

    /**
     * Copy constructor.
     *
     * @param[in] rhs right-hand side
     */
    cgul_exception_cxx(const cgul_exception_cxx& rhs)
    {
        /* This object must be initialized before make calls to other
         * methods. */
        obj = NULL;

        /* Use operator=() to copy obj to this object. */
        *this = rhs;
    }

    /**
     * Destructor.
     */
    virtual
    ~cgul_exception_cxx()
    {
        if (obj) {
            cgul_exception__delete(obj);
            obj = NULL;
        }
    }

    /**
     * This method assigns the exception in <code>rhs</code> to this
     * object and then return this object.
     *
     * @param[in] rhs right-hand side
     * @return *this
     */
    cgul_exception_cxx& operator=(const cgul_exception_cxx& rhs)
    {
        if (this != &rhs) {
            /* Delete the underlying exception. */
            if (obj) {
                cgul_exception__delete(obj);
            }
            /* Create a new underlying exception based on rhs. */
            obj = cgul_exception__new("%s", rhs.get_message());
        }
        return *this;
    }

    /**
     * Get the message associated with the exception.
     *
     * @return message associated with the exception
     */
    const char*
    get_message() const
    {
        return cgul_exception__get_message(obj);
    }

    /**
     * Get the underlying <code>cgul_exception</code> object.
     *
     * @return underlying object
     */
    virtual cgul_exception_t
    get_obj() const
    {
        return obj;
    }

    /**
     * Take the underlying <code>cgul_exception</code> object.  This
     * means the underlying object will not be deleted when the
     * wrapper goes out of scope.  Also, because you have taken the
     * underlying object, no other methods should be called on this
     * wrapper's instance.  Lastly, after taking the underlying
     * object, it is the caller's responsibility to delete the
     * underlying object by calling
     * <code>cgul_exception__delete()</code>.
     *
     * @return underlying object
     */
    virtual cgul_exception_t
    take_obj()
    {
        cgul_exception_t rv = obj;
        obj = NULL;
        return rv;
    }

    /**
     * Set the new underlying object to <code>rhs</code>.  This causes
     * the old underlying object to be deleted which invalidates any
     * outstanding pointers to or iterators for the old underlying
     * object. <p>
     *
     * This instance takes ownership of <code>rhs</code> which means
     * <code>rhs</code> will be automatically deleted when the C++
     * wrapper is deleted.  To prevent automatic deletion of
     * <code>rhs</code>, call <code>take_obj()</code> when the C++
     * wrapper is no longer needed. <p>
     *
     * @param[in] rhs right-hand side
     */
    virtual void
    set_obj(cgul_exception_t rhs)
    {
        if (rhs && (rhs != obj)) {
            if (obj) {
                cgul_exception__delete(obj);
            }
            obj = rhs;
        }
    }

private:

    /* Hide these. */
    cgul_exception_cxx();

    /** Underlying <code>cgul_exception</code> object. */
    cgul_exception_t obj;

};

#endif  /*  CGUL_EXCEPTION_CXX_H  */
