#ifndef CGUL_GET_LINE_H
#define CGUL_GET_LINE_H

/**
 * Get an arbitrarily large line of text from a file.
 *
 * @file
 * @author Paul Serice
 * @brief get an arbitrarily large line of text from a file
 * @see cgul_crlf__get_line()
 * @see cgul_crlf_file__get_line()
 */

#include "cgul_common.h"
#include "cgul_exception.h"
#include "cgul_string.h"

CGUL_BEGIN_C

/**
 * Read an arbitrarily long line of text from <code>f</code> and
 * append it to <code>s</code>.  A line of text is terminated by '\\n'
 * only.  Supporting just what the platform strictly considers to be
 * an <code>EOL</code> allows this method to be much faster than
 * <code>cgul_get_line_relaxed()</code>.  On success, 1 is returned.
 * On EOF, 0 is returned.  If an error occurs, 0 is returned and an
 * exception is thrown. <p>
 *
 * If you are reading directly from a block device, e.g., a
 * traditional file, you can use <code>cgul_crlf_file</code> which is
 * even faster than this function and, unlike this function, is
 * capable of reading text files in DOS, Mac, and Unix formats;
 * however, it does require that you can read your file in large
 * blocks. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] f input file
 * @param[in] s where to store the text
 * @return whether EOF was not reached
 */
extern CGUL_EXPORT int
cgul_get_line(cgul_exception_t* cex,
              FILE* f,
              cgul_string_t s);

/**
 * Read an arbitrarily long line of text from <code>f</code> and
 * append it to <code>s</code>.  It is more relaxed in what it allows
 * as the <code>EOL</code> sequence.  A line of text is terminated by
 * "\n", "\r", or "\r\n".  Support for the three main
 * <code>EOL</code> conventions allows you to read in text files
 * generated by most common operating systems.  On success, 1 is
 * returned.  On EOF, 0 is returned.  If an error occurs, 0 is
 * returned and an exception is thrown. <p>
 *
 * If you are reading directly from a block device, e.g., a regular
 * file, you can use <code>cgul_crlf_file</code> which is very fast
 * (even faster than <code>cgul_get_line()</code> above) and also
 * capable of reading text files in DOS, Mac, and Unix formats;
 * however, it does require that you can read your file in large
 * blocks. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] f input file
 * @param[in] s where to store the line of text
 * @return whether EOF was not reached
 */
extern CGUL_EXPORT int
cgul_get_line_relaxed(cgul_exception_t* cex,
                      FILE* f,
                      cgul_string_t s);

CGUL_END_C

#endif  /*  CGUL_GET_LINE_H  */
