#ifndef CGUL_MULTIMAP_NODE_H
#define CGUL_MULTIMAP_NODE_H

/**
 * This class implements a multimap node which holds one key/value
 * pair and enough information to find the next node.
 *
 * @file
 * @author Paul Serice
 * @brief multimap node
 */

#include "cgul_common.h"
#include "cgul_exception.h"
#include "cgul_multimap_node.h"

CGUL_BEGIN_C

/** Opaque pointer to a <code>cgul_multimap_node</code> instance. */
typedef struct cgul_multimap_node* cgul_multimap_node_t;

/*
 * The constructor and destructor are both private.  Only the
 * <code>cgul_multimap</code> class should create are destroy its
 * nodes.
 */

/**
 * Return the key stored in this node.  The key is used to determine
 * the sort order of the node.  If you want to set the value of the
 * key, there is no method to do so because changing the value of a
 * key requires changing the node's position in the
 * <code>cgul_multimap</code>.  Thus, if you need to change the value
 * of a key, you should delete the node from the
 * <code>cgul_multimap</code> and insert a new node with the new value
 * for the key.
 *
 * @param[in] cex c-style exception
 * @param[in] n <code>cgul_multimap_node</code> instance
 * @return key
 */
extern CGUL_EXPORT void*
cgul_multimap_node__get_key(cgul_exception_t* cex,
                            cgul_multimap_node_t n);

/**
 * Return the value stored in this node.
 *
 * @param[in] cex c-style exception
 * @param[in] n <code>cgul_multimap_node</code> instance
 * @return value
 */
extern CGUL_EXPORT void*
cgul_multimap_node__get_value(cgul_exception_t* cex,
                              cgul_multimap_node_t n);

/**
 * Set the value stored in this node.
 *
 * @param[in] cex c-style exception
 * @param[in] n <code>cgul_multimap_node</code> instance
 * @param[in] value
 */
extern CGUL_EXPORT void
cgul_multimap_node__set_value(cgul_exception_t* cex,
                              cgul_multimap_node_t n,
                              const void* value);

/**
 * Return the previous node.  You can use this method to iterate over
 * the multimap.  When there are no more nodes, <code>NULL</code> is
 * returned.  You can use <code>cgul_multimap__get_back()</code> to
 * get the last node in a <code>cgul_multimap</code> in order to start
 * iterating.
 *
 * The primary sort criterion is the user's <code>compare</code>
 * function, of course, but the <code>older_is_greater</code> boolean
 * is used as a secondary sort criterion.  When a key maps to many
 * values, the <code>older_is_greater</code> boolean determines how to
 * sort the older nodes relative to the newer nodes. <p>
 *
 * The way the <code>older_is_greater</code> boolean works in this
 * implementation is that each key maps to a linked list of values.
 * The <code>compare</code> function only determines the relative
 * order of the keys.  To order the values, this implementation always
 * pushes newer values onto the front of the list.  Thus, the newest
 * value is always at the front of the list, and the oldest value is
 * always at the back of the list.  If <code>older_is_greater</code>
 * is true, the sort order is from front to back.  If
 * <code>older_is_greater</code> is false, the sort order is from back
 * to front. <p>
 *
 * For example, if you want to iterate over the values for a
 * particular key in the same order they were inserted, you would call
 * this method with <code>older_is_greater</code> set to
 * <code>0</code>.  This way, the older values are considered to be
 * less than the newer values. <p>
 *
 * @param[in] cex c-style exception
 * @param[in] n <code>cgul_multimap_node</code> instance
 * @param[in] older_is_greater whether older nodes are greater than newer nodes
 * @return previous node
 */
extern CGUL_EXPORT cgul_multimap_node_t
cgul_multimap_node__get_prev(cgul_exception_t* cex,
                             cgul_multimap_node_t n,
                             int older_is_greater);

/**
 * Return the next node.  You can use this method to iterate over the
 * multimap.  When there are no more nodes, <code>NULL</code> is
 * returned.  You can use <code>cgul_multimap__get_front()</code> to
 * get the first node in a <code>cgul_multimap</code> in order to
 * start iterating.
 *
 * The primary sort criterion is the user's <code>compare</code>
 * function, of course, but the <code>older_is_greater</code> boolean
 * is used as a secondary sort criterion.  When a key maps to many
 * values, the <code>older_is_greater</code> boolean determines how to
 * sort the older nodes relative to the newer nodes. <p>
 *
 * The way the <code>older_is_greater</code> boolean works in this
 * implementation is that each key maps to a linked list of values.
 * The <code>compare</code> function only determines the relative
 * order of the keys.  To order the values, this implementation always
 * pushes newer values onto the front of the list.  Thus, the newest
 * value is always at the front of the list, and the oldest value is
 * always at the back of the list.  If <code>older_is_greater</code>
 * is true, the sort order is from front to back.  If
 * <code>older_is_greater</code> is false, the sort order is from back
 * to front. <p>
 *
 * For example, if you want to iterate over the values for a
 * particular key in the same order they were inserted, you would call
 * this method with <code>older_is_greater</code> set to
 * <code>0</code>.  This way, the older values are considered to be
 * less than the newer values. <p>
 *
 * @param[in] cex c-style exception
 * @param[in] n <code>cgul_multimap_node</code> instance
 * @param[in] older_is_greater whether older nodes are greater than newer nodes
 * @return next node
 */
extern CGUL_EXPORT cgul_multimap_node_t
cgul_multimap_node__get_next(cgul_exception_t* cex,
                             cgul_multimap_node_t n,
                             int older_is_greater);

/**
 * Return the next older node.  Together with
 * <code>cgul_multimap_node__get_younger()</code>, this method lets
 * you traverse the multimap in chronological order. If there is no
 * older node, this method returns <code>NULL</code>. <p>
 *
 * The following example shows how to iterate over the entire multimap
 * in reverse chronological order.  Notice that you have to start with
 * the youngest node when calling
 * <code>cgul_multimap_node__get_older()</code>:
 *
 * <pre>
 *     cgul_multimap_node_t n = cgul_multimap__get_youngest(cex, mm);
 *     for ( ; n ; n = cgul_multimap_node__get_older(cex, n)) {
 *         ...
 *     }
 * </pre>
 *
 * @param[in] cex c-style exception
 * @param[in] n <code>cgul_multimap_node</code> instance
 * @return older node
 */
extern CGUL_EXPORT cgul_multimap_node_t
cgul_multimap_node__get_older(cgul_exception_t* cex,
                              cgul_multimap_node_t n);

/**
 * Return the next younger node.  Together with
 * <code>cgul_multimap_node__get_older()</code>, this method lets you
 * traverse the multimap in chronological order. If there is no
 * younger node, this method returns <code>NULL</code>. <p>
 *
 * The following example shows how to iterate over the entire multimap in
 * chronological order.  Notice that you have to start with the oldest
 * node when calling <code>cgul_multimap_node__get_younger()</code>:
 *
 * <pre>
 *     cgul_multimap_node_t n = cgul_multimap__get_oldest(cex, mm);
 *     for ( ; n ; n = cgul_multimap_node__get_younger(cex, n)) {
 *         ...
 *     }
 * </pre>
 *
 * @param[in] cex c-style exception
 * @param[in] n <code>cgul_multimap_node</code> instance
 * @return younger node
 */
extern CGUL_EXPORT cgul_multimap_node_t
cgul_multimap_node__get_younger(cgul_exception_t* cex,
                                cgul_multimap_node_t n);

CGUL_END_C

#endif  /*  CGUL_MULTIMAP_NODE_H  */
