#ifndef CGUL_RADIX_SORT_H
#define CGUL_RADIX_SORT_H

/**
 * Radix sort an array.
 *
 * @file
 * @author Paul Serice
 * @brief radix sort
 */

#include "cgul_common.h"
#include "cgul_exception.h"
#include "cgul_int.h"

CGUL_BEGIN_C

/**
 * This typedef is the interface the client must define in order for
 * <code>cgul_radix_sort()</code> or
 * <code>cgul_radix_sort_pointers()</code> to sort an array.  This
 * function will be called repeatedly during the sorting process, and
 * it should consistently return the same <code>unsigned long</code>
 * value that is associated with the element <code>element</code> from
 * the array being sorted.  The elements from the array will then be
 * sorted in ascending order of the enumerations. <p>
 *
 * If a signed enumeration is more natural for the array being sorted,
 * just have your enumeration function cast the signed enumeration to
 * an <code>unsigned long</code> and use the
 * <code>are_enumerations_signed</code> parameter that is provided for
 * all functions that accept an enumeration function.  Alternatively,
 * your enumeration function could just shift the signed values into
 * the unsigned range and return only unsigned values.  If the shifted
 * sort order is the same as the original sort order, the end result
 * will be the same. <p>
 *
 * @param[in] element element from the array being sorted
 * @return enumeration for <code>element</code>
 */
typedef unsigned long int
(*cgul_radix_sort__enumerate_t)(const void* element);

/**
 * Perform an in-place radix sort on the array <code>v</code> which
 * holds <code>v_count</code> instances of values having size
 * <code>element_size</code> (in bytes).  To sort the array of values,
 * the client must supply the enumeration function <code>ef</code>
 * that maps each value to an integer.  This function sorts the values
 * in ascending order of their enumerations if
 * <code>ascending_order</code> is true or in descending order
 * otherwise.  If <code>are_enumerations_signed</code> is true, the
 * enumeration returned by <code>ef</code> will be treated is a signed
 * type and sorted accordingly; otherwise, it will be treated as an
 * unsigned type.  Because of all the callbacks, this function is
 * about 5x slower than the functions (below) that are optimized for
 * sorting arrays of integers.  (This function uses
 * <code>cgul_radix_sort_pointers()</code>, but it can be faster than
 * <code>cgul_radix_sort_pointers()</code> because calling this
 * function means all the values are located in contigous memory which
 * improves locality of reference.)  This functions internally
 * allocates (and releases) an extra array large enough to hold a copy
 * of <code>v</code>.  If an error occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] v array of elements to sort
 * @param[in] v_count number of elements in <code>v</code>
 * @param[in] element_size size of each element in <code>v</code> (in bytes)
 * @param[in] ascending_order whether to sort in ascending order
 * @param[in] ef enumeration function
 * @param[in] are_enumerations_signed whether the enumerations are signed
 */
extern CGUL_EXPORT void
cgul_radix_sort(cgul_exception_t* cex,
                void* v,
                size_t v_count,
                size_t element_size,
                int ascending_order,
                cgul_radix_sort__enumerate_t ef,
                int are_enumerations_signed);

/**
 * Perform an in-place radix sort on the array <code>v</code> which
 * holds <code>v_count</code> instances of pointers to values.  To
 * sort the array of pointers, the client must supply the enumeration
 * function <code>ef</code> that maps each pointer to an integer.
 * This function sorts the pointers in ascending order of their
 * enumerations if <code>ascending_order</code> is true or in
 * descending order otherwise.  If
 * <code>are_enumerations_signed</code> is true, the enumeration
 * returned by <code>ef</code> will be treated is a signed type and
 * sorted accordingly; otherwise, it will be treated as an unsigned
 * type.  Because of all the callbacks, this function is about 10x
 * slower than the functions (below) that are optimized for sorting
 * arrays of integers.  This functions internally allocates (and
 * releases) an extra array large enough to hold a copy of
 * <code>v</code>.  If an error occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] v array of elements to sort
 * @param[in] v_count number of elements in <code>v</code>
 * @param[in] ascending_order whether to sort in ascending order
 * @param[in] ef enumeration function
 * @param[in] are_enumerations_signed whether the enumerations are signed
 */
extern CGUL_EXPORT void
cgul_radix_sort_pointers(cgul_exception_t* cex,
                         void** v,
                         size_t v_count,
                         int ascending_order,
                         cgul_radix_sort__enumerate_t ef,
                         int are_enumerations_signed);

/**
 * Perform an in-place radix sort on the array <code>v</code> which
 * holds <code>v_count</code> instances of <code>signed char</code>
 * values.  On return, <code>v</code> will be in ascending order if
 * <code>ascending_order</code> is true or in descending order
 * otherwise.  This functions internally allocates (and releases) an
 * extra array large enough to hold a copy of <code>v</code>.  This
 * function assumes the computer represents negative values using
 * two's complement (which should be true for the vast majority of
 * computers).  If an error occurs, an exception is thrown.
 *
 * If another sort order is desired, use <code>cgul_radix_sort()</code>
 * or <code>cgul_radix_sort_pointers()</code>, but they are much slower
 * than this function. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] v array of elements to sort
 * @param[in] v_count number of elements in <code>v</code>
 * @param[in] ascending_order whether to sort in ascending order
 */
extern CGUL_EXPORT void
cgul_radix_sort_schar(cgul_exception_t* cex,
                      signed char* v,
                      size_t v_count,
                      int ascending_order);

/**
 * Perform an in-place radix sort on the array <code>v</code> which
 * holds <code>v_count</code> instances of <code>unsigned char</code>
 * values.  On return, <code>v</code> will be in ascending order if
 * <code>ascending_order</code> is true or in descending order
 * otherwise.  This functions internally allocates (and releases) an
 * extra array large enough to hold a copy of <code>v</code>.  If an
 * error occurs, an exception is thrown. <p>
 *
 * If another sort order is desired, use <code>cgul_radix_sort()</code>
 * or <code>cgul_radix_sort_pointers()</code>, but they are much slower
 * than this function. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] v array of elements to sort
 * @param[in] v_count number of elements in <code>v</code>
 * @param[in] ascending_order whether to sort in ascending order
 */
extern CGUL_EXPORT void
cgul_radix_sort_uchar(cgul_exception_t* cex,
                      unsigned char* v,
                      size_t v_count,
                      int ascending_order);

/**
 * Perform an in-place radix sort on the array <code>v</code> which
 * holds <code>v_count</code> instances of <code>short</code> values.
 * On return, <code>v</code> will be in ascending order if
 * <code>ascending_order</code> is true or in descending order
 * otherwise.  This functions internally allocates (and releases) an
 * extra array large enough to hold a copy of <code>v</code>.  This
 * function assumes the computer represents negative values using
 * two's complement (which should be true for the vast majority of
 * computers).  If an error occurs, an exception is thrown.
 *
 * If another sort order is desired, use <code>cgul_radix_sort()</code>
 * or <code>cgul_radix_sort_pointers()</code>, but they are much slower
 * than this function. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] v array of elements to sort
 * @param[in] v_count number of elements in <code>v</code>
 * @param[in] ascending_order whether to sort in ascending order
 */
extern CGUL_EXPORT void
cgul_radix_sort_short(cgul_exception_t* cex,
                      short int* v,
                      size_t v_count,
                      int ascending_order);

/**
 * Perform an in-place radix sort on the array <code>v</code> which
 * holds <code>v_count</code> instances of <code>unsigned short</code>
 * values.  On return, <code>v</code> will be in ascending order if
 * <code>ascending_order</code> is true or in descending order
 * otherwise.  This functions internally allocates (and releases) an
 * extra array large enough to hold a copy of <code>v</code>.  If an
 * error occurs, an exception is thrown. <p>
 *
 * If another sort order is desired, use <code>cgul_radix_sort()</code>
 * or <code>cgul_radix_sort_pointers()</code>, but they are much slower
 * than this function. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] v array of elements to sort
 * @param[in] v_count number of elements in <code>v</code>
 * @param[in] ascending_order whether to sort in ascending order
 */
extern CGUL_EXPORT void
cgul_radix_sort_ushort(cgul_exception_t* cex,
                       unsigned short int* v,
                       size_t v_count,
                       int ascending_order);

/**
 * Perform an in-place radix sort on the array <code>v</code> which
 * holds <code>v_count</code> instances of <code>int</code> values.
 * On return, <code>v</code> will be in ascending order if
 * <code>ascending_order</code> is true or in descending order
 * otherwise.  This functions internally allocates (and releases) an
 * extra array large enough to hold a copy of <code>v</code>.  This
 * function assumes the computer represents negative values using
 * two's complement (which should be true for the vast majority of
 * computers).  If an error occurs, an exception is thrown.
 *
 * If another sort order is desired, use <code>cgul_radix_sort()</code>
 * or <code>cgul_radix_sort_pointers()</code>, but they are much slower
 * than this function. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] v array of elements to sort
 * @param[in] v_count number of elements in <code>v</code>
 * @param[in] ascending_order whether to sort in ascending order
 */
extern CGUL_EXPORT void
cgul_radix_sort_int(cgul_exception_t* cex,
                    int* v,
                    size_t v_count,
                    int ascending_order);

/**
 * Perform an in-place radix sort on the array <code>v</code> which
 * holds <code>v_count</code> instances of <code>unsigned int</code>
 * values.  On return, <code>v</code> will be in ascending order if
 * <code>ascending_order</code> is true or in descending order
 * otherwise.  This functions internally allocates (and releases) an
 * extra array large enough to hold a copy of <code>v</code>.  If an
 * error occurs, an exception is thrown. <p>
 *
 * If another sort order is desired, use <code>cgul_radix_sort()</code>
 * or <code>cgul_radix_sort_pointers()</code>, but they are much slower
 * than this function. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] v array of elements to sort
 * @param[in] v_count number of elements in <code>v</code>
 * @param[in] ascending_order whether to sort in ascending order
 */
extern CGUL_EXPORT void
cgul_radix_sort_uint(cgul_exception_t* cex,
                     unsigned int* v,
                     size_t v_count,
                     int ascending_order);

/**
 * Perform an in-place radix sort on the array <code>v</code> which
 * holds <code>v_count</code> instances of <code>long</code> values.
 * On return, <code>v</code> will be in ascending order if
 * <code>ascending_order</code> is true or in descending order
 * otherwise.  This functions internally allocates (and releases) an
 * extra array large enough to hold a copy of <code>v</code>.  This
 * function assumes the computer represents negative values using
 * two's complement (which should be true for the vast majority of
 * computers).  If an error occurs, an exception is thrown.
 *
 * If another sort order is desired, use <code>cgul_radix_sort()</code>
 * or <code>cgul_radix_sort_pointers()</code>, but they are much slower
 * than this function. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] v array of elements to sort
 * @param[in] v_count number of elements in <code>v</code>
 * @param[in] ascending_order whether to sort in ascending order
 */
extern CGUL_EXPORT void
cgul_radix_sort_long(cgul_exception_t* cex,
                     long int* v,
                     size_t v_count,
                     int ascending_order);

/**
 * Perform an in-place radix sort on the array <code>v</code> which
 * holds <code>v_count</code> instances of <code>unsigned long</code>
 * values.  On return, <code>v</code> will be in ascending order if
 * <code>ascending_order</code> is true or in descending order
 * otherwise.  This functions internally allocates (and releases) an
 * extra array large enough to hold a copy of <code>v</code>.  If an
 * error occurs, an exception is thrown. <p>
 *
 * If another sort order is desired, use <code>cgul_radix_sort()</code>
 * or <code>cgul_radix_sort_pointers()</code>, but they are much slower
 * than this function. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] v array of elements to sort
 * @param[in] v_count number of elements in <code>v</code>
 * @param[in] ascending_order whether to sort in ascending order
 */
extern CGUL_EXPORT void
cgul_radix_sort_ulong(cgul_exception_t* cex,
                      unsigned long int* v,
                      size_t v_count,
                      int ascending_order);

/**
 * Perform an in-place radix sort on the array <code>v</code> which
 * holds <code>v_count</code> instances of <code>cgul_int8_t</code>
 * values.  On return, <code>v</code> will be in ascending order if
 * <code>ascending_order</code> is true or in descending order
 * otherwise.  This functions internally allocates (and releases) an
 * extra array large enough to hold a copy of <code>v</code>.  This
 * function assumes the computer represents negative values using
 * two's complement (which should be true for the vast majority of
 * computers).  If an error occurs, an exception is thrown.
 *
 * If another sort order is desired, use <code>cgul_radix_sort()</code>
 * or <code>cgul_radix_sort_pointers()</code>, but they are much slower
 * than this function. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] v array of elements to sort
 * @param[in] v_count number of elements in <code>v</code>
 * @param[in] ascending_order whether to sort in ascending order
 */
extern CGUL_EXPORT void
cgul_radix_sort_int8(cgul_exception_t* cex,
                     cgul_int8_t* v,
                     size_t v_count,
                     int ascending_order);

/**
 * Perform an in-place radix sort on the array <code>v</code> which
 * holds <code>v_count</code> instances of <code>cgul_uint8_t</code>
 * values.  On return, <code>v</code> will be in ascending order if
 * <code>ascending_order</code> is true or in descending order
 * otherwise.  This functions internally allocates (and releases) an
 * extra array large enough to hold a copy of <code>v</code>.  If an
 * error occurs, an exception is thrown. <p>
 *
 * If another sort order is desired, use <code>cgul_radix_sort()</code>
 * or <code>cgul_radix_sort_pointers()</code>, but they are much slower
 * than this function. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] v array of elements to sort
 * @param[in] v_count number of elements in <code>v</code>
 * @param[in] ascending_order whether to sort in ascending order
 */
extern CGUL_EXPORT void
cgul_radix_sort_uint8(cgul_exception_t* cex,
                      cgul_uint8_t* v,
                      size_t v_count,
                      int ascending_order);

/**
 * Perform an in-place radix sort on the array <code>v</code> which
 * holds <code>v_count</code> instances of <code>cgul_int16_t</code>
 * values.  On return, <code>v</code> will be in ascending order if
 * <code>ascending_order</code> is true or in descending order
 * otherwise.  This functions internally allocates (and releases) an
 * extra array large enough to hold a copy of <code>v</code>.  This
 * function assumes the computer represents negative values using
 * two's complement (which should be true for the vast majority of
 * computers).  If an error occurs, an exception is thrown.
 *
 * If another sort order is desired, use <code>cgul_radix_sort()</code>
 * or <code>cgul_radix_sort_pointers()</code>, but they are much slower
 * than this function. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] v array of elements to sort
 * @param[in] v_count number of elements in <code>v</code>
 * @param[in] ascending_order whether to sort in ascending order
 */
extern CGUL_EXPORT void
cgul_radix_sort_int16(cgul_exception_t* cex,
                      cgul_int16_t* v,
                      size_t v_count,
                      int ascending_order);

/**
 * Perform an in-place radix sort on the array <code>v</code> which
 * holds <code>v_count</code> instances of <code>cgul_uint16_t</code>
 * values.  On return, <code>v</code> will be in ascending order if
 * <code>ascending_order</code> is true or in descending order
 * otherwise.  This functions internally allocates (and releases) an
 * extra array large enough to hold a copy of <code>v</code>.  If an
 * error occurs, an exception is thrown. <p>
 *
 * If another sort order is desired, use <code>cgul_radix_sort()</code>
 * or <code>cgul_radix_sort_pointers()</code>, but they are much slower
 * than this function. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] v array of elements to sort
 * @param[in] v_count number of elements in <code>v</code>
 * @param[in] ascending_order whether to sort in ascending order
 */
extern CGUL_EXPORT void
cgul_radix_sort_uint16(cgul_exception_t* cex,
                       cgul_uint16_t* v,
                       size_t v_count,
                       int ascending_order);

/**
 * Perform an in-place radix sort on the array <code>v</code> which
 * holds <code>v_count</code> instances of <code>cgul_int32_t</code>
 * values.  On return, <code>v</code> will be in ascending order if
 * <code>ascending_order</code> is true or in descending order
 * otherwise.  This functions internally allocates (and releases) an
 * extra array large enough to hold a copy of <code>v</code>.  This
 * function assumes the computer represents negative values using
 * two's complement (which should be true for the vast majority of
 * computers).  If an error occurs, an exception is thrown.
 *
 * If another sort order is desired, use <code>cgul_radix_sort()</code>
 * or <code>cgul_radix_sort_pointers()</code>, but they are much slower
 * than this function. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] v array of elements to sort
 * @param[in] v_count number of elements in <code>v</code>
 * @param[in] ascending_order whether to sort in ascending order
 */
extern CGUL_EXPORT void
cgul_radix_sort_int32(cgul_exception_t* cex,
                      cgul_int32_t* v,
                      size_t v_count,
                      int ascending_order);

/**
 * Perform an in-place radix sort on the array <code>v</code> which
 * holds <code>v_count</code> instances of <code>cgul_uint32_t</code>
 * values.  On return, <code>v</code> will be in ascending order if
 * <code>ascending_order</code> is true or in descending order
 * otherwise.  This functions internally allocates (and releases) an
 * extra array large enough to hold a copy of <code>v</code>.  If an
 * error occurs, an exception is thrown. <p>
 *
 * If another sort order is desired, use <code>cgul_radix_sort()</code>
 * or <code>cgul_radix_sort_pointers()</code>, but they are much slower
 * than this function. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] v array of elements to sort
 * @param[in] v_count number of elements in <code>v</code>
 * @param[in] ascending_order whether to sort in ascending order
 */
extern CGUL_EXPORT void
cgul_radix_sort_uint32(cgul_exception_t* cex,
                       cgul_uint32_t* v,
                       size_t v_count,
                       int ascending_order);

#ifdef CGUL_INT64
/**
 * Perform an in-place radix sort on the array <code>v</code> which
 * holds <code>v_count</code> instances of <code>cgul_int64_t</code>
 * values.  On return, <code>v</code> will be in ascending order if
 * <code>ascending_order</code> is true or in descending order
 * otherwise.  This functions internally allocates (and releases) an
 * extra array large enough to hold a copy of <code>v</code>.  This
 * function assumes the computer represents negative values using
 * two's complement (which should be true for the vast majority of
 * computers).  If an error occurs, an exception is thrown.
 *
 * If another sort order is desired, use <code>cgul_radix_sort()</code>
 * or <code>cgul_radix_sort_pointers()</code>, but they are much slower
 * than this function. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] v array of elements to sort
 * @param[in] v_count number of elements in <code>v</code>
 * @param[in] ascending_order whether to sort in ascending order
 */
extern CGUL_EXPORT void
cgul_radix_sort_int64(cgul_exception_t* cex,
                      cgul_int64_t* v,
                      size_t v_count,
                      int ascending_order);
#endif /* CGUL_INT64 */

#ifdef CGUL_UINT64
/**
 * Perform an in-place radix sort on the array <code>v</code> which
 * holds <code>v_count</code> instances of <code>cgul_uint64_t</code>
 * values.  On return, <code>v</code> will be in ascending order if
 * <code>ascending_order</code> is true or in descending order
 * otherwise.  This functions internally allocates (and releases) an
 * extra array large enough to hold a copy of <code>v</code>.  If an
 * error occurs, an exception is thrown. <p>
 *
 * If another sort order is desired, use <code>cgul_radix_sort()</code>
 * or <code>cgul_radix_sort_pointers()</code>, but they are much slower
 * than this function. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] v array of elements to sort
 * @param[in] v_count number of elements in <code>v</code>
 * @param[in] ascending_order whether to sort in ascending order
 */
extern CGUL_EXPORT void
cgul_radix_sort_uint64(cgul_exception_t* cex,
                       cgul_uint64_t* v,
                       size_t v_count,
                       int ascending_order);
#endif /* CGUL_UINT64 */

CGUL_END_C

#endif  /*  CGUL_RADIX_SORT_H  */
