#ifndef CGUL_RADIX_SORT_CXX_H
#define CGUL_RADIX_SORT_CXX_H

/**
 * C++ wrapper for <code>cgul_radix_sort</code>.
 *
 * @file
 * @author Paul Serice
 * @brief C++ wrapper for <code>cgul_radix_sort</code>
 */

#include "cgul_exception_cxx.h"
#include "cgul_radix_sort.h"

/**
 * This class provides the C++ bindings for
 * <code>cgul_radix_sort</code>.  The main purpose of this class is to
 * convert the C-style function calls and exception handling into
 * C++-style function calls and exception handling.
 *
 * @brief C++ bindings for <code>cgul_radix_sort</code>
 */
class cgul_radix_sort_cxx {

public:

    /**
     * This typedef is the interface the client must define in order for
     * <code>cgul_radix_sort()</code> or
     * <code>cgul_radix_sort_pointers()</code> to sort an array.  This
     * function will be called repeatedly during the sorting process, and
     * it should consistently return the same <code>unsigned long</code>
     * value that is associated with the element <code>element</code> from
     * the array being sorted.  The elements from the array will then be
     * sorted in ascending order of the enumerations. <p>
     *
     * If a signed enumeration is more natural for the array being sorted,
     * just have your enumeration function cast the signed enumeration to an <code>unsigned long</code>
     * and use the <code>are_enumerations_signed</code> parameter that is
     * provided for all functions that accept an enumeration function.
     * Alternatively, your enumeration function could just shift the
     * signed values into the unsigned range and return only unsigned
     * values.  If the shifted sort order is the same as the original sort
     * order, the end result will be the same. <p>
     *
     * @param[in] element element from the array being sorted
     * @return enumeration for <code>element</code>
     */
    typedef cgul_radix_sort__enumerate_t enumerate_t;

    /**
     * Perform an in-place radix sort on the array <code>v</code> which
     * holds <code>v_count</code> instances of values having size
     * <code>element_size</code> (in bytes).  To sort the array of values,
     * the client must supply the enumeration function <code>ef</code>
     * that maps each value to an integer.  This function sorts the values
     * in ascending order of their enumerations if
     * <code>ascending_order</code> is true or in descending order
     * otherwise.  If <code>are_enumerations_signed</code> is true, the
     * enumeration returned by <code>ef</code> will be treated is a signed
     * type and sorted accordingly; otherwise, it will be treated as an
     * unsigned type.  Because of all the callbacks, this function is
     * about 5x slower than the functions (below) that are optimized for
     * sorting arrays of integers.  (This function uses
     * <code>cgul_radix_sort_pointers()</code>, but it can be faster than
     * <code>cgul_radix_sort_pointers()</code> because calling this
     * function means all the values are located in contigous memory which
     * improves locality of reference.)  This functions internally
     * allocates (and releases) an extra array large enough to hold a copy
     * of <code>v</code>.  If an error occurs, an exception is thrown.
     *
     * @param[in] v array of elements to sort
     * @param[in] v_count number of elements in <code>v</code>
     * @param[in] element_size size of each element in <code>v</code> (in bytes)
     * @param[in] ascending_order whether to sort in ascending order
     * @param[in] ef enumeration function
     * @param[in] are_enumerations_signed whether the enumerations are signed
     */
    static void
    cgul_radix_sort(void* v,
                    size_t v_count,
                    size_t element_size,
                    int ascending_order,
                    enumerate_t ef,
                    int are_enumerations_signed)
    {
        cgul_exception_t local = NULL;
        ::cgul_radix_sort(&local,
                          v,
                          v_count,
                          element_size,
                          ascending_order,
                          ef,
                          are_enumerations_signed);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Perform an in-place radix sort on the array <code>v</code> which
     * holds <code>v_count</code> instances of pointers to values.  To
     * sort the array of pointers, the client must supply the enumeration
     * function <code>ef</code> that maps each pointer to an integer.
     * This function sorts the pointers in ascending order of their
     * enumerations if <code>ascending_order</code> is true or in
     * descending order otherwise.  If
     * <code>are_enumerations_signed</code> is true, the enumeration
     * returned by <code>ef</code> will be treated is a signed type and
     * sorted accordingly; otherwise, it will be treated as an unsigned
     * type.  Because of all the callbacks, this function is about 10x
     * slower than the functions (below) that are optimized for sorting
     * arrays of integers.  This functions internally allocates (and
     * releases) an extra array large enough to hold a copy of
     * <code>v</code>.  If an error occurs, an exception is thrown.
     *
     * @param[in] v array of elements to sort
     * @param[in] v_count number of elements in <code>v</code>
     * @param[in] ascending_order whether to sort in ascending order
     * @param[in] ef enumeration function
     * @param[in] are_enumerations_signed whether the enumerations are signed
     */
    static void
    cgul_radix_sort_pointers(void** v,
                             size_t v_count,
                             int ascending_order,
                             enumerate_t ef,
                             int are_enumerations_signed)
    {
        cgul_exception_t local = NULL;
        ::cgul_radix_sort_pointers(&local,
                                   v,
                                   v_count,
                                   ascending_order,
                                   ef,
                                   are_enumerations_signed);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Perform an in-place radix sort on the array <code>v</code> which
     * holds <code>v_count</code> instances of <code>signed char</code>
     * values.  On return, <code>v</code> will be in ascending order if
     * <code>ascending_order</code> is true or in descending order
     * otherwise.  This functions internally allocates (and releases) an
     * extra array large enough to hold a copy of <code>v</code>.  This
     * function assumes the computer represents negative values using
     * two's complement (which should be true for the vast majority of
     * computers).  If an error occurs, an exception is thrown.
     *
     * If another sort order is desired, use <code>cgul_radix_sort()</code>
     * or <code>cgul_radix_sort_pointers()</code>, but they are much slower
     * than this function. <p>
     *
     * @param[in] v array of elements to sort
     * @param[in] v_count number of elements in <code>v</code>
     * @param[in] ascending_order whether to sort in ascending order
     */
    static void
    cgul_radix_sort_schar(signed char* v,
                          size_t v_count,
                          int ascending_order)
    {
        cgul_exception_t local = NULL;
        ::cgul_radix_sort_schar(&local, v, v_count, ascending_order);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Perform an in-place radix sort on the array <code>v</code> which
     * holds <code>v_count</code> instances of <code>unsigned char</code>
     * values.  On return, <code>v</code> will be in ascending order if
     * <code>ascending_order</code> is true or in descending order
     * otherwise.  This functions internally allocates (and releases) an
     * extra array large enough to hold a copy of <code>v</code>.  If an
     * error occurs, an exception is thrown. <p>
     *
     * If another sort order is desired, use <code>cgul_radix_sort()</code>
     * or <code>cgul_radix_sort_pointers()</code>, but they are much slower
     * than this function. <p>
     *
     * @param[in] v array of elements to sort
     * @param[in] v_count number of elements in <code>v</code>
     * @param[in] ascending_order whether to sort in ascending order
     */
    static void
    cgul_radix_sort_uchar(unsigned char* v,
                          size_t v_count,
                          int ascending_order)
    {
        cgul_exception_t local = NULL;
        ::cgul_radix_sort_uchar(&local, v, v_count, ascending_order);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Perform an in-place radix sort on the array <code>v</code> which
     * holds <code>v_count</code> instances of <code>short</code> values.
     * On return, <code>v</code> will be in ascending order if
     * <code>ascending_order</code> is true or in descending order
     * otherwise.  This functions internally allocates (and releases) an
     * extra array large enough to hold a copy of <code>v</code>.  This
     * function assumes the computer represents negative values using
     * two's complement (which should be true for the vast majority of
     * computers).  If an error occurs, an exception is thrown.
     *
     * If another sort order is desired, use <code>cgul_radix_sort()</code>
     * or <code>cgul_radix_sort_pointers()</code>, but they are much slower
     * than this function. <p>
     *
     * @param[in] v array of elements to sort
     * @param[in] v_count number of elements in <code>v</code>
     * @param[in] ascending_order whether to sort in ascending order
     */
    static void
    cgul_radix_sort_short(short int* v,
                          size_t v_count,
                          int ascending_order)
    {
        cgul_exception_t local = NULL;
        ::cgul_radix_sort_short(&local, v, v_count, ascending_order);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Perform an in-place radix sort on the array <code>v</code> which
     * holds <code>v_count</code> instances of <code>unsigned short</code>
     * values.  On return, <code>v</code> will be in ascending order if
     * <code>ascending_order</code> is true or in descending order
     * otherwise.  This functions internally allocates (and releases) an
     * extra array large enough to hold a copy of <code>v</code>.  If an
     * error occurs, an exception is thrown. <p>
     *
     * If another sort order is desired, use <code>cgul_radix_sort()</code>
     * or <code>cgul_radix_sort_pointers()</code>, but they are much slower
     * than this function. <p>
     *
     * @param[in] v array of elements to sort
     * @param[in] v_count number of elements in <code>v</code>
     * @param[in] ascending_order whether to sort in ascending order
     */
    static void
    cgul_radix_sort_ushort(unsigned short int* v,
                           size_t v_count,
                           int ascending_order)
    {
        cgul_exception_t local = NULL;
        ::cgul_radix_sort_ushort(&local, v, v_count, ascending_order);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Perform an in-place radix sort on the array <code>v</code> which
     * holds <code>v_count</code> instances of <code>int</code> values.
     * On return, <code>v</code> will be in ascending order if
     * <code>ascending_order</code> is true or in descending order
     * otherwise.  This functions internally allocates (and releases) an
     * extra array large enough to hold a copy of <code>v</code>.  This
     * function assumes the computer represents negative values using
     * two's complement (which should be true for the vast majority of
     * computers).  If an error occurs, an exception is thrown.
     *
     * If another sort order is desired, use <code>cgul_radix_sort()</code>
     * or <code>cgul_radix_sort_pointers()</code>, but they are much slower
     * than this function. <p>
     *
     * @param[in] v array of elements to sort
     * @param[in] v_count number of elements in <code>v</code>
     * @param[in] ascending_order whether to sort in ascending order
     */
    static void
    cgul_radix_sort_int(int* v,
                        size_t v_count,
                        int ascending_order)
    {
        cgul_exception_t local = NULL;
        ::cgul_radix_sort_int(&local, v, v_count, ascending_order);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Perform an in-place radix sort on the array <code>v</code> which
     * holds <code>v_count</code> instances of <code>unsigned int</code>
     * values.  On return, <code>v</code> will be in ascending order if
     * <code>ascending_order</code> is true or in descending order
     * otherwise.  This functions internally allocates (and releases) an
     * extra array large enough to hold a copy of <code>v</code>.  If an
     * error occurs, an exception is thrown. <p>
     *
     * If another sort order is desired, use <code>cgul_radix_sort()</code>
     * or <code>cgul_radix_sort_pointers()</code>, but they are much slower
     * than this function. <p>
     *
     * @param[in] v array of elements to sort
     * @param[in] v_count number of elements in <code>v</code>
     * @param[in] ascending_order whether to sort in ascending order
     */
    static void
    cgul_radix_sort_uint(unsigned int* v,
                         size_t v_count,
                         int ascending_order)
    {
        cgul_exception_t local = NULL;
        ::cgul_radix_sort_uint(&local, v, v_count, ascending_order);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Perform an in-place radix sort on the array <code>v</code> which
     * holds <code>v_count</code> instances of <code>long</code> values.
     * On return, <code>v</code> will be in ascending order if
     * <code>ascending_order</code> is true or in descending order
     * otherwise.  This functions internally allocates (and releases) an
     * extra array large enough to hold a copy of <code>v</code>.  This
     * function assumes the computer represents negative values using
     * two's complement (which should be true for the vast majority of
     * computers).  If an error occurs, an exception is thrown.
     *
     * If another sort order is desired, use <code>cgul_radix_sort()</code>
     * or <code>cgul_radix_sort_pointers()</code>, but they are much slower
     * than this function. <p>
     *
     * @param[in] v array of elements to sort
     * @param[in] v_count number of elements in <code>v</code>
     * @param[in] ascending_order whether to sort in ascending order
     */
    static void
    cgul_radix_sort_long(long int* v,
                         size_t v_count,
                         int ascending_order)
    {
        cgul_exception_t local = NULL;
        ::cgul_radix_sort_long(&local, v, v_count, ascending_order);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Perform an in-place radix sort on the array <code>v</code> which
     * holds <code>v_count</code> instances of <code>unsigned long</code>
     * values.  On return, <code>v</code> will be in ascending order if
     * <code>ascending_order</code> is true or in descending order
     * otherwise.  This functions internally allocates (and releases) an
     * extra array large enough to hold a copy of <code>v</code>.  If an
     * error occurs, an exception is thrown. <p>
     *
     * If another sort order is desired, use <code>cgul_radix_sort()</code>
     * or <code>cgul_radix_sort_pointers()</code>, but they are much slower
     * than this function. <p>
     *
     * @param[in] v array of elements to sort
     * @param[in] v_count number of elements in <code>v</code>
     * @param[in] ascending_order whether to sort in ascending order
     */
    static void
    cgul_radix_sort_ulong(unsigned long int* v,
                          size_t v_count,
                          int ascending_order)
    {
        cgul_exception_t local = NULL;
        ::cgul_radix_sort_ulong(&local, v, v_count, ascending_order);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Perform an in-place radix sort on the array <code>v</code> which
     * holds <code>v_count</code> instances of <code>cgul_int8_t</code>
     * values.  On return, <code>v</code> will be in ascending order if
     * <code>ascending_order</code> is true or in descending order
     * otherwise.  This functions internally allocates (and releases) an
     * extra array large enough to hold a copy of <code>v</code>.  This
     * function assumes the computer represents negative values using
     * two's complement (which should be true for the vast majority of
     * computers).  If an error occurs, an exception is thrown.
     *
     * If another sort order is desired, use <code>cgul_radix_sort()</code>
     * or <code>cgul_radix_sort_pointers()</code>, but they are much slower
     * than this function. <p>
     *
     * @param[in] v array of elements to sort
     * @param[in] v_count number of elements in <code>v</code>
     * @param[in] ascending_order whether to sort in ascending order
     */
    static void
    cgul_radix_sort_int8(cgul_int8_t* v,
                         size_t v_count,
                         int ascending_order)
    {
        cgul_exception_t local = NULL;
        ::cgul_radix_sort_int8(&local, v, v_count, ascending_order);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Perform an in-place radix sort on the array <code>v</code> which
     * holds <code>v_count</code> instances of <code>cgul_uint8_t</code>
     * values.  On return, <code>v</code> will be in ascending order if
     * <code>ascending_order</code> is true or in descending order
     * otherwise.  This functions internally allocates (and releases) an
     * extra array large enough to hold a copy of <code>v</code>.  If an
     * error occurs, an exception is thrown. <p>
     *
     * If another sort order is desired, use <code>cgul_radix_sort()</code>
     * or <code>cgul_radix_sort_pointers()</code>, but they are much slower
     * than this function. <p>
     *
     * @param[in] v array of elements to sort
     * @param[in] v_count number of elements in <code>v</code>
     * @param[in] ascending_order whether to sort in ascending order
     */
    static void
    cgul_radix_sort_uint8(cgul_uint8_t* v,
                          size_t v_count,
                          int ascending_order)
    {
        cgul_exception_t local = NULL;
        ::cgul_radix_sort_uint8(&local, v, v_count, ascending_order);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Perform an in-place radix sort on the array <code>v</code> which
     * holds <code>v_count</code> instances of <code>cgul_int16_t</code>
     * values.  On return, <code>v</code> will be in ascending order if
     * <code>ascending_order</code> is true or in descending order
     * otherwise.  This functions internally allocates (and releases) an
     * extra array large enough to hold a copy of <code>v</code>.  This
     * function assumes the computer represents negative values using
     * two's complement (which should be true for the vast majority of
     * computers).  If an error occurs, an exception is thrown.
     *
     * If another sort order is desired, use <code>cgul_radix_sort()</code>
     * or <code>cgul_radix_sort_pointers()</code>, but they are much slower
     * than this function. <p>
     *
     * @param[in] v array of elements to sort
     * @param[in] v_count number of elements in <code>v</code>
     * @param[in] ascending_order whether to sort in ascending order
     */
    static void
    cgul_radix_sort_int16(cgul_int16_t* v,
                          size_t v_count,
                          int ascending_order)
    {
        cgul_exception_t local = NULL;
        ::cgul_radix_sort_int16(&local, v, v_count, ascending_order);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Perform an in-place radix sort on the array <code>v</code> which
     * holds <code>v_count</code> instances of <code>cgul_uint16_t</code>
     * values.  On return, <code>v</code> will be in ascending order if
     * <code>ascending_order</code> is true or in descending order
     * otherwise.  This functions internally allocates (and releases) an
     * extra array large enough to hold a copy of <code>v</code>.  If an
     * error occurs, an exception is thrown. <p>
     *
     * If another sort order is desired, use <code>cgul_radix_sort()</code>
     * or <code>cgul_radix_sort_pointers()</code>, but they are much slower
     * than this function. <p>
     *
     * @param[in] v array of elements to sort
     * @param[in] v_count number of elements in <code>v</code>
     * @param[in] ascending_order whether to sort in ascending order
     */
    static void
    cgul_radix_sort_uint16(cgul_uint16_t* v,
                           size_t v_count,
                           int ascending_order)
    {
        cgul_exception_t local = NULL;
        ::cgul_radix_sort_uint16(&local, v, v_count, ascending_order);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Perform an in-place radix sort on the array <code>v</code> which
     * holds <code>v_count</code> instances of <code>cgul_int32_t</code>
     * values.  On return, <code>v</code> will be in ascending order if
     * <code>ascending_order</code> is true or in descending order
     * otherwise.  This functions internally allocates (and releases) an
     * extra array large enough to hold a copy of <code>v</code>.  This
     * function assumes the computer represents negative values using
     * two's complement (which should be true for the vast majority of
     * computers).  If an error occurs, an exception is thrown.
     *
     * If another sort order is desired, use <code>cgul_radix_sort()</code>
     * or <code>cgul_radix_sort_pointers()</code>, but they are much slower
     * than this function. <p>
     *
     * @param[in] v array of elements to sort
     * @param[in] v_count number of elements in <code>v</code>
     * @param[in] ascending_order whether to sort in ascending order
     */
    static void
    cgul_radix_sort_int32(cgul_int32_t* v,
                          size_t v_count,
                          int ascending_order)
    {
        cgul_exception_t local = NULL;
        ::cgul_radix_sort_int32(&local, v, v_count, ascending_order);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Perform an in-place radix sort on the array <code>v</code> which
     * holds <code>v_count</code> instances of <code>cgul_uint32_t</code>
     * values.  On return, <code>v</code> will be in ascending order if
     * <code>ascending_order</code> is true or in descending order
     * otherwise.  This functions internally allocates (and releases) an
     * extra array large enough to hold a copy of <code>v</code>.  If an
     * error occurs, an exception is thrown. <p>
     *
     * If another sort order is desired, use <code>cgul_radix_sort()</code>
     * or <code>cgul_radix_sort_pointers()</code>, but they are much slower
     * than this function. <p>
     *
     * @param[in] v array of elements to sort
     * @param[in] v_count number of elements in <code>v</code>
     * @param[in] ascending_order whether to sort in ascending order
     */
    static void
    cgul_radix_sort_uint32(cgul_uint32_t* v,
                           size_t v_count,
                           int ascending_order)
    {
        cgul_exception_t local = NULL;
        ::cgul_radix_sort_uint32(&local, v, v_count, ascending_order);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

#ifdef CGUL_INT64
    /**
     * Perform an in-place radix sort on the array <code>v</code> which
     * holds <code>v_count</code> instances of <code>cgul_int64_t</code>
     * values.  On return, <code>v</code> will be in ascending order if
     * <code>ascending_order</code> is true or in descending order
     * otherwise.  This functions internally allocates (and releases) an
     * extra array large enough to hold a copy of <code>v</code>.  This
     * function assumes the computer represents negative values using
     * two's complement (which should be true for the vast majority of
     * computers).  If an error occurs, an exception is thrown.
     *
     * If another sort order is desired, use <code>cgul_radix_sort()</code>
     * or <code>cgul_radix_sort_pointers()</code>, but they are much slower
     * than this function. <p>
     *
     * @param[in] v array of elements to sort
     * @param[in] v_count number of elements in <code>v</code>
     * @param[in] ascending_order whether to sort in ascending order
     */
    static void
    cgul_radix_sort_int64(cgul_int64_t* v,
                          size_t v_count,
                          int ascending_order)
    {
        cgul_exception_t local = NULL;
        ::cgul_radix_sort_int64(&local, v, v_count, ascending_order);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }
#endif /* CGUL_INT64 */

#ifdef CGUL_UINT64
    /**
     * Perform an in-place radix sort on the array <code>v</code> which
     * holds <code>v_count</code> instances of <code>cgul_uint64_t</code>
     * values.  On return, <code>v</code> will be in ascending order if
     * <code>ascending_order</code> is true or in descending order
     * otherwise.  This functions internally allocates (and releases) an
     * extra array large enough to hold a copy of <code>v</code>.  If an
     * error occurs, an exception is thrown. <p>
     *
     * If another sort order is desired, use <code>cgul_radix_sort()</code>
     * or <code>cgul_radix_sort_pointers()</code>, but they are much slower
     * than this function. <p>
     *
     * @param[in] v array of elements to sort
     * @param[in] v_count number of elements in <code>v</code>
     * @param[in] ascending_order whether to sort in ascending order
     */
    static void
    cgul_radix_sort_uint64(cgul_uint64_t* v,
                           size_t v_count,
                           int ascending_order)
    {
        cgul_exception_t local = NULL;
        ::cgul_radix_sort_uint64(&local, v, v_count, ascending_order);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }
#endif /* CGUL_UINT64 */

};

#endif  /*  CGUL_RADIX_SORT_CXX_H  */
