#ifndef CGUL_STRTOK_CXX_H
#define CGUL_STRTOK_CXX_H

/**
 * C++ wrapper for <code>cgul_strtok</code>.
 *
 * @file
 * @author Paul Serice
 * @brief C++ wrapper for <code>cgul_strtok</code>
 */

#include "cgul_exception_cxx.h"
#include "cgul_strtok.h"

/**
 * This class provides the C++ bindings for C
 * <code>cgul_strtok</code> objects.  The main purpose of this class
 * is to convert the C-style function calls and exception handling in
 * <code>cgul_strtok</code> into C++-style function calls and
 * exception handling.
 *
 * @brief C++ bindings for <code>cgul_strtok</code>
 */
class cgul_strtok_cxx {

public:

    /**
     * Default constructor.
     */
    cgul_strtok_cxx() {
        cgul_exception_t local = NULL;
        obj = cgul_strtok__new(&local);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Destructor.
     */
    virtual
    ~cgul_strtok_cxx() {
        if (obj) {
            cgul_strtok__delete(obj);
            obj = NULL;
        }
    }

    /**
     * Set the value to tokenize.  This method makes an internal copy of
     * <code>value</code> out of which tokens will be returned.
     *
     * @param[in] value value to tokenize
     */
    virtual void
    set_value(const char* value)
    {
        cgul_exception_t local = NULL;
        cgul_strtok__set_value(&local, obj, value);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Get the next token that is delimited by the current list of
     * <code>separators</code>.  The list of separators can change each
     * time this method is called.  If all the tokens have been consumed,
     * <code>NULL</code> is returned.  <p>
     *
     * If <code>allow_empty_tokens</code> is true, this method will return
     * empty tokens when two consecutive separators are encountered
     * including the case where a separator immediately precedes the end
     * of string.  If <code>allow_empty_tokens</code> is false,
     * consecutive separators will be telescoped into one, and all tokens
     * returned will be non-empty. <p>
     *
     * @param[in] separators list of separators that delimit tokens
     * @param[in] allow_empty_tokens whether to allow returning empty tokens
     * @return next token
     */
    virtual const char*
    get_next(const char* separators,
             int allow_empty_tokens)
    {
        const char* rv = NULL;
        cgul_exception_t local = NULL;
        rv = cgul_strtok__get_next(&local, obj, separators, allow_empty_tokens);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Get the trailing separator for the current token.  This value can
     * be <code>nul</code> if the current token terminated because it
     * reached the end of the string.
     *
     * @return trailing separator for the current token
     */
    virtual char
    get_separator() const
    {
        char rv = '\0';
        cgul_exception_t local = NULL;
        rv = cgul_strtok__get_separator(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Get the starting index for the current token.
     *
     * @return starting index for the current token 
     */
    virtual size_t
    get_start() const
    {
        size_t rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_strtok__get_start(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Get the ending index for the current token.  The value returned is
     * actually one beyond the end of the current token.  So in effect,
     * the value returned is the index of the trailing separator for the
     * current token.
     *
     * @return ending index for the current token 
     */
    virtual size_t
    get_end() const
    {
        size_t rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_strtok__get_end(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Get the underlying <code>cgul_strtok</code> object.
     *
     * @return underlying object
     */
    virtual cgul_strtok_t
    get_obj() const
    {
        return obj;
    }

    /**
     * Take the underlying <code>cgul_strtok</code> object.  This
     * means the underlying object will not be deleted when the
     * wrapper goes out of scope.  Also, because you have taken the
     * underlying object, no other methods should be called on this
     * wrapper's instance.  Lastly, after taking the underlying
     * object, it is the caller's responsibility to delete the
     * underlying object by calling
     * <code>cgul_strtok__delete()</code>.
     *
     * @return underlying object
     */
    virtual cgul_strtok_t
    take_obj()
    {
        cgul_strtok_t rv = obj;
        obj = NULL;
        return rv;
    }

    /**
     * Set the new underlying object to <code>rhs</code>.  This causes
     * the old underlying object to be deleted which invalidates any
     * outstanding pointers to or iterators for the old underlying
     * object. <p>
     *
     * This instance takes ownership of <code>rhs</code> which means
     * <code>rhs</code> will be automatically deleted when the C++
     * wrapper is deleted.  To prevent automatic deletion of
     * <code>rhs</code>, call <code>take_obj()</code> when the C++
     * wrapper is no longer needed. <p>
     *
     * @param[in] rhs right-hand side
     */
    virtual void
    set_obj(cgul_strtok_t rhs)
    {
        if (rhs && (rhs != obj)) {
            if (obj) {
                cgul_strtok__delete(obj);
            }
            obj = rhs;
        }
    }

private:

    /* Hide these. */
    cgul_strtok_cxx(const cgul_strtok_cxx&);
    cgul_strtok_cxx& operator=(const cgul_strtok_cxx&);

    /** Underlying <code>cgul_strtok</code> object. */
    cgul_strtok_t obj;

};

#endif  /*  CGUL_STRTOK_CXX_H  */
