#ifndef CGUL_SUBSTRING
#define CGUL_SUBSTRING

/**
 * Substring search.  This class performs a substring search using the
 * Rabin-Karp algorithm which is effectively O(m + n) where m is the
 * length of the substring and n is the length of the string being
 * searched.
 *
 * @file
 * @author Paul Serice
 * @brief substring search
 */

#include "cgul_common.h"
#include "cgul_exception.h"

CGUL_BEGIN_C

/** Opaque pointer to a <code>cgul_substring</code> instance. */
typedef struct cgul_substring* cgul_substring_t;

/*************************************************************************
 * EZ API
 *************************************************************************/

/**
 * This function is the easy way to search because it does not require
 * the user to allocate a <code>cgul_substring</code> instance.
 * Instead, just pass in the binary substring <code>ss</code> and its
 * length <code>ss_length</code> in bytes along with the binary block
 * <code>block</code> to search and its length
 * <code>block_length</code> in bytes.  If the substring is found, a
 * pointer to the first matching character of the substring in
 * <code>block</code> is returned; otherwise, <code>NULL</code> is
 * returned. <p>
 *
 * If you are only going to search <code>block</code> once, in
 * addition to being easier to use, this function will be faster than
 * <code>cgul_substring__search()</code> because this function does
 * not copy the substring. <p>
 *
 * @param[in] cex c-style exception
 * @param[in] ss binary substring
 * @param[in] ss_length substring length in bytes
 * @param[in] block block of binary data to search
 * @param[in] block_length block length in bytes
 * @return pointer to first match or <code>NULL</code> if no match
 */
extern CGUL_EXPORT const char*
cgul_substring__search_ez(cgul_exception_t* cex,
                          const char* ss,
                          size_t ss_length,
                          const char* block,
                          size_t block_length);

/*************************************************************************
 * Multiple Search API
 *************************************************************************/

/**
 * Create a new <code>cgul_substring</code> instance.  The caller is
 * responsible for calling <code>cgul_substring__delete()</code> on
 * the pointer returned.  If an error occurs, <code>NULL</code> is
 * returned, and an exception is thrown. <p>
 *
 * @param[in,out] cex c-style exception
 * @return new <code>cgul_substring</code> instance
 */
extern CGUL_EXPORT cgul_substring_t
cgul_substring__new(cgul_exception_t* cex);

/**
 * Create a new <code>cgul_substring</code> instance initializing the
 * substring from the C-style string <code>substring</code>.  The
 * caller is responsible for calling
 * <code>cgul_substring__delete()</code> on the pointer returned.  If
 * an error occurs, <code>NULL</code> is returned, and an exception is
 * thrown. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] substring c-style substring
 * @return new <code>cgul_substring</code> instance
 */
extern CGUL_EXPORT cgul_substring_t
cgul_substring__new_from_cstring(cgul_exception_t* cex,
                                 const char* substring);

/**
 * Create a new <code>cgul_substring</code> instance initializing the
 * substring from the binary string <code>substring</code> extending
 * for <code>length</code> bytes.  The caller is responsible for
 * calling <code>cgul_substring__delete()</code> on the pointer
 * returned.  If an error occurs, <code>NULL</code> is returned, and
 * an exception is thrown. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] substring binary substring
 * @param[in] length length of <code>substring</code>
 * @return new <code>cgul_substring</code> instance
 */
extern CGUL_EXPORT cgul_substring_t
cgul_substring__new_from_bstring(cgul_exception_t* cex,
                                 const char* substring,
                                 size_t length);

/**
 * Delete the <code>cgul_substring</code> instance <code>ss</code>
 * freeing all internally allocated resources.  The caller must not
 * attempt to dereference <code>ss</code> after calling this function.
 *
 * @param[in] ss <code>cgul_substring</code> instance
 */
extern CGUL_EXPORT void
cgul_substring__delete(cgul_substring_t ss);

/**
 * Get the value used as the substring in subsequent searches.  The
 * value can be set by calling
 * <code>cgul_substring__set_value()</code>.  If the value is not set,
 * <code>NULL</code> is returned.
 *
 * @param[in] cex c-style exception
 * @param[in] ss <code>cgul_substring</code> instance
 * @return value used as the substring in subsequent searches
 */
extern CGUL_EXPORT const char*
cgul_substring__get_value(cgul_exception_t* cex,
                          cgul_substring_t ss);

/**
 * Set the value to use as the substring in subsequent searches.  This
 * function must be called before calling
 * <code>cgul_substring__search()</code>.  It caches the
 * pre-computations based on the substring so that subsequent searches
 * for the same substring will not have to recompute the same values.
 * If an error occurs, an exception is thrown. <p>
 *
 * In addition to caching the pre-computed values, this function also
 * caches a copy of <code>substring</code> for use with
 * <code>cgul_substring__search()</code>.  Because
 * <code>cgul_substring__search_ez()</code> does not copy the
 * substring, it is not always clear whether
 * <code>cgul_substring__search()</code> or
 * <code>cgul_substring__search_ez()</code> will run faster. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] ss <code>cgul_substring</code> instance
 * @param[in] substring binary substring
 * @param[in] length length of <code>substring</code>
 */
extern CGUL_EXPORT void
cgul_substring__set_value(cgul_exception_t* cex,
                          cgul_substring_t ss,
                          const char* substring,
                          size_t length);

/**
 * Get the length of the substring that will be used in subsequent
 * searches.  The substring can be set by calling
 * <code>cgul_substring__set_value()</code>.  If the substring is not
 * set, <code>0</code> is returned.
 *
 * @param[in] cex c-style exception
 * @param[in] ss <code>cgul_substring</code> instance
 * @return length of the substring that will be used in subsequent searches
 */
extern CGUL_EXPORT size_t
cgul_substring__get_length(cgul_exception_t* cex,
                           cgul_substring_t ss);

/**
 * Search the first <code>length</code> bytes of block
 * <code>block</code> for the substring set by the most recent call to
 * <code>cgul_substring__set_value()</code>.  If the substring is
 * found, a pointer to the first matching character in
 * <code>block</code> is returned; otherwise, <code>NULL</code> is
 * returned.
 *
 * @param[in] cex c-style exception
 * @param[in] ss <code>cgul_substring</code> instance
 * @param[in] block block of binary data to search
 * @param[in] length block length in bytes
 * @return pointer to first match or <code>NULL</code> if no match
 */
extern CGUL_EXPORT const char* 
cgul_substring__search(cgul_exception_t* cex,
                       cgul_substring_t ss,
                       const char* block,
                       size_t length);

CGUL_END_C

#endif  /*  CGUL_SUBSTRING  */
