#include "cgul_crc16_p.h"
#include "cgul_defensive.h"
#include "cgul_libc.h"
#include "cgul_rbtree.h"
#include "cgul_string.h"

/**
 * This class generates all the lookup tables it needs.  It then sets
 * up each object that can share tables to use the same generated
 * table.  The tables that are stored in this tree are referenced
 * counted so that each table can be freed when it is no longer in
 * use.
 */
static cgul_rbtree_t cgul_crc16__tables = NULL;

/**
 * This function compares two keys that index into
 * <code>cgul_crc16__tables</code>.
 *
 * @param[in] key1 first key
 * @param[in] key2 second key
 * @return comparison of <code>key1</code> with </code>key2</code>
 */
static int
cgul_crc16__tables_key_compare(const void* key1, const void* key2)
{
    cgul_crc16__tables_key_t k1 = (cgul_crc16__tables_key_t)key1;
    cgul_crc16__tables_key_t k2 = (cgul_crc16__tables_key_t)key2;
    /* Sort first by generator polynomial ... */
    int rv = k1->g - k2->g;
    if (rv == 0) {
        /* ... then by "reflected".  For no good reason, the object
         * that is not reflected is consider to be greater than the
         * object that is reflected.*/
        if (k1->reflected != k2->reflected) {
            if (!k1->reflected) {
                rv = 1;
            } else {
                rv = -1;
            }
        }
    }
    return rv;
}

/**
 * This function returns a table suitable for <code>crc</code> to use
 * for table lookups when calculating the CRC.  The caller is
 * responsible for calling <code>free()</code> on the pointer that is
 * returned.
 *
 * @param[in] cex c-style exception
 * @param[in] crc <code>cgul_crc16</code> instance
 * @return lookup table
 */
static unsigned short*
cgul_crc16__generate_table(cgul_exception_t* cex,
                           cgul_crc16_t crc)
{
    unsigned short i = 0;
    unsigned short j = 0;
    unsigned short n = 0;
    unsigned short* rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(crc);

    /* Allocate the lookup table. */
    rv = (unsigned short*)cgul_libc__calloc(cex, 1, sizeof(*rv) * 256);
    if (*cex) {
        goto out;
    }

    if (crc->reflected) {
        /* Generate the table for the reflected algorithm. */
        for (i = 0 ; i < 256 ; ++i) {
            n = i;
            for (j = 0 ; j < 8 ; ++j) {
                n = (n & 0x01) ? (n >> 1) ^ crc->g : (n >> 1);
            }
            rv[i] = n;
        }
    } else {
        /* Generate the regular table. */
        for (i = 0 ; i < 256 ; ++i) {
            n = i << 8;
            for (j = 0 ; j < 8 ; ++j) {
                n = (n & 0x8000) ? (n << 1) ^ crc->g : (n << 1);
            }
            rv[i] = n & 0xffffU;
        }
    }

 out:
    return rv;
}

/**
 * The lookup table is set when the class is first created.  If a
 * suitable table has already been created, this object will share
 * that table; otherwise, a suitable table will be generated.
 *
 * @param[in,out] cex c-style exception
 * @param[in] crc <code>cgul_crc16</code> instance
 */
static void
cgul_crc16__set_table(cgul_exception_t* cex,
                      cgul_crc16_t crc)
{
    int new_key_is_inserted = 0;
    cgul_crc16__tables_key_t new_key = NULL;
    cgul_crc16__tables_key_t old_key = NULL;
    cgul_rbtree_node_t old_node = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(crc);

    /* Create a new key that does not have a table. */
    new_key = (cgul_crc16__tables_key_t)
                  cgul_libc__calloc(cex, 1, sizeof(*new_key));
    if (*cex) {
        goto out;
    }

    new_key->g = crc->g;
    new_key->reflected = crc->reflected;

    /* Use the new node to try to find an identical node that already
     * has a table.  If you cannot find an existing node, insert the
     * new node into the table.  Remember that the
     * cgul_rbtree__insert() method combines the "find" and "insert"
     * operation for efficiency.  (See the cgul_rbtree.h header for
     * more.) */
    new_key_is_inserted = cgul_rbtree__insert(cex,
                                              cgul_crc16__tables,
                                              new_key,
                                              &old_node);
    if (*cex) {
        goto out;
    }

    /* If the new key was added, generate and save a new lookup table. */
    if (new_key_is_inserted) {
        new_key->table = cgul_crc16__generate_table(cex, crc);
        if (*cex) {
            goto out;
        }
        crc->table = new_key->table;
        ++new_key->ref_count;
        goto out;
    }

    /* The key was not inserted because it already exists in the tree.
     * Use the lookup table that already exists. */
    assert(old_node);
    old_key = (cgul_crc16__tables_key_t)
                  cgul_rbtree_node__get_key(cex, old_node);
    assert(old_key);
    assert(old_key->table);
    crc->table = old_key->table;
    old_key->ref_count =
        cgul_defensive__add_size_type(cex, old_key->ref_count, 1);
    if (*cex) {
        goto out;
    }

 out:

    /* Remove "new_key" if an exception was thrown after inserting it. */
    if (*cex && new_key_is_inserted) {
        cgul_exception_t local = NULL;
        cgul_rbtree__remove(&local, cgul_crc16__tables,
                            new_key, NULL, NULL);
    }

    /* Clean up "new_key" if an exception was thrown after creating it
     * or if a the key wasn't inserted at all. */
    if ((*cex && new_key) || (!new_key_is_inserted && new_key)) {
        free(new_key);
    }

    return;
}

/************************************************************************
 * MSVC++ does not like the "inline" key word.  So we resort to
 * macros.  Inline functions are much nicer, but macros seems to be
 * more portable.
 ************************************************************************/
#define CGUL_CRC16__RESET(crc) \
do { \
    (crc)->accumulator = (crc)->init; \
    (crc)->size = 0; \
} while(0)

#define CGUL_CRC16__UPDATE_BYTE_REFLECTED(acc, table, c) \
do { \
    (acc) = ((acc) >> 8) ^ (table)[((acc) ^ (c)) & 0xff]; \
} while(0)

#define CGUL_CRC16__UPDATE_BYTE_REGULAR(acc, table, c) \
do { \
    (acc) = ((acc) << 8) ^ (table)[(((acc) >> 8) ^ (c)) & 0xff]; \
} while(0)

#define CGUL_CRC16__UPDATE_BYTE(acc, reflected, table, c) \
do { \
    if (reflected) { \
        /* Update the CRC using the reflected algorithm. */ \
        CGUL_CRC16__UPDATE_BYTE_REFLECTED(acc, table, c); \
    } else { \
        /* Update the CRC using the regular algorithm. */ \
        CGUL_CRC16__UPDATE_BYTE_REGULAR(acc, table, c); \
    } \
} while(0)

#define CGUL_CRC16__UPDATE_BLOCK(crc, block, length) \
do { \
    /* Limit dereferencing "crc" to improve performance. */ \
    unsigned int accumulator = (crc)->accumulator; \
    unsigned short* table = (crc)->table; \
\
    /* Update the size before we modify "length". */ \
    (crc)->size = cgul_defensive__add_size_type(cex, (crc)->size, (length)); \
    if (*cex) { \
        goto out; \
    } \
\
    /* We can save some time if we only check the "reflected" boolean
     * once.  Also notice that we are passing *(block)++ to a macro.
     * It is generally a bad idea to pass x++ to a macro, but this
     * time it is safe because the parameter in question is only used
     * once. */ \
    if ((crc)->reflected) { \
        while ((length)--) { \
            CGUL_CRC16__UPDATE_BYTE_REFLECTED(accumulator, table, *(block)++);\
        } \
    } else { \
        while ((length)--) { \
            CGUL_CRC16__UPDATE_BYTE_REGULAR(accumulator, table, *(block)++); \
        } \
    } \
\
    /* You must remember to update "crc". */ \
    (crc)->accumulator = (unsigned short)accumulator; \
} while(0)

#define CGUL_CRC16__CRC(crc, rv) \
do { \
    /* Append size if necessary. */ \
    if ((crc)->append_size) { \
        /* Remember to do the size XOR. */ \
        size_t tmp = (crc)->size ^ (crc)->size_xor; \
        assert(sizeof(tmp) >= 2); \
        if (tmp == 0) { \
            /* Degenerate case. */ \
            CGUL_CRC16__UPDATE_BYTE((crc)->accumulator, \
                                    (crc)->reflected, \
                                    (crc)->table, \
                                    0); \
        } else { \
            unsigned int i = 0; \
            for (i = 0 ; tmp && (i < sizeof(tmp)) ; ++i) { \
                CGUL_CRC16__UPDATE_BYTE((crc)->accumulator, \
                                        (crc)->reflected, \
                                        (crc)->table, \
                                        (unsigned char)tmp); \
                tmp >>= 8; \
            } \
        } \
    } \
\
    /* Remember to do the final XOR.  Elsewhere, we are able to do the
     * CRC calculation without regard to the size of the accumulator,
     * but here, we must "and" with 0xffffU because we need 16 bits
     * and the accumulator may be wider than 16 bits. */ \
    (rv) = ((crc)->accumulator ^ (crc)->final_xor) & 0xffffU; \
\
    /* Reset this object so it can be reused. */ \
    CGUL_CRC16__RESET(crc); \
} while(0)

cgul_crc16_t
cgul_crc16__new(cgul_exception_t* cex,
                unsigned short g,
                unsigned short init,
                unsigned short final_xor,
                int reflected)
{
    cgul_crc16_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* First, make sure the static variables are initialized. */
    if (cgul_crc16__tables == NULL) {
        cgul_crc16__tables =
            cgul_rbtree__new(cex, &cgul_crc16__tables_key_compare);
        if (*cex) {
            goto out;
        }
    }

    /* Now you can initialize this object. */
    rv = (cgul_crc16_t)cgul_libc__calloc(cex, 1, sizeof(*rv));
    if(*cex) {
        goto out;
    }

    rv->g = g;
    rv->init = init;
    rv->final_xor = final_xor;
    rv->reflected = reflected;

    cgul_crc16__set_table(cex, rv);
    if (*cex) {
        goto out;
    }

    CGUL_CRC16__RESET(rv);
    
 out:

    /* Clean up partially constructed objects. */
    if (*cex && rv) {
        cgul_crc16__delete(rv);
        rv = NULL;
    }

    return rv;
}

cgul_crc16_t
cgul_crc16__new_appends_size(cgul_exception_t* cex,
                             unsigned short g,
                             unsigned short init,
                             unsigned short final_xor,
                             int reflected,
                             size_t size_xor)
{
    cgul_crc16_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    rv = cgul_crc16__new(cex, g, init, final_xor, reflected);
    if (*cex) {
        goto out;
    }

    rv->append_size = 1;
    rv->size_xor = size_xor;

 out:
    return rv;
}

void
cgul_crc16__delete(cgul_crc16_t crc)
{
    cgul_rbtree_node_t node = NULL;
    cgul_crc16__tables_key_t key;
    struct cgul_crc16__tables_key search_key;
    cgul_exception_t local = NULL;

    /*
     * Keep the cgul_crc16__tables tree up to date.
     */
    if (crc && cgul_crc16__tables) {

        /* Search for the node that holds the table this object is using. */
        search_key.g = crc->g;
        search_key.reflected = crc->reflected;
        node = cgul_rbtree__find(&local, cgul_crc16__tables, &search_key);

        /* Decrement the node's reference count. */
        if (node) {
            key = (cgul_crc16__tables_key_t)
                      cgul_rbtree_node__get_key(&local, node);
            assert(key);
            assert(key->ref_count);
            if (--key->ref_count == 0) {
                /* If the node's reference count drops to zero, remove
                 * the node from the tree, free the lookup table, and
                 * free the key. */
                cgul_rbtree__remove(&local, cgul_crc16__tables,
                                    key, NULL, NULL);
                free(key->table);
                free(key);
            }
        }

        /* If the rbtee is empty, delete the tree. */
        if (cgul_rbtree__is_empty(&local, cgul_crc16__tables)) {
            cgul_rbtree__delete(cgul_crc16__tables);
            cgul_crc16__tables = NULL;
        }

    }

    /*
     * Free this object.
     */
    if (crc) {
        free(crc);
    }
}

void
cgul_crc16__assign(cgul_exception_t* cex,
                   cgul_crc16_t lhs,
                   cgul_crc16_t rhs)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(lhs);
    assert(rhs);

    /* There is nothing to do if lhs and rhs point to the same object. */
    if (lhs == rhs) {
        goto out;
    }

    /* Make sure both objects use the same lookup table. */
    if (lhs->table != rhs->table) {
        *cex = cgul_exception__new(
                   "cgul_crc16__assign: lookup table not shared");
        goto out;
    }

    /* Copy the size and accumulator from rhs to lhs. */
    lhs->size = rhs->size;
    lhs->accumulator = rhs->accumulator;

 out:
    return;
}

unsigned short
cgul_crc16__crc(cgul_exception_t* cex,
                cgul_crc16_t crc)
{
    unsigned short rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(crc);

    CGUL_CRC16__CRC(crc, rv);

 out:
    return rv;
}

unsigned short
cgul_crc16__crc_block(cgul_exception_t* cex,
                      cgul_crc16_t crc,
                      const void* block,
                      size_t length)
{
    unsigned short rv = 0;
    unsigned char* buf = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(crc);
    assert(block);

    /* "buf" is used to protect "block" from the macro expansion. */
    buf = (unsigned char*)block;
    CGUL_CRC16__UPDATE_BLOCK(crc, buf, length);
    CGUL_CRC16__CRC(crc, rv);

 out:
    return rv;
}

unsigned short
cgul_crc16__crc_range_of_block(cgul_exception_t* cex,
                               cgul_crc16_t crc,
                               const void* block,
                               size_t offset,
                               size_t length)
{
    unsigned short rv = 0;
    unsigned char* buf = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(crc);
    assert(block);

    /* "buf" is used to protect "block" from the macro expansion. */
    buf = (unsigned char*)block + offset;
    CGUL_CRC16__UPDATE_BLOCK(crc, buf, length);
    CGUL_CRC16__CRC(crc, rv);

 out:
    return rv;
}

unsigned long
cgul_crc16__crc_file(cgul_exception_t* cex,
                     cgul_crc16_t crc,
                     FILE* f)
{
    unsigned long rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(crc);
    assert(f);

    /* Update the CRC from the file. */
    cgul_crc16__update_file(cex, crc, f);
    if (*cex) {
        goto out;
    }

    /* Return the final CRC value. */
    CGUL_CRC16__CRC(crc, rv);

 out:
    return rv;
}

unsigned long
cgul_crc16__crc_fname(cgul_exception_t* cex,
                      cgul_crc16_t crc,
                      const char* fname)
{
    unsigned long rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(crc);
    assert(fname);

    /* Update the CRC given the file fname. */
    cgul_crc16__update_fname(cex, crc, fname);
    if (*cex) {
        goto out;
    }

    /* Return the final CRC value. */
    CGUL_CRC16__CRC(crc, rv);

 out:
    return rv;
}

void
cgul_crc16__update_byte(cgul_exception_t* cex,
                        cgul_crc16_t crc,
                        unsigned char c)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(crc);

    CGUL_CRC16__UPDATE_BYTE(crc->accumulator, crc->reflected, crc->table, c);
    crc->size = cgul_defensive__add_size_type(cex, crc->size, 1);

 out:
    return;
}

void
cgul_crc16__update_block(cgul_exception_t* cex,
                         cgul_crc16_t crc,
                         const void* block,
                         size_t length)
{
    unsigned char* buf = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(crc);
    assert(block);

    /* "buf" is used to protect "block" from the macro expansion. */
    buf = (unsigned char*)block;
    CGUL_CRC16__UPDATE_BLOCK(crc, buf, length);

 out:
    return;
}

void
cgul_crc16__update_range_of_block(cgul_exception_t* cex,
                                  cgul_crc16_t crc,
                                  const void* block,
                                  size_t offset,
                                  size_t length)
{
    unsigned char* buf = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(crc);
    assert(block);

    /* "buf" is used to protect "block" from the macro expansion. */
    buf = (unsigned char*)block + offset;
    CGUL_CRC16__UPDATE_BLOCK(crc, buf, length);

 out:
    return;
}

void
cgul_crc16__update_file(cgul_exception_t* cex,
                        cgul_crc16_t crc,
                        FILE* f)
{
    size_t rcount = 0;
    size_t block_size = 16384;
    unsigned char* block = NULL;
    unsigned char* buf = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(crc);
    assert(f);

    /* Allocate. */
    block = (unsigned char*)cgul_libc__calloc(cex, 1, block_size);
    if (*cex) {
        goto out;
    }

    /* Iterate over each block in the file. */
    for (;;) {

        /* Read the next block. */
        rcount = cgul_libc__fread(cex, block, 1, block_size, f);
        if (*cex) {
            goto out;
        }

        /* Check for EOF. */
        if (rcount == 0) {
            goto out;
        }

        /* "buf" is used to protect "block" from the macro expansion. */
        buf = block;
        CGUL_CRC16__UPDATE_BLOCK(crc, buf, rcount);
    }

 out:

    /* Clean up. */
    if (block) {
        free(block);
    }

    return;
}

void
cgul_crc16__update_fname(cgul_exception_t* cex,
                         cgul_crc16_t crc,
                         const char* fname)
{
    FILE* f = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(crc);
    assert(fname);

    /* Open the file. */
    f = cgul_libc__fopen(cex, fname, "rb");
    if (*cex) {
        goto out;
    }

    /* Update the CRC with the file. */
    cgul_crc16__update_file(cex, crc, f);
    if (*cex) {
        goto out;
    }

 out:

    /* Clean up. */
    if (f) {
        fclose(f);
    }

    return;
}

void
cgul_crc16__reset(cgul_exception_t* cex,
                  cgul_crc16_t crc)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(crc);

    CGUL_CRC16__RESET(crc);

 out:
    return;
}

char*
cgul_crc16__get_table_as_string(cgul_exception_t* cex,
                                cgul_crc16_t crc)
{
    char* rv = NULL;
    unsigned short i = 0;
    unsigned short mask = 0;
    cgul_string_t s = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(crc);

    s = cgul_string__new(cex);
    if (*cex) {
        goto out;
    }

    for (i = 0 ; i < 256 ; ++i) {
        unsigned short tmp = crc->table[i];

        cgul_string__append(cex, s, "0x");
        if (*cex) {
            goto out;
        }

        for (mask = 0x1000 ; mask >= 4 ; mask >>= 4) {
            if (tmp < mask) {
                cgul_string__append_char(cex, s, '0');
                if (*cex) {
                    goto out;
                }
            }
        }

        cgul_string__append_hexadecimal(cex, s, tmp, 0 /* caps */);
        if (*cex) {
            goto out;
        }

        if ((i + 1) % 8 == 0) {
            if (i + 1 < 256) {
                cgul_string__append_char(cex, s, ',');
                if (*cex) {
                    goto out;
                }
            }
            cgul_string__append_char(cex, s, '\n');
            if (*cex) {
                goto out;
            }
        } else {
            cgul_string__append(cex, s, ",  ");
            if (*cex) {
                goto out;
            }
        }
    }

    /* Success. */
    rv = cgul_string__take_value(cex, s);

 out:

    if (s) {
        cgul_string__delete(s);
    }

    return rv;
}
