#ifndef CGUL_CRC16_P_H
#define CGUL_CRC16_P_H

#include "cgul_common.h"
#include "cgul_crc16.h"

CGUL_BEGIN_C

/*
 * NOTE: The C standard guarantees that the "unsigned short" type is
 * at least 16 bits wide.
 */

/**
 * Data encapsulated by the <code>cgul_crc16</code> class.
 */
struct cgul_crc16 {

    /**
     * "g" is the Generator Polynomial (with the most significant bit
     * removed).  The common values for g are:
     *
     *     CRC-CCITT: x^16 + x^12 + x^5 + 1  ==>  g = 0x1021
     *    IBM-CRC-16: x^16 + x^15 + x^2 + 1  ==>  g = 0x8005
     *
     * The convention when dealing with the reflected CRC algorithm is
     * to specify the reflection of the real generator polynomial.
     * (For an example, take a look at the implementation of FCS-16 in
     * RFC 1331.)  Because of this convention, if you are using the
     * reflected CRC algorithm, the common values you should set g to
     * are 0x8408 and 0xA001 (which are the reflections of 0x1021 and
     * 0x8005 respectively).
     */
    unsigned short g;

    /**
       "init" is used to initialize the accumulator register.  By
     * initializing the accumulator to a non-zero value, you prevent a
     * common blind spot caused by invalid leading zeros.  However, by
     * convention, the "init" value is usually 0x0000 except when g is
     * 0x1021.  Then, "init" should be set to 0xffff.
     */
    unsigned short init;

    /**
     * "final_xor" is the final value to XOR the calculated CRC with.
     * This prevents a common blind spot caused by invalid trailing
     * zeros.  By convention however, this is always 0x0000, but don't
     * let that stop you from using 0xffff for private protocols.
     */
    unsigned short final_xor;

    /**
     * Some hardware uses the most significant bit as a parity bit.
     * Thus, the hardware transmits the least significant bits first
     * starting with bit0.  Instead of reflecting the incoming bits and
     * calculating the checksum normally, the bits are fed through the
     * CRC shift register as they arrive.  To compensate, the algorithm
     * to calculate the CRC is different, and the CRC calculated is
     * actually the reflection of the real CRC.  FCS-16 (specified
     * starting with RFC 1331) is an example of a reflected CRC.
     */
    int reflected;

    /**
     * We may need to append the size to the byte stream that gets
     * CRCed.
     */
    int append_size;

    /**
     * We need to keep a count of the bytes we have processed so that
     * we can append it if necessary.
     */
    size_t size;

    /**
     * We may need to XOR the size with sizeXOR before appending.
     */
    size_t size_xor;

    /**
     * The lookup table used by this object.  The table will be
     * generated only once and shared among all other CRC16 objects
     * that use the same table.  Do not try to free this pointer
     * because it points to a lookup table saved in the
     * cgul_crc16__tables tree.
     */
    unsigned short* table;

    /**
     * We need a general purpose register where we can accumulate XOR
     * operations.  This variable is initialized with "init" above.
     */
    unsigned short accumulator;
};

/**
 * This structure keeps track of all the tables we have already
 * generated.
 */
typedef struct cgul_crc16__tables_key {
    unsigned short g;
    int reflected;
    unsigned short* table;
    size_t ref_count;
} *cgul_crc16__tables_key_t;

CGUL_END_C

#endif  /*  CGUL_CRC16_P_H  */
