#include "cgul_hash_p.h"
#include "cgul_hash_node_p.h"
#include "cgul_libc.h"
#include "cgul_merge_sort.h"
#include "cgul_string.h"

/* Forward declaration. */
static unsigned int
cgul_hash__calculate_total_search_distance(cgul_exception_t* cex,
                                           cgul_hash_t h);

/*
 * Notes on Testing:
 * ================
 *
 * The best place to start testing a prospective hash function is with
 * the following program:
 *
 *     tests/main_hash_zeros
 *
 * It simulates passing blocks of zeros to the hash function where
 * each block is one byte longer than the previous one.  It requires a
 * little bit of extra coding to adapt the hash function for use, but
 * it is well worth it as this one test can weed out many bad hash
 * functions.
 *
 * If you don't get any collisions when running test/main_hash_zeros,
 * try feeding the same string of hash values to "dieharder" which can
 * be found at the following URL:
 *
 *     http://www.phy.duke.edu/~rgb/General/dieharder.php
 *
 * To feed hash values into dieharder, look at the following sample in
 * the dieharder source (which I submitted by the way):
 *
 *     libdieharder/rng_jenkins.c
 *
 * You can then test the hash function as follows:
 *
 *     rng_jenkins | dieharder -a -g 200
 *
 * A good way to test the speed of various hash functions under a
 * real-world usage scenario is to use the source in
 * tests/main_hash_list_new which reads lines from an "old" file into
 * a cgul_hash.  It then reads lines from a "new" file and searches
 * the "old" hash for those lines.  It outputs the lines that are in
 * the "new" file but not in the "old" file hence the name "list_new".
 */

/**
 * Default hash capacity.  When the hash is resized, the new capacity
 * is always forced to be a power of two so that index values for
 * buckets can be calculated using a simple mask rather than a modulo
 * operation.
 */
#define CGUL_HASH__DEFAULT_CAPACITY 16

/*
 * Hash Functions -- Pick one.
 */

/*************************************************************************
 * Hash Function -- Jenkins One-at-a-Time
 *************************************************************************/
#if 1
/**
 * Jenkins One-at-a-Time Hash -- This amazing public domain hash
 * function is known to have very good uniform distribution and
 * avalanche behavior: <p>
 *
 * <pre>
 *     http://www.burtleburtle.net/bob/hash/doobs.html
 * </pre>
 *
 * @param[in] s string to hash
 * @param[out] hv hash value
 */
#define CGUL_HASH__HASH(s, hv) \
do { \
    unsigned char c; \
    unsigned char* tmp = ((unsigned char*)(s)); \
    (hv) = 0xaa87231f; \
    for (c = *tmp ; c ; c = *++tmp) { \
        (hv) += c; \
        (hv) += ((hv) << 10); \
        (hv) ^= ((hv) >> 6); \
    } \
    (hv) += ((hv) << 3); \
    (hv) ^= ((hv) >> 11); \
    (hv) += ((hv) << 15); \
} while (0)

unsigned int
cgul_hash__hash_block(void* block,
                      unsigned int block_size)
{
    unsigned int i;
    unsigned int rv = 0xaa87231f;
    unsigned char* tmp = (unsigned char*)block;
    for (i = 0 ; i < block_size ; ++i) {
        rv += tmp[i];
        rv += (rv << 10);
        rv ^= (rv >> 6);
    }
    rv += (rv << 3);
    rv ^= (rv >> 11);
    rv += (rv << 15);
    return rv;
}
#endif

/*************************************************************************
 * Hash Function -- Xorshift
 *************************************************************************/
#if 0
/**
 * Xorshift Hash -- This is an adaptation of Marsaglia's Xorshift PRNG
 * for use as a hash function.  The PRNG is described in the following
 * paper: <p>
 *
 * <pre>
 *     http://www.jstatsoft.com/v08/i14/paper
 * </pre>
 *
 * @param[in] s string to hash
 * @param[out] hv hash value
 */
#define CGUL_HASH__HASH(s, hv) \
do { \
    unsigned char c; \
    unsigned char* tmp = ((unsigned char*)(s)); \
    (hv) = 0xaa87231f; \
    for (c = *tmp ; c ; c = *++tmp) { \
        (hv) ^= ((hv) << 13); \
        (hv) ^= ((hv) >> 17); \
        (hv) ^= ((hv) << 5); \
        (hv) ^= c; \
    } \
} while (0)

unsigned int
cgul_hash__hash_block(void* block,
                      unsigned int block_size)
{
    unsigned int i;
    unsigned int rv = 0xaa87231f;
    unsigned char* tmp = (unsigned char*)block;
    for (i = 0 ; i < block_size ; ++i) {
        rv ^= (rv << 13);
        rv ^= (rv >> 17);
        rv ^= (rv << 5);
        rv ^= tmp[i];
    }
    return rv;
}
#endif

/*************************************************************************
 * Hash Function -- Cyclic Polynomial Hash
 *************************************************************************/
#if 0
#include "cgul_sbox.h"
/**
 * Cyclic Polynomial Hash -- This is the same cyclic polynomial hash
 * function used by <code>cgul_rolling_hash</code>. <p>
 *
 * @param[in] s string to hash
 * @param[out] hv hash value
 */
#define CGUL_HASH__HASH(s, hv) \
do { \
    unsigned char c; \
    unsigned char* tmp = ((unsigned char*)(s)); \
    (hv) = 0; \
    for (c = *tmp ; c ; c = *++tmp) { \
        (hv) = (((hv) << 1) | ((hv) >> 31)) ^ cgul_sbox[c]; \
    } \
} while (0)

unsigned int
cgul_hash__hash_block(void* block,
                      unsigned int block_size)
{
    unsigned int i;
    unsigned int rv = 0;
    unsigned char* tmp = (unsigned char*)block;
    for (i = 0 ; i < block_size ; ++i) {
        rv = ((rv << 1) | (rv >> 31)) ^ cgul_sbox[(unsigned char)tmp[i]];
    }
    return rv;
}
#endif

/*************************************************************************
 * Hash Function -- Multiplicative Hash
 *************************************************************************/
#if 0
/**
 * The following hash function is probably the most popular.  It does
 * not do well when subjected to rigorous statistical analysis, but it
 * is fast and seems to function well in practice.  Whether to
 * multiply by 33 or some other odd number larger than 9 is largely
 * arbitrary.  33 and 31 are often chosen because they can be
 * implemented (as shown below) by shifting followed by addition or
 * some form of subtraction. <p>
 *
 * @param[in] s string to hash
 * @param[out] hv hash value
 */
#define CGUL_HASH__HASH(s, hv) \
do { \
    unsigned char c; \
    unsigned char* tmp = ((unsigned char*)(s)); \
    (hv) = 0xaa87231f; \
    for (c = *tmp ; c ; c = *++tmp) { \
        (hv) += ((hv) << 5);  /* Multiply by 33. */ \
        (hv) += c; \
    } \
} while (0)

unsigned int
cgul_hash__hash_block(void* block,
                      unsigned int block_size)
{
    unsigned int i;
    unsigned int rv = 0xaa87231f;
    unsigned char* tmp = (unsigned char*)block;
    for (i = 0 ; i < block_size ; ++i) {
        rv += rv << 5;
        rv += tmp[i];
    }
    return rv;
}
#endif

/**************************************************************************/

/**
 * This macro calculates the hash taking into account whether the user
 * has registered a callback.
 *
 * @param[in] h <code>cgul_hash</code> instance
 * @param[in] key key to hash
 * @param[out] hv hash value
 */
#define CGUL_HASH__CALCULATE_HASH(h, key, hv) \
do { \
    /* Do not confuse "h", "hf", and "hv". */ \
    if ((h)->hf) { \
        /* Calculate the hash value using the callback hash function. */ \
        (hv) = ((h)->hf)(((void*)(key)), ((h)->hf_data)); \
    } else { \
        /* Calculate the hash value using the in-line hash function. */ \
        CGUL_HASH__HASH((key), (hv)); \
    } \
} while (0)

/**
 * This macro calculates the index to use for the <code>h->v</code>
 * array of buckets.  <code>index</code> should be an <code>unsigned
 * int</code>.  Because the number of buckets is forced to always be a
 * power of two, the index can be calculated using a simple
 * <code>key_mask</code> instead of a relatively expensive modulo
 * operation.
 *
 * @param[in] hv immutable hash value of the key
 * @param[in] key_mask key mask
 * @param[out] index (or bucket) to which <code>s</code> belongs
 */
#define CGUL_HASH__CALCULATE_INDEX(hv, key_mask, index) \
do { \
    (index) = (hv) & (key_mask); \
} while (0)

/**
 * Find the node associated with <code>key</code>.  The caller should
 * have already calculated <code>index</code> by calling
 * <code>CGUL_HASH__CALCULATE_INDEX()</code>. <p>
 *
 * <code>index</code> should be an <code>unsigned int</code>, and
 * <code>node</code> should be a <code>cgul_hash_node_t</code>. <p>
 *
 * @param[in] h <code>cgul_hash</code> instance
 * @param[in] key key
 * @param[in] index index (or bucket) to which <code>key</code> belongs
 * @param[out] node node that was found; othewise, <code>NULL</code>
 */
#define CGUL_HASH__FIND_USING_INDEX(h, key, index, node) \
do { \
    cgul_hash_node_t tmp; \
    (node) = NULL; \
    for (tmp = (h)->v[(index)] ; tmp ; tmp = tmp->next) { \
        /* Call the comparison function. */ \
        if (((h)->cf)((key), tmp->key) == 0) { \
            /* Success. */ \
            (node) = tmp; \
            break; \
        } \
    } \
} while (0)

/**
 * Insert <code>hash_node</code> at the front of the bucket in
 * <code>h</code> that corresponds to <code>v_index</code>.  This code
 * only inserts the hash node and updates the accounting information
 * related to hash nodes.  This code is factored out because
 * <code>cgul_hash__insert()</code> and
 * <code>cgul_hash__resize()</code> both need it. <p>
 *
 * Notice that this macro does not affect the embedded "older/younger"
 * linked list because resizing the hash should not have any effect on
 * the relative ages of the nodes. <p>
 *
 * @param[in] h <code>cgul_hash</code> instance
 * @param[in] v_index index into <code>h->v</code>
 * @param[in] hash_node hash node to insert
 */
#define CGUL_HASH__INSERT_HASH_NODE(h, v_index, hash_node) \
do { \
    /* Keep count of the total number of collisions. */ \
    if ((h)->v[(v_index)]) { \
        ++((h)->total_collisions); \
    } \
    /* Insert the node at the front of the list for the bucket. */ \
    (hash_node)->prev = NULL; \
    (hash_node)->next = (h)->v[(v_index)]; \
    (h)->v[(v_index)] = (hash_node); \
    if ((hash_node)->next) { \
        (hash_node)->next->prev = (hash_node); \
    } \
    /* Increment the hash size. */ \
    ++((h)->size); \
} while (0)

/**
 * Remove <code>hash_node</code> from <code>h</code>.  This code only
 * removes the hash node and updates the accounting information
 * related to hash nodes.  This code is factored out because
 * (originally) <code>cgul_hash__remove_node()</code> and
 * <code>cgul_hash__resize()</code> both needed it.  (This is no
 * longer the case.) <p>
 *
 * Notice is that this macro does not affect the embedded
 * "older/younger" linked list because resizing the hash should not
 * have any effect on the relative ages of the nodes. <p>
 *
 * @param[in] h <code>cgul_hash</code> instance
 * @param[in] v_index index into <code>h->v</code>
 * @param[in] hash_node hash node to remove
 */
#define CGUL_HASH__REMOVE_HASH_NODE(h, v_index, hash_node)  \
do { \
    /* Remove the node from the list for the bucket. */ \
    if ((hash_node)->next) { \
        (hash_node)->next->prev = (hash_node)->prev; \
    } \
    if ((hash_node)->prev) { \
        (hash_node)->prev->next = (hash_node)->next; \
    }\
    if ((h)->v[(v_index)] == (hash_node)) { \
        (h)->v[(v_index)] = (hash_node)->next; \
    } \
    /* Just in case. */ \
    (hash_node)->prev = NULL; \
    (hash_node)->next = NULL; \
    /* Decrement the hash size. */ \
    --((h)->size); \
} while (0)

/**
 * This function rounds <code>x</code> up to the next power of 2
 * unless that would cause overflow.  Then it rounds down to the
 * previous power of 2.
 *
 * @param[in] x
 * @return <code>x</code> rounded to power of 2
 */
static unsigned int
cgul_hash__round_to_power_of_2(unsigned int x)
{
    unsigned int rv = 0;

    /* Shift rv left until it is larger than x (careful to avoid overflow). */
    for (rv = 1 ; (rv << 1) != 0 ; rv <<= 1) {
        if (rv >= x) {
            break;
        }
    }

    return rv;
}

/**
 * This method is used by <code>cgul_hash__resize()</code> to change
 * the physical size of the hash.  If successful, all of the buckets
 * in the hash table are reset to <code>NULL</code>, and the caller
 * must reconstruct the hash from the oldest/youngest linked list.  If
 * an out-of-memory error occurs, the exception is quashed.  In this
 * case, it is intended that the hash will just increase in size
 * beyond what <code>h->load_factor</code> specifies.
 *
 * @param[in] cex c-style exception
 * @param[in] h <code>cgul_hash</code> instance
 * @param[in] suggested_capacity suggested capacity
 * @return whether the resize occurred
 */
static int
cgul_hash__resize_physical_only(cgul_exception_t* cex,
                                cgul_hash_t h,
                                unsigned int suggested_capacity)
{
    int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(h);

#if 0
    if (h && h->size) {
        unsigned int d = cgul_hash__calculate_total_search_distance(cex, h);
        printf("average search distance: %g\n",
               (float)d / (float)h->size);
        fflush(stdout);
    }
#endif

    /* Round suggested_capacity to the next power of 2 (unless it
     * would cause overflow then round it to the previous power of 2). */
    suggested_capacity = cgul_hash__round_to_power_of_2(suggested_capacity);

    /* Attempt to allocate the new capacity. */
    h->v = (cgul_hash_node_t*)cgul_libc__realloc(
               cex, h->v, suggested_capacity * sizeof(cgul_hash_node_t));
    if (*cex) {
        /* Quash the exception.  The hash just won't resize in this case. */
        cgul_exception__delete(*cex);
        *cex = NULL;
        goto out;
    }

    /* Remember the new capacity. */
    h->v_capacity = suggested_capacity;

    /* Clear the size. */
    h->size = 0;

    /* Calculate the new size limit. */
    h->size_limit = (size_t)(h->load_factor * h->v_capacity);

    /* Calculate the new key mask. */
    h->key_mask = h->v_capacity - 1;

    /* Clear the buckets.  (The hash table will be reconstructed by
     * the caller from the oldest/youngest linked list.) */
    memset(h->v, 0, h->v_capacity * sizeof(cgul_hash_node_t));

    /* Success. */
    rv = 1;

 out:
    return rv;
}

/**
 * This method resizes the hash without altering its contents.  It
 * increases or decreases the number of available buckets and rehashes
 * all nodes so they are placed in the correct, new bucket.  The
 * number of buckets will be at least <code>suggested_capacity</code>.
 * If an error occurs, an exception is thrown, but the hash is still
 * safe to use. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] h <code>cgul_hash</code> instance
 * @param[in] suggested_capacity suggested capacity
 * @return whether the hash could be resized
 */
static int
cgul_hash__resize(cgul_exception_t* cex,
                  cgul_hash_t h,
                  unsigned int suggested_capacity)
{
    int rv = 0;
    unsigned int key_mask = 0;
    unsigned int new_index = 0;
    cgul_hash_node_t curr = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(h);

    /* Sanity check. */
    if (suggested_capacity == 0) {
        goto out;
    }

    /* Try to physically resize the hash without rehashing the nodes.
     * This can fail if we run out of memory.  This is the only point
     * of failure in this method. */
    rv = cgul_hash__resize_physical_only(cex, h, suggested_capacity);
    if (!rv) {
        goto out;
    }

    /* It is faster to cache the key mask. */
    key_mask = h->key_mask;

    /* Reconstruct the hash table by rehashing each node.  Note that
     * the call above to cgul_hash__resize_physical_only() sets all
     * the buckets in the new hash table to NULL so we only have to
     * insert the existing hash nodes.  (We do not need to remove the
     * hash nodes from the old hash table.) */
    for (curr = h->oldest ; curr ; curr = curr->younger) {

        /* Calculate the new index. */
        CGUL_HASH__CALCULATE_INDEX(curr->hv, key_mask, new_index);

        /* Set the new index and insert the node in its new location. */
        curr->index = new_index;
        CGUL_HASH__INSERT_HASH_NODE(h, curr->index, curr);
    }

 out:
    return rv;
}

/**
 * Get the number of keys in the bucket in <code>h</code> specified by
 * <code>index</code>.  This method is not efficient and should only
 * be used sparingly.  If <code>index</code> is out of bounds, an
 * exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] h <code>cgul_hash</code> instance
 * @param[in] index index for bucket
 * @return bucket size
 */
static size_t
cgul_hash__get_bucket_size(cgul_exception_t* cex,
                           cgul_hash_t h,
                           unsigned int index)
{
    size_t rv = 0;
    cgul_hash_node_t n = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(h);

    /* Sanity check. */
    if (index >= h->v_capacity) {
        *cex = cgul_exception__new(
                   "cgul_hash__get_bucket_size: Invalid bucket index: %u",
                   index);
        goto out;
    }

    /* We have to iterate to count because the embeded "prev/next"
     * linked list is a minimal implementation that are not able to
     * tell you how many nodes exist in the list. */
    for (n = h->v[index] ; n ; n = n->next) {
        ++rv;
    }

 out:
    return rv;
}

/**
 * Calculate the total search distance for all the keys in
 * <code>h</code>.  This method is not efficient and should only be
 * used sparingly.
 *
 * @param[in,out] cex c-style exception
 * @param[in] h <code>cgul_hash</code> instance
 * @return average search distance
 */
static unsigned int
cgul_hash__calculate_total_search_distance(cgul_exception_t* cex,
                                           cgul_hash_t h)
{
    unsigned int rv = 0;
    unsigned int i = 0;
    unsigned int j = 0;
    cgul_hash_node_t n = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(h);

    /* Iterate over the buckets. */
    for (i = 0 ; i < h->v_capacity ; ++i) {
        /* Iterate over the linked list calculating the exact search
         * distance to each node and adding this value to the total
         * search distance (rv).  For example, the first node on each
         * list has an exact search distance of 1.  The second node
         * has an exact search distance of 2.  Etc. */
        j = 0;
        for (n = h->v[i] ; n ; n = n->next) {
            rv += ++j;
        }
    }

 out:
    return rv;
}

/**
 * Create a new custom <code>cgul_hash</code> object that uses
 * <code>hf</code> as the hash function and <code>cf</code> as the
 * comparison function and that is initialized with the approximate
 * <code>capacity</code> requested.  With the help of an adaptor, the
 * <code>cgul_hash__hash_block()</code> function is a good choice for
 * <code>hf</code>.  If memory cannot be allocated, <code>NULL</code>
 * is returned, and an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] hf hash function
 * @param[in] hf_data hash function client data
 * @param[in] cf comparison function
 * @param[in] capacity initial capacity
 * @return new <code>cgul_hash</code> instance
 * @see cgul_hash__set_load_factor()
 */
static cgul_hash_t
cgul_hash__new_custom_with_capacity(cgul_exception_t* cex,
                                    cgul_hash__hash_t hf,
                                    void* hf_data,
                                    cgul_hash__compare_t cf,
                                    unsigned int capacity)
{
    cgul_hash_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Allocate memory for the new cgul_hash object. */
    rv = (cgul_hash_t)cgul_libc__calloc(cex, 1, sizeof(*rv));
    if (*cex) {
        goto out;
    }

    /* Remember the hash function callback. */
    rv->hf = hf;
    rv->hf_data = hf_data;

    /* Remember the comparison function callback. */
    rv->cf = cf;

    /* Set the default load factor. */
    cgul_hash__set_load_factor(cex, rv, 0.75);

    /* Resize the hash. */
    capacity = CGUL_MAX(capacity, 1);
    cgul_hash__resize(cex, rv, capacity);
    if (*cex) {
        goto out;
    }

 out:

    /* Clean up if an exception is thrown. */
    if (*cex && rv) {
        cgul_hash__delete(rv);
        rv = NULL;
    }

    return rv;
}

cgul_hash_t
cgul_hash__new(cgul_exception_t* cex)
{
    cgul_hash_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    rv = cgul_hash__new_custom_with_capacity(
             cex,
             (cgul_hash__hash_t)NULL,
             NULL,
             (cgul_hash__compare_t)strcmp,
             CGUL_HASH__DEFAULT_CAPACITY);

 out:
    return rv;
}

cgul_hash_t
cgul_hash__new_custom(cgul_exception_t* cex,
                      cgul_hash__hash_t hf,
                      void* hf_data,
                      cgul_hash__compare_t cf)
{
    cgul_hash_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    rv = cgul_hash__new_custom_with_capacity(
             cex, hf, hf_data, cf, CGUL_HASH__DEFAULT_CAPACITY);

 out:
    return rv;
}

void
cgul_hash__delete(cgul_hash_t h)
{
    cgul_exception_t local = NULL;

    if (h) {

        /* Free the all the hash nodes. */
        cgul_hash__clear(&local, h);

        /* Free the vector of buckets. */
        if (h->v) {
            free(h->v);
        }

        /* Free the hash itself. */
        free(h);
    }
}

void
cgul_hash__free_keys(cgul_hash_t h)
{
    cgul_hash_node_t curr = NULL;

    if (h) {
        /* Iterate over each node. */
        for (curr = h->oldest ; curr ; curr = curr->younger) {
            /* Free the key. */
            free((void*)curr->key);
        }
    }
}

void
cgul_hash__free_values(cgul_hash_t h)
{
    cgul_hash_node_t curr = NULL;

    if (h) {
        /* Iterate over each node. */
        for (curr = h->oldest ; curr ; curr = curr->younger) {
            /* Free the value. */
            free((void*)curr->value);
        }
    }
}

float
cgul_hash__get_load_factor(cgul_exception_t* cex,
                           cgul_hash_t h)
{
    float rv = 0.0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(h);

    rv = h->load_factor;

 out:
    return rv;
}

void
cgul_hash__set_load_factor(cgul_exception_t* cex,
                           cgul_hash_t h,
                           float load_factor)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(h);

    /* Only change the load factor if the new value is sane. */
    if (load_factor > 0.0) {

        /* Save the load factor. */
        h->load_factor = load_factor;

        /* Calculate the new size limit. */
        h->size_limit = (size_t)(load_factor * h->v_capacity);
    }

 out:
    return;
}

void
cgul_hash__swap(cgul_exception_t* cex,
                cgul_hash_t h1,
                cgul_hash_t h2)
{
    struct cgul_hash tmp;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(h1);
    assert(h2);

    /* Nothing to swap if they point to the same objects. */
    if (h1 == h2) {
        goto out;
    }

    /* A shallow copy, by design, is exactly what is needed. */
    tmp = *h1;
    *h1 = *h2;
    *h2 = tmp;

 out:
    return;
}

int
cgul_hash__is_empty(cgul_exception_t* cex,
                    cgul_hash_t h)
{
    int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(h);

    rv = (h->size == (size_t)0);

 out:
    return rv;
}

const void**
cgul_hash__get_keys(cgul_exception_t* cex,
                    cgul_hash_t h,
                    size_t* key_count)
{
    unsigned int i = 0;
    cgul_hash_node_t n = NULL;
    const void** rv = NULL;
    size_t key_count_tmp = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(h);

    /* Assume failure. */
    if (key_count) {
        *key_count = 0;
    }

    /* Allocate the array that will be returned.  Note that the array
     * has an extra slot to hold a NULL sentinel. */
    key_count_tmp = h->size;
    rv = (const void**)cgul_libc__calloc(cex,
                                         key_count_tmp + 1,
                                         sizeof(*rv));
    if (*cex) {
        goto out;
    }

    /* Add the nodes to the array plus a trailing sentinel. */
    for( i = 0, n = h->oldest ;
         (i < key_count_tmp) && n ;
         ++i, n = n->younger )
    {
        rv[i] = n->key;
    }
    rv[i] = NULL;

    /* Return the key count. */
    if (key_count) {
        *key_count = key_count_tmp;
    }

 out:
    return rv;
}

const void**
cgul_hash__get_sorted_keys(cgul_exception_t* cex,
                           cgul_hash_t h,
                           size_t* key_count)
{
    const void** rv = NULL;
    size_t key_count_tmp = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(h);

    /* Assume failure. */
    if (key_count) {
        *key_count = 0;
    }

    /* Get the unsorted pointers. */
    rv = cgul_hash__get_keys(cex, h, &key_count_tmp);
    if (*cex) {
        goto out;
    }

    /* Sort the pointers. */
    cgul_merge_sort_pointers(cex,
                             (void**)rv,
                             key_count_tmp,
                             (cgul_merge_sort__compare_t)(h->cf));
    if (*cex) {
        goto out;
    }

    if (key_count) {
        *key_count = key_count_tmp;
    }

 out:

    /* Clean up. */
    if (*cex && rv) {
        free((void**)rv);
        rv = NULL;
    }

    return rv;
}

cgul_hash_node_t
cgul_hash__get_front(cgul_exception_t* cex,
                     cgul_hash_t h)
{
    cgul_hash_node_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(h);

    /* "front" is a synonym for "oldest." */
    rv = h->oldest;

 out:
    return rv;
}

cgul_hash_node_t
cgul_hash__get_back(cgul_exception_t* cex,
                    cgul_hash_t h)
{
    cgul_hash_node_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(h);

    /* "back" is a synonym for "youngest." */
    rv = h->youngest;

 out:
    return rv;
}

cgul_hash_node_t
cgul_hash__get_oldest(cgul_exception_t* cex,
                      cgul_hash_t h)
{
    cgul_hash_node_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(h);

    rv = h->oldest;

 out:
    return rv;
}

void
cgul_hash__set_oldest(cgul_exception_t* cex,
                      cgul_hash_t h,
                      cgul_hash_node_t n)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(h);
    assert(h->oldest);
    assert(n);

    /* There is nothing to do if n is already the oldest. */
    if (n == h->oldest) {
        goto out;
    }

    /* If n is currently the youngest, it will not be when this method
     * returns.  In this case, we need to remember which node is the
     * new youngest node. */
    if (n == h->youngest) {
        h->youngest = n->older;
    }

    /* Carefully remove n from the list. */
    if (n->older) {
        n->older->younger = n->younger;
    }
    if (n->younger) {
        n->younger->older = n->older;
    }

    /* Insert n as the oldest node.  We assert h->oldest above because
     * any non-empty hash has an oldest node, and because n is a
     * node in h, h is non-empty. */
    h->oldest->older = n;
    n->younger = h->oldest;
    n->older = NULL;

    /* Remember that n is now the oldest node. */
    h->oldest = n;

 out:
    return;
}

cgul_hash_node_t
cgul_hash__get_youngest(cgul_exception_t* cex,
                        cgul_hash_t h)
{
    cgul_hash_node_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(h);

    rv = h->youngest;

 out:
    return rv;
}

void
cgul_hash__set_youngest(cgul_exception_t* cex,
                        cgul_hash_t h,
                        cgul_hash_node_t n)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(h);
    assert(h->youngest);
    assert(n);

    /* There is nothing to do if n is already the youngest. */
    if (n == h->youngest) {
        goto out;
    }

    /* If n is currently the oldest, it will not be when this method
     * returns.  In this case, we need to remember which node is the
     * new oldest node. */
    if (n == h->oldest) {
        h->oldest = n->younger;
    }

    /* Carefully remove n from the list. */
    if (n->older) {
        n->older->younger = n->younger;
    }
    if (n->younger) {
        n->younger->older = n->older;
    }

    /* Insert n as the youngest node.  We assert h->youngest above
     * because any non-empty hash has a youngest node, and
     * because n is a node in h, h is non-empty. */
    h->youngest->younger = n;
    n->older = h->youngest;
    n->younger = NULL;

    /* Remember that n is now the youngest node. */
    h->youngest = n;

 out:
    return;
}

int
cgul_hash__insert(cgul_exception_t* cex,
                  cgul_hash_t h,
                  const void* key,
                  cgul_hash_node_t* n)
{
    int rv = 0;
    unsigned int hv = 0;
    unsigned int index = 0;
    cgul_hash_node_t tmp1 = NULL;
    cgul_hash_node_t tmp2 = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(h);
    assert(key);

    /* Reset the return node in case an exception is thrown. */
    if (n) {
        *n = NULL;
    }

    /* Try to only hash the key once.  If the hash gets resized below
     * we'll have to calculate it again however. */
    CGUL_HASH__CALCULATE_HASH(h, key, hv);
    CGUL_HASH__CALCULATE_INDEX(hv, h->key_mask, index);

    /*
     * If a node with the same "key" already exists, use it.
     */

    CGUL_HASH__FIND_USING_INDEX(h, key, index, tmp1);
    if (tmp1) {
        /* A node already exists for this key.  Return the existing
         * node in n, and do not return 1 because nothing was
         * inserted. */
        if (n) {
            *n = tmp1;
        }
        /* A return value of 0 means the key could not be inserted
         * because one already exists.  It doesn't mean failure.  It
         * just means that a new node wasn't inserted and that we are
         * returning a pre-existing node instead. */
        rv = 0;
        goto out;
    }
    
    /*
     * We could not find an existing node with the same key as "key"
     * so initialize a new cgul_hash_node object and insert it.
     */

    /* Decide if the hash needs to be resized.  You need to do this
     * before inserting the new node because resizing the hash
     * invalidates any existing cgul_hash_node object including the
     * node you would have just inserted that needs to be returned to
     * the caller.  */
    assert(h->v_capacity);
    if (h->size > h->size_limit) {
        /* Try to double the capacity. */
        if (cgul_hash__resize(cex, h, (h->v_capacity << 1))) {
            /* We have to calculate the index again using the new capacity. */
            CGUL_HASH__CALCULATE_INDEX(hv, h->key_mask, index);
        }
    }

    /* Create a new node that we can insert.  If the operation is
     * successful, this is the node that will be returned to the
     * caller. */
    tmp2 = cgul_hash_node__new(cex);
    if (*cex) {
        goto out;
    }

    /* Set the key. */
    tmp2->key = key;

    /* Remember the hash value. */
    tmp2->hv = hv;

    /* Remember the index. */
    tmp2->index = index;

    /* Insert the node into the hash without affecting the embedded
     * "older/younger" linked list. */
    CGUL_HASH__INSERT_HASH_NODE(h, index, tmp2);
    if (*cex) {
        goto out;
    }

    /* Keep the "oldest" pointer up to date.  If h->oldest is NULL, it
     * means the hash is empty.  When we insert n, it becomes both the
     * oldest and youngest node. */
    if (h->oldest == NULL) {
        h->oldest = tmp2;
    }

    /* Update "youngest" because we just inserted a new node.  Also
     * remember to set the "younger" pointer of the node that was
     * previously the youngest and the "older" pointer of the node
     * that is now the youngest. */
    if (h->youngest) {
        h->youngest->younger = tmp2;
        tmp2->older = h->youngest;
    }
    h->youngest = tmp2;

    /* If "oldest" is NULL, it means the hash was empty and we just
     * inserted the first node which is both the youngest and oldest
     * node. */
    if (h->oldest == NULL) {
        h->oldest = tmp2;
    }

    /*
     * Success.
     */

    /* Return the new node to the caller. */
    if (n) {
        *n = tmp2;
    }

    /* Make sure the new node is not deleted when we clean up. */
    tmp2 = NULL;

    /* Let the caller know that we are returning a new node, not a
     * pre-existing node. */
    rv = 1;

 out:

    /* Clean up tmp if it was not inserted. */
    if (tmp2) {
        cgul_hash_node__delete(tmp2);
    }

    return rv;
}

cgul_hash_node_t
cgul_hash__find(cgul_exception_t* cex,
                cgul_hash_t h,
                const void* key)
{
    unsigned int hv = 0;
    unsigned int index = 0;
    cgul_hash_node_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(h);
    assert(key);

    CGUL_HASH__CALCULATE_HASH(h, key, hv);
    CGUL_HASH__CALCULATE_INDEX(hv, h->key_mask, index);
    CGUL_HASH__FIND_USING_INDEX(h, key, index, rv);

 out:
    return rv;
}

int
cgul_hash__remove(cgul_exception_t* cex,
                  cgul_hash_t h,
                  const void* key_in,
                  void** key_out,
                  void** value_out)
{
    int rv = 0;
    cgul_hash_node_t n = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(h);
    assert(key_in);

    /* Clear the user's output parameters. */
    if (key_out) {
        *key_out = NULL;
    }
    if (value_out) {
        *value_out = NULL;
    }

    /* Find the node for the hash and remove it returning "*key_out"
     * and "*value_out" to the user. */
    n = cgul_hash__find(cex, h, key_in);
    if (n)  {
        if (key_out) {
            *key_out = (void*)n->key;
        }
        if (value_out) {
            *value_out = (void*)n->value;
        }
        cgul_hash__remove_node(cex, h, n);
        rv = 1;
    }

 out:
    return rv;
}

void
cgul_hash__remove_node(cgul_exception_t* cex,
                       cgul_hash_t h,
                       cgul_hash_node_t n)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(h);
    assert(n);

    /* Remove the node.  This removes the node from the bucket and
     * updates the accounting information to reflect the change.
     * (Originally, this was factored out because cgul_hash__resize()
     * needs this feature in order to be efficient, but this is no
     * longer the case.) */
    CGUL_HASH__REMOVE_HASH_NODE(h, n->index, n);

    /* While "n" is still valid, update the "oldest" and "youngest"
     * pointers for the entire hash. */
    if (n == h->oldest) {
        h->oldest = n->younger;
    }
    if (n == h->youngest) {
        h->youngest = n->older;
    }

    /* While "n" is still valid, update the "older/younger" linked
     * list that runs through all the nodes. */
    if (n->older) {
        n->older->younger = n->younger;
    }
    if (n->younger) {
        n->younger->older = n->older;
    }

    /* Free the hash node. */
    cgul_hash_node__delete(n);

    /* Decide if the hash needs to be resized.  We try to reduce the
     * capacity by 1/2 when the current size is 1/4 the size limit
     * unless it would reduce the capacity below the default capacity. */
    assert(h->v_capacity);
    if (   ((h->size << 2) <= h->size_limit)
        && (h->v_capacity > CGUL_HASH__DEFAULT_CAPACITY)) {
        /* Try to halve the capacity. */
        cgul_hash__resize(cex, h, (h->v_capacity >> 1));
    }

 out:
    return;
}

int
cgul_hash__remove_front(cgul_exception_t* cex,
                        cgul_hash_t h,
                        void** key_out,
                        void** value_out)
{
    int rv = 0;
    cgul_hash_node_t n = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(h);

    /* Clear the user's output parameters. */
    if (key_out) {
        *key_out = NULL;
    }
    if (value_out) {
        *value_out = NULL;
    }

    /* Get the front node. */
    n = h->oldest;
    if (!n) {
        goto out;
    }

    /* Return the key and value. */
    if (key_out) {
        *key_out = (void*)n->key;
    }
    if (value_out) {
        *value_out = (void*)n->value;
    }

    /* Remove the node. */
    cgul_hash__remove_node(cex, h, n);

    /* Success. */
    rv = 1;

 out:
    return rv;
}

int
cgul_hash__remove_back(cgul_exception_t* cex,
                       cgul_hash_t h,
                       void** key_out,
                       void** value_out)
{
    int rv = 0;
    cgul_hash_node_t n = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(h);

    /* Clear the user's output parameters. */
    if (key_out) {
        *key_out = NULL;
    }
    if (value_out) {
        *value_out = NULL;
    }

    /* Get the back node. */
    n = h->youngest;
    if (!n) {
        goto out;
    }

    /* Return the key and value. */
    if (key_out) {
        *key_out = (void*)n->key;
    }
    if (value_out) {
        *value_out = (void*)n->value;
    }

    /* Remove the node. */
    cgul_hash__remove_node(cex, h, n);

    /* Success. */
    rv = 1;

 out:
    return rv;
}

int
cgul_hash__remove_oldest(cgul_exception_t* cex,
                         cgul_hash_t h,
                         void** key_out,
                         void** value_out)
{
    int rv = 0;
    cgul_hash_node_t n = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(h);

    /* Clear the user's output parameters. */
    if (key_out) {
        *key_out = NULL;
    }
    if (value_out) {
        *value_out = NULL;
    }

    /* Get the oldest node. */
    n = h->oldest;
    if (!n) {
        goto out;
    }

    /* Return the key and value. */
    if (key_out) {
        *key_out = (void*)n->key;
    }
    if (value_out) {
        *value_out = (void*)n->value;
    }

    /* Remove the node. */
    cgul_hash__remove_node(cex, h, n);

    /* Success. */
    rv = 1;

 out:
    return rv;
}

int
cgul_hash__remove_youngest(cgul_exception_t* cex,
                           cgul_hash_t h,
                           void** key_out,
                           void** value_out)
{
    int rv = 0;
    cgul_hash_node_t n = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(h);

    /* Clear the user's output parameters. */
    if (key_out) {
        *key_out = NULL;
    }
    if (value_out) {
        *value_out = NULL;
    }

    /* Get the youngest node. */
    n = h->youngest;
    if (!n) {
        goto out;
    }

    /* Return the key and value. */
    if (key_out) {
        *key_out = (void*)n->key;
    }
    if (value_out) {
        *value_out = (void*)n->value;
    }

    /* Remove the node. */
    cgul_hash__remove_node(cex, h, n);

    /* Success. */
    rv = 1;

 out:
    return rv;
}

void
cgul_hash__clear(cgul_exception_t* cex,
                 cgul_hash_t h)
{
    cgul_hash_node_t curr = NULL;
    cgul_hash_node_t next = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(h);

    /* Carefully delete each hash node. */
    for (curr = h->youngest ; curr ; curr = next) {
        /* Get the next node while we still can. */
        next = curr->older;
        /* Delete this node. */
        cgul_hash_node__delete(curr);
    }

    /* Reset the buckets. */
    memset(h->v, 0, h->v_capacity * sizeof(cgul_hash_node_t));

    /* Reset the size. */
    h->size = 0;

    /* Reset the "oldest" and "youngest" pointers. */
    h->oldest = NULL;
    h->youngest = NULL;

 out:
    return;
}

size_t
cgul_hash__get_size(cgul_exception_t* cex,
                    cgul_hash_t h)
{
    size_t rv = (size_t)0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(h);

    rv = h->size;

 out:
    return rv;
}

size_t
cgul_hash__get_total_collisions(cgul_exception_t* cex,
                                cgul_hash_t h)
{
    size_t rv = (size_t)0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(h);

    rv = h->total_collisions;

 out:
    return rv;
}

void
cgul_hash__foldl_keys(cgul_exception_t* cex,
                      cgul_hash_t h,
                      cgul_hash__fold_key_t f,
                      void* data)
{
    int f_rv = 0;
    cgul_hash_node_t curr = NULL;
    cgul_hash_node_t younger = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(h);
    assert(f);

    /* Iterate over the nodes in the hash. */
    for (curr = h->oldest ; curr ; curr = younger) {
        /* Get the younger node now in case f deletes curr. */
        younger = curr->younger;
        /* Call the combining function. */
        f_rv = (*f)(cex, curr->key, data);
        if (*cex || !f_rv) {
            goto out;
        }
    }

 out:
    return;
}

void
cgul_hash__foldr_keys(cgul_exception_t* cex,
                      cgul_hash_t h,
                      cgul_hash__fold_key_t f,
                      void* data)
{
    int f_rv = 0;
    cgul_hash_node_t curr = NULL;
    cgul_hash_node_t older = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(h);
    assert(f);

    /* Iterate over the nodes in the hash. */
    for (curr = h->youngest ; curr ; curr = older) {
        /* Get the older node now in case f deletes curr. */
        older = curr->older;
        /* Call the combining function. */
        f_rv = (*f)(cex, curr->key, data);
        if (*cex || !f_rv) {
            goto out;
        }
    }

 out:
    return;
}

void
cgul_hash__foldl_values(cgul_exception_t* cex,
                        cgul_hash_t h,
                        cgul_hash__fold_value_t f,
                        void* data)
{
    int f_rv = 0;
    cgul_hash_node_t curr = NULL;
    cgul_hash_node_t younger = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(h);
    assert(f);

    /* Iterate over the nodes in the hash. */
    for (curr = h->oldest ; curr ; curr = younger) {
        /* Get the younger node now in case f deletes curr. */
        younger = curr->younger;
        /* Call the combining function. */
        f_rv = (*f)(cex, (void*)curr->value, data);
        if (*cex || !f_rv) {
            goto out;
        }
    }

 out:
    return;
}

void
cgul_hash__foldr_values(cgul_exception_t* cex,
                        cgul_hash_t h,
                        cgul_hash__fold_value_t f,
                        void* data)
{
    int f_rv = 0;
    cgul_hash_node_t curr = NULL;
    cgul_hash_node_t older = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(h);
    assert(f);

    /* Iterate over the nodes in the hash. */
    for (curr = h->youngest ; curr ; curr = older) {
        /* Get the older node now in case f deletes curr. */
        older = curr->older;
        /* Call the combining function. */
        f_rv = (*f)(cex, (void*)curr->value, data);
        if (*cex || !f_rv) {
            goto out;
        }
    }

 out:
    return;
}

void
cgul_hash__foldl_pairs(cgul_exception_t* cex,
                       cgul_hash_t h,
                       cgul_hash__fold_pair_t f,
                       void* data)
{
    int f_rv = 0;
    cgul_hash_node_t curr = NULL;
    cgul_hash_node_t younger = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(h);
    assert(f);

    /* Iterate over the nodes in the hash. */
    for (curr = h->oldest ; curr ; curr = younger) {
        /* Get the younger node now in case f deletes curr. */
        younger = curr->younger;
        /* Call the combining function. */
        f_rv = (*f)(cex, curr->key, (void*)curr->value, data);
        if (*cex || !f_rv) {
            goto out;
        }
    }

 out:
    return;
}

void
cgul_hash__foldr_pairs(cgul_exception_t* cex,
                       cgul_hash_t h,
                       cgul_hash__fold_pair_t f,
                       void* data)
{
    int f_rv = 0;
    cgul_hash_node_t curr = NULL;
    cgul_hash_node_t older = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(h);
    assert(f);

    /* Iterate over the nodes in the hash. */
    for (curr = h->youngest ; curr ; curr = older) {
        /* Get the older node now in case f deletes curr. */
        older = curr->older;
        /* Call the combining function. */
        f_rv = (*f)(cex, curr->key, (void*)curr->value, data);
        if (*cex || !f_rv) {
            goto out;
        }
    }

 out:
    return;
}

void
cgul_hash__traverse(cgul_exception_t* cex,
                    cgul_hash_t h,
                    cgul_hash__traverse_t f,
                    void* data)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    cgul_hash__traverse_range(cex, h, NULL, NULL, f, data);

 out:
    return;
}

void
cgul_hash__traverse_range(cgul_exception_t* cex,
                          cgul_hash_t h,
                          cgul_hash_node_t first,
                          cgul_hash_node_t last,
                          cgul_hash__traverse_t f,
                          void* data)
{
    int f_rv = 0;
    cgul_hash_node_t curr = NULL;
    cgul_hash_node_t younger = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Do not assert(first) nor assert(last) because a NULL value
     * indicate that we should begin or end with the first or last
     * node of the entire tree respectively. */
    assert(h);
    assert(f);

    if (first == NULL) {
        first = h->oldest;
    }

    if (last == NULL) {
        last = h->youngest;
    }

    /* In order to provide "f" with a context in which it is safe for
     * "f" to delete curr, you have to be careful to dereference
     * curr before calling "f". */
    for (curr = first ; curr ; curr = younger) {
        younger = curr->younger;
        f_rv = (*f)(cex, h, curr, data);
        if (*cex) {
            goto out;
        }
        /* We are done if f returned false or if we just processed the
         * last node. */
        if (!f_rv || (curr == last)) {
            break;
        }
    }

 out:
    return;
}

char*
cgul_hash__get_stats(cgul_exception_t* cex,
                     cgul_hash_t h)
{
    char* rv = NULL;
    cgul_string_t s = NULL;
    unsigned int i = 0;
    unsigned int total_search_distance = 0;
    unsigned int bucket_count_size = 0;
    size_t bucket_count[17] = {0};

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(h);

    /* Determine the number of entries to report based on the size of
     * the "bucket_count" array. */
    bucket_count_size = sizeof(bucket_count) / sizeof(bucket_count[0]);

    /* Allocate the return value. */
    s = cgul_string__new(cex);

    /* Count the number of entries in each bucket. */
    for (i = 0 ; i < h->v_capacity ; ++i) {
        size_t index = cgul_hash__get_bucket_size(cex, h, i);
        if (index >= bucket_count_size) {
            index = bucket_count_size - 1;
        }
        ++bucket_count[index];
    }

    /* Append the overview. */
    cgul_string__append_sprintf(
        cex, s, "Hash Stats: %lu entries, %lu buckets\n",
        (unsigned long)h->size,
        (unsigned long)h->v_capacity);

    /* Append the bucket classifications. */
    for (i = 0 ; i < bucket_count_size ; ++i) {
        cgul_string__append_sprintf(
            cex, s, "%10lu buckets with %3u %sentries.\n",
            (unsigned long)bucket_count[i],
            i,
            (i == bucket_count_size - 1) ? "or more " : "");
    }

    /* Append the total and average search distance. */
    total_search_distance = cgul_hash__calculate_total_search_distance(cex, h);
    cgul_string__append_sprintf(cex, s, "total search distance: %u\n",
                                total_search_distance);
    cgul_string__append_sprintf(cex, s, "average search distance: %g\n",
                                (float)total_search_distance / (float)h->size);

    /* Append the load factor. */
    assert(h->v_capacity);
    cgul_string__append_sprintf(cex, s, "load factor: %g\n",
                                (float)h->size / (float)h->v_capacity);

    /* Append the total number of collisions. */
    assert(h->v_capacity);
    cgul_string__append_sprintf(cex, s, "total collisions: %lu\n",
                                (unsigned long int)h->total_collisions);

    /* Success. */
    rv = cgul_string__take_value(cex, s);

 out:

    if (s) {
        cgul_string__delete(s);
    }

    return rv;
}
