#include "cgul_heap_primitives.h"

void
cgul_heap_primitives__heapify(cgul_exception_t* cex,
                              void** v,
                              size_t v_count,
                              cgul_heap_primitives__compare_t cf)
{
    size_t i = 0;
    size_t index = 0;
    
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(v);

    /* Create the heap bottom-up from subheaps.  We know that all the
     * leaf nodes are valid subheaps (because they have only one
     * element) so we start with the last node that has children and
     * work our way back to the front of the array.  Because we are
     * dealing with unsigned type and iterating in reverse, i is
     * one-beyond the index. */
    for (i = (v_count >> 1) ; i > 0 ; --i) {

        /* Determine the index. */
        index = i - 1;

        /* Shift down.  Do *not* use Floyd's Method here because
         * shift_up() assumes the heap is fully constructed. */
        cgul_heap_primitives__shift_down(cex,
                                         v,
                                         v_count,  /* size */
                                         index,    /* index */
                                         0,        /* use_floyds_method */
                                         cf);
    }

 out:
    return;
}

void
cgul_heap_primitives__shift_down(cgul_exception_t* cex,
                                 void** v,
                                 size_t v_count,
                                 size_t index,
                                 int use_floyds_method,
                                 cgul_heap_primitives__compare_t cf)
{
    int compare_rv = 0;
    void* tmp = NULL;
    size_t left_child_index = 0;
    size_t right_child_index = 0;
    size_t larger_child_index = 0;
    
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(v);
    assert(v_count > 0);
    assert(index < v_count);

    while (index < v_count) {

        /* We do not yet know which child is larger yet.  (Note that
         * v[0] is the root node so it cannot be a child of any node.) */
        larger_child_index = 0;

        /* Get the index for the left and right children (if any). */
        left_child_index = (index << 1) + 1;
        right_child_index = left_child_index + 1;
        
        /* Determine which child is larger. */
        if (right_child_index < v_count) {
            compare_rv = cf(v[left_child_index], v[right_child_index]);
            if (compare_rv >= 0) {
                larger_child_index = left_child_index;
            } else {
                larger_child_index = right_child_index;
            }
        } else if (left_child_index < v_count) {
            larger_child_index = left_child_index;
        }

        /* If there are no children, this is a leaf node, and we are done. */
        if (larger_child_index == 0) {
            break;
        }

        /* If we are not using Floyd's Method, we have to do a real
         * comparison of v[index] with the larger child in order to
         * maintain heap order. */
        if (!use_floyds_method) {
            /* Compare v[index] with its larger child. */
            compare_rv = cf(v[index], v[larger_child_index]);
            if (compare_rv >= 0) {
                /* v[index] is larger than both children so we are done. */
                break;
            }
        }

        /* Swap v[index] with the larger child. */
        tmp = v[index];
        v[index] = v[larger_child_index];
        v[larger_child_index] = tmp;

        /* Update index and continue. */
        index = larger_child_index;
    }

    /* If using Floyd's Method, v[index] has reached a leaf node, but
     * it has done so by assuming v[index] is smaller than all the
     * values along the path taken.  Now, we have to do a real shift
     * up of v[index] to make sure heap order is maintained.  This
     * improves performance in cases where v[index] was pulled from
     * the back of the heap to replace the root value at the front of
     * the heap because it is likely in this case that v[index] is in
     * fact smaller than all the values along the path allowing us to
     * often replace O(log n) comparisons with O(1) comparisons. */
    if (use_floyds_method) {
        cgul_heap_primitives__shift_up(cex, v, v_count, index, cf);
    }

 out:
    return;
}

void
cgul_heap_primitives__shift_up(cgul_exception_t* cex,
                               void** v,
                               size_t v_count,
                               size_t index,
                               cgul_heap_primitives__compare_t cf)
{
    int compare_rv = 0;
    void* tmp = NULL;
    size_t parent_index = 0;
    
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(v);
    assert(v_count > 0);
    assert(index < v_count);

    while ((index > 0) && (index < v_count)) {

        /* Get the index for the parent. */
        parent_index = ((index + 1) >> 1) - 1;
        
        /* Compare v[index] with its parent. */
        compare_rv = cf(v[index], v[parent_index]);
        if (compare_rv <= 0) {
            /* v[index] is less than or equal to its parent so we are done. */
            break;
        }

        /* Swap v[index] with the parent. */
        tmp = v[index];
        v[index] = v[parent_index];
        v[parent_index] = tmp;

        /* Update index and continue. */
        index = parent_index;
    }

 out:
    return;
}
