#include "cgul_hstring_p.h"
#include "cgul_defensive.h"
#include "cgul_libc.h"
#include "cgul_unicode.h"

/* The default capacity for each newly allocated string.  This is the
 * number of characters _not_ including the NUL terminator.  When
 * realloc() is called one is added to the capacity to make room for
 * the NUL terminator. */
static const size_t CGUL_HSTRING__DEFAULT_MINIMUM_CAPACITY = 128;

/* Comments are in cgul_hstring_p.h. */
void
cgul_hstring__set_capacity(cgul_exception_t* cex,
                           cgul_hstring_t hs,
                           size_t capacity)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(hs);

    /* Sanity check. */
    if (capacity == 0) {
        capacity = hs->minimum_capacity;
    }

    /* Make sure the capacity is at least large enough to hold the
     * current number of elements plus 1 for the NUL terminator. */
    capacity = CGUL_MAX(capacity, hs->length + 1);

#if 0
    fprintf(stderr, "cgul_hstring__set_capacity = %lu\n", capacity);
    fflush(stderr);
#endif

    /* realloc */
    hs->value = (cgul_hchar_t*)cgul_libc__realloc(
                    cex,
                    hs->value,
                    sizeof(*hs->value) * capacity);
    if (*cex) {
        goto out;
    }

    /* Rememeber the new capacity. */
    hs->capacity = capacity;

    /* Clear the extra elements. */
    if (capacity > hs->length) {
        memset(hs->value + hs->length,
               0,
               sizeof(*hs->value) * (hs->capacity - hs->length));
    }

 out:
    return;
}

/* Comments are in cgul_hstring_p.h. */
void
cgul_hstring__decrease_capacity_if_necessary(cgul_exception_t* cex,
                                             cgul_hstring_t hs)
{
    size_t actual_new_capacity = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* We only have excess capacity when 1/4 capacity is greater than
     * or equal to the size.  So, do nothing when 1/4 capacity is less
     * than the size. */
    if ((hs->capacity >> 2) < (hs->length + 1)) {
        goto out;
    }

    /* Keep halving the capacity to get the best fit.  It should be at
     * least double the size to avoid thrashing the allocator.  We
     * know from the if() above that we can start iterating with 1/2
     * of hs->capacity. */
    actual_new_capacity = hs->capacity >> 1;
    while (   ((actual_new_capacity >> 1) > 0)
           && ((actual_new_capacity >> 1) >= ((hs->length + 1) << 1)))
    {
        actual_new_capacity >>= 1;
    }

    /* The actual new capacity should never drop below the minimum
     * capacity. */
    actual_new_capacity = CGUL_MAX(actual_new_capacity, hs->minimum_capacity);
    assert(actual_new_capacity > 0);

    /* If the capacity is already less than or equal to the actual new
     * capacity, do nothing. */
    if (hs->capacity <= actual_new_capacity) {
        goto out;
    }

    /* Sanity check to make sure we are not going to accidentally
     * reduce the capacity below the necessary size. */
    if (actual_new_capacity <= (hs->length + 1)) {
        goto out;
    }

    /* Set the reduced capacity. */
    cgul_hstring__set_capacity(cex, hs, actual_new_capacity);
    if (*cex) {
        goto out;
    }

 out:
    return;
}

/**
 * Wrapper around <code>cgul_hstring__increase_capacity_if_necessary()</code>
 * that performs a quick check to determine if there is already enough
 * excess capacity to accommodate the extra size.  If not, the real
 * method is called.
 *
 * @param[in,out] cex c-style exception
 * @param[in] hs <code>cgul_hstring</code> instance
 * @param[in] extra_size extra size in characters excluding NUL-terminator
 */
#define CGUL_HSTRING__INCREASE_CAPACITY_IF_NECESSARY(cex, hs, extra_size) \
do {                                                                    \
    size_t excess_capacity = (hs)->capacity - (hs)->length - 1;         \
    size_t local_extra_size = (extra_size);                             \
    if (excess_capacity < local_extra_size) {                           \
        cgul_hstring__increase_capacity_if_necessary(                   \
            (cex), (hs), local_extra_size);                             \
    }                                                                   \
} while (0)

/* Comments are in cgul_hstring_p.h. */
void
cgul_hstring__increase_capacity_if_necessary(cgul_exception_t* cex,
                                             cgul_hstring_t hs,
                                             size_t extra_size)
{
    size_t needed_capacity = 0;
    size_t target_capacity = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(hs);

    /* Sanity check. */
    if (extra_size == 0) {
        goto out;
    }

    /*
     * NOTE:
     * ====
     *
     * It is more efficient (on average) to call this method
     * through the CGUL_HSTRING__INCREASE_CAPACITY_IF_NECESSARY() macro
     * defined above.
     */

    /* Add 1 to extra_size to account for the NUL terminator. */
    extra_size = cgul_defensive__add_size_type(cex, extra_size, 1);
    if (*cex) {
        goto out;
    }

    /* Calculate the needed capacity. */
    needed_capacity =
        cgul_defensive__add_size_type(cex, hs->length, extra_size);
    if (*cex) {
        goto out;
    }

    /* If we need more capacity, realloc() it.*/
    if (needed_capacity > hs->capacity) {

        /* Try to double the capacity to allow room to grow.  The test
         * against zero is used to detect overflow. */
        target_capacity = hs->capacity;
        while (   ((target_capacity << 1) > 0)
               && ((target_capacity << 1) <= (needed_capacity << 1)))
        {
            target_capacity <<= 1;
        }

        /* If there is overflow, just use needed_capacity as-is which
         * doesn't allow for room to grow without reallocating but is
         * large enough to hold the new elements.  Also, we know that
         * needed_capacity did not overflow when calculated above
         * because an exception would have been thrown. */
        if (target_capacity < needed_capacity) {
            target_capacity = needed_capacity;
        }

        /* Set the new capacity. */
        cgul_hstring__set_capacity(cex, hs, target_capacity);
    }

 out:
    return;
}

/**
 * Initialize <code>hs</code>.
 *
 * @param[in,out] cex c-style exception
 * @param[in] hs cgul wide-character string
 * @param[in] value initial value
 */
static void
cgul_hstring__initialize(cgul_exception_t* cex,
                         cgul_hstring_t hs,
                         const cgul_hchar_t* value)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(hs);

    /* Free the old underlying array if any. */
    if (hs->value) {
        free(hs->value);
    }

    /* Initialize all member variables. */
    hs->value = NULL;
    hs->length = 0;
    hs->capacity = 0;
    hs->minimum_capacity = CGUL_HSTRING__DEFAULT_MINIMUM_CAPACITY;

    /* Initialize the capacity. */
    cgul_hstring__set_capacity(cex, hs, hs->minimum_capacity);
    if (*cex) {
        goto out;
    }

    /* Now that it is safe to call methods, use set_value() to set the value. */
    cgul_hstring__set_value(cex, hs, value);

 out:
    return;
}

cgul_hstring_t
cgul_hstring__new(cgul_exception_t* cex)
{
    cgul_hstring_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    rv = (cgul_hstring_t)cgul_libc__calloc(cex, 1, sizeof(*rv));
    if (*cex) {
        goto out;
    }

    cgul_hstring__initialize(cex, rv, CGUL_HCHAR__EMPTY_STRING);
    if (*cex) {
        goto out;
    }

 out:

    /* Clean up partially constructed objects. */
    if (*cex && rv) {
        cgul_hstring__delete(rv);
        rv = NULL;
    }

    return rv;
}

cgul_hstring_t
cgul_hstring__new_from_hcstring(cgul_exception_t* cex,
                                const cgul_hchar_t* value)
{
    cgul_hstring_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(value);

    /* Create an empty string object. */
    rv = cgul_hstring__new(cex);
    if (*cex) {
        goto out;
    }

    /* Set the value. */
    cgul_hstring__set_value(cex, rv, value);
    if (*cex) {
        goto out;
    }

 out:

    /* Clean up partially constructed objects. */
    if (*cex && rv) {
        cgul_hstring__delete(rv);
        rv = NULL;
    }

    return rv;
}

void
cgul_hstring__delete(cgul_hstring_t hs)
{
    if (hs) {
        if (hs->value) {
            free(hs->value);
        }
        free(hs);
    }
}

cgul_hchar_t*
cgul_hstring__get_value(cgul_exception_t* cex,
                        cgul_hstring_t hs)
{
    cgul_hchar_t* rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(hs);

    rv = hs->value;

 out:
    return rv;
}

cgul_hchar_t*
cgul_hstring__take_value(cgul_exception_t* cex, cgul_hstring_t hs)
{
    cgul_hchar_t* rv= NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(hs);
    assert(hs->value);

    /* The client is taking ownership of hs->value. */
    rv = hs->value;
    hs->value = NULL;

    /* Reset this object. */
    cgul_hstring__initialize(cex, hs, CGUL_HCHAR__EMPTY_STRING);

 out:
    return rv;
}

void
cgul_hstring__set_value(cgul_exception_t* cex,
                        cgul_hstring_t hs,
                        const cgul_hchar_t* value)
{
    int is_interior_ptr = 0;
    size_t i = 0;
    size_t value_length = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(hs);
    assert(hs->value);
    assert(value);

    /* Do not copy the string to itself. */
    if (hs->value == value) {
        goto out;
    }

    /* Determine if "value" is a pointer to a substring inside of hs->value. */
    is_interior_ptr = (   (value >= hs->value)
                       && (value <= hs->value + hs->length));

    /* Make sure the underlying buffer is large enough but only if
     * "value" is not an interior pointer.  Without this extra check,
     * the realloc() call could invalidate "value", and in this case
     * there is no reason to call realloc() because we know hs->value
     * is large enough to hold any of its own substrings. */
    if (!is_interior_ptr) {
        value_length = cgul_hchar__hcslen(cex, value);
        CGUL_HSTRING__INCREASE_CAPACITY_IF_NECESSARY(cex, hs, value_length);
        if (*cex) {
            goto out;
        }
    }

    /* Copy "value" to "hs->value" including the NUL terminator. */
    for (i = 0 ; i <= value_length ; ++i) {
        hs->value[i] = value[i];
    }
    hs->length = value_length;

    /* See if the capacity can be decreased.  (Note that we *increase*
     * the capacity if necessary before attempting to copy, but
     * sometimes we are setting the string to a shorter value.) */
    cgul_hstring__decrease_capacity_if_necessary(cex, hs);
    if (*cex) {
        goto out;
    }

 out:
    return;
}

cgul_hchar_t
cgul_hstring__get_value_at(cgul_exception_t* cex,
                           cgul_hstring_t hs,
                           size_t index)
{
    cgul_hchar_t rv = CGUL_HCHAR__NUL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(hs);

    if (index >= hs->length) {
        *cex = cgul_exception__new("cgul_hstring__get_value_at: "
                                   "index out of bounds: %lu",
                                   (unsigned long)index);
        goto out;
    }

    rv = hs->value[index];

 out:
    return rv;
}

extern void
cgul_hstring__set_value_at(cgul_exception_t* cex,
                           cgul_hstring_t hs,
                           size_t index,
                           cgul_hchar_t hc)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(hs);

    /* Verify that "index" is within bounds. */
    if (index >= hs->length) {
        *cex = cgul_exception__new("cgul_hstring__set_value_at: "
                                   "index out of bounds: %lu",
                                   (unsigned long)index);
        goto out;
    }

    hs->value[index] = hc;

 out:
    return;
}

/* This method is part of the "cacheable" interface defined by
 * "cgul_cache".  It is required to not throw any exception. */
void
cgul_hstring__clear(cgul_hstring_t hs)
{
    cgul_exception_t local = NULL;
    cgul_hstring__set_length(&local, hs, 0);
}

signed char
cgul_hstring__to_schar(cgul_exception_t* cex,
                       cgul_hstring_t hs,
                       int base)
{
    signed char rv = 0;
    cgul_string_t tmp = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(hs);

    /* Allocate. */
    tmp = cgul_string__new(cex);
    if (*cex) {
        goto out;
    }

    /* Convert to UTF-8. */
    cgul_unicode__hcstombs(cex, hs->value, hs->length, /* skip_bom */ 1, tmp);
    if (*cex) {
        goto out;
    }

    /* Do the numberic conversion. */
    rv = cgul_string__to_schar(cex, tmp, base);
    if (*cex) {
        goto out;
    }

 out:

    /* Clean up. */
    if (tmp) {
        cgul_string__delete(tmp);
    }

    return rv;
}

unsigned char
cgul_hstring__to_uchar(cgul_exception_t* cex,
                       cgul_hstring_t hs,
                       int base)
{
    unsigned char rv = 0;
    cgul_string_t tmp = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(hs);

    /* Allocate. */
    tmp = cgul_string__new(cex);
    if (*cex) {
        goto out;
    }

    /* Convert to UTF-8. */
    cgul_unicode__hcstombs(cex, hs->value, hs->length, /* skip_bom */ 1, tmp);
    if (*cex) {
        goto out;
    }

    /* Do the numberic conversion. */
    rv = cgul_string__to_uchar(cex, tmp, base);
    if (*cex) {
        goto out;
    }

 out:

    /* Clean up. */
    if (tmp) {
        cgul_string__delete(tmp);
    }

    return rv;
}

short int
cgul_hstring__to_short(cgul_exception_t* cex,
                       cgul_hstring_t hs,
                       int base)
{
    short int rv = 0;
    cgul_string_t tmp = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(hs);

    /* Allocate. */
    tmp = cgul_string__new(cex);
    if (*cex) {
        goto out;
    }

    /* Convert to UTF-8. */
    cgul_unicode__hcstombs(cex, hs->value, hs->length, /* skip_bom */ 1, tmp);
    if (*cex) {
        goto out;
    }

    /* Do the numberic conversion. */
    rv = cgul_string__to_short(cex, tmp, base);
    if (*cex) {
        goto out;
    }

 out:

    /* Clean up. */
    if (tmp) {
        cgul_string__delete(tmp);
    }

    return rv;
}

unsigned short int
cgul_hstring__to_ushort(cgul_exception_t* cex,
                        cgul_hstring_t hs,
                        int base)
{
    unsigned short int rv = 0;
    cgul_string_t tmp = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(hs);

    /* Allocate. */
    tmp = cgul_string__new(cex);
    if (*cex) {
        goto out;
    }

    /* Convert to UTF-8. */
    cgul_unicode__hcstombs(cex, hs->value, hs->length, /* skip_bom */ 1, tmp);
    if (*cex) {
        goto out;
    }

    /* Do the numberic conversion. */
    rv = cgul_string__to_ushort(cex, tmp, base);
    if (*cex) {
        goto out;
    }

 out:

    /* Clean up. */
    if (tmp) {
        cgul_string__delete(tmp);
    }

    return rv;
}

int
cgul_hstring__to_int(cgul_exception_t* cex,
                     cgul_hstring_t hs,
                     int base)
{
    int rv = 0;
    cgul_string_t tmp = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(hs);

    /* Allocate. */
    tmp = cgul_string__new(cex);
    if (*cex) {
        goto out;
    }

    /* Convert to UTF-8. */
    cgul_unicode__hcstombs(cex, hs->value, hs->length, /* skip_bom */ 1, tmp);
    if (*cex) {
        goto out;
    }

    /* Do the numberic conversion. */
    rv = cgul_string__to_int(cex, tmp, base);
    if (*cex) {
        goto out;
    }

 out:

    /* Clean up. */
    if (tmp) {
        cgul_string__delete(tmp);
    }

    return rv;
}

unsigned int
cgul_hstring__to_unsigned(cgul_exception_t* cex,
                          cgul_hstring_t hs,
                          int base)
{
    unsigned int rv = 0;
    cgul_string_t tmp = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(hs);

    /* Allocate. */
    tmp = cgul_string__new(cex);
    if (*cex) {
        goto out;
    }

    /* Convert to UTF-8. */
    cgul_unicode__hcstombs(cex, hs->value, hs->length, /* skip_bom */ 1, tmp);
    if (*cex) {
        goto out;
    }

    /* Do the numberic conversion. */
    rv = cgul_string__to_unsigned(cex, tmp, base);
    if (*cex) {
        goto out;
    }

 out:

    /* Clean up. */
    if (tmp) {
        cgul_string__delete(tmp);
    }

    return rv;
}

long int
cgul_hstring__to_long(cgul_exception_t* cex,
                      cgul_hstring_t hs,
                      int base)
{
    long int rv = 0;
    cgul_string_t tmp = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(hs);

    /* Allocate. */
    tmp = cgul_string__new(cex);
    if (*cex) {
        goto out;
    }

    /* Convert to UTF-8. */
    cgul_unicode__hcstombs(cex, hs->value, hs->length, /* skip_bom */ 1, tmp);
    if (*cex) {
        goto out;
    }

    /* Do the numberic conversion. */
    rv = cgul_string__to_long(cex, tmp, base);
    if (*cex) {
        goto out;
    }

 out:

    /* Clean up. */
    if (tmp) {
        cgul_string__delete(tmp);
    }

    return rv;
}

unsigned long int
cgul_hstring__to_unsigned_long(cgul_exception_t* cex,
                               cgul_hstring_t hs,
                               int base)
{
    unsigned long int rv = 0;
    cgul_string_t tmp = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(hs);

    /* Allocate. */
    tmp = cgul_string__new(cex);
    if (*cex) {
        goto out;
    }

    /* Convert to UTF-8. */
    cgul_unicode__hcstombs(cex, hs->value, hs->length, /* skip_bom */ 1, tmp);
    if (*cex) {
        goto out;
    }

    /* Do the numberic conversion. */
    rv = cgul_string__to_unsigned_long(cex, tmp, base);
    if (*cex) {
        goto out;
    }

 out:

    /* Clean up. */
    if (tmp) {
        cgul_string__delete(tmp);
    }

    return rv;
}

double
cgul_hstring__to_double(cgul_exception_t* cex,
                        cgul_hstring_t hs)
{
    double rv = 0;
    cgul_string_t tmp = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(hs);

    /* Allocate. */
    tmp = cgul_string__new(cex);
    if (*cex) {
        goto out;
    }

    /* Convert to UTF-8. */
    cgul_unicode__hcstombs(cex, hs->value, hs->length, /* skip_bom */ 1, tmp);
    if (*cex) {
        goto out;
    }

    /* Do the numberic conversion. */
    rv = cgul_string__to_double(cex, tmp);
    if (*cex) {
        goto out;
    }

 out:

    /* Clean up. */
    if (tmp) {
        cgul_string__delete(tmp);
    }

    return rv;
}

void
cgul_hstring__append(cgul_exception_t* cex,
                     cgul_hstring_t hs,
                     const cgul_hchar_t* value)
{
    size_t value_length = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(hs);
    assert(value);

    /* Calculate the string length of "value". */
    value_length = cgul_hchar__hcslen(cex, value);

    /* Add capacity if necessary. */
    CGUL_HSTRING__INCREASE_CAPACITY_IF_NECESSARY(cex, hs, value_length);
    if (*cex) {
        goto out;
    }

    /* Because we have already calculated the length of "value", we
     * can use memcpy(). */
    memcpy(&hs->value[hs->length], value, value_length * sizeof(cgul_hchar_t));

    /* Calculate the new length. */
    hs->length += value_length;

    /* Null-terminate the string. */
    hs->value[hs->length] = CGUL_HCHAR__NUL;

 out:
    return;
}

void
cgul_hstring__append_block(cgul_exception_t* cex,
                           cgul_hstring_t hs,
                           const void* block,
                           size_t block_size)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(hs);
    assert(hs->value);
    assert(block);

    CGUL_HSTRING__INCREASE_CAPACITY_IF_NECESSARY(cex, hs, block_size);
    if (*cex) {
        goto out;
    }

    memcpy(&hs->value[hs->length], block, block_size * sizeof(cgul_hchar_t));
    hs->length += block_size;
    hs->value[hs->length] = CGUL_HCHAR__NUL;

 out:
    return;
}

void
cgul_hstring__append_char(cgul_exception_t* cex,
                          cgul_hstring_t hs,
                          cgul_hchar_t c)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(hs);
    assert(hs->value);

    CGUL_HSTRING__INCREASE_CAPACITY_IF_NECESSARY(cex, hs, 1);
    if (*cex) {
        goto out;
    }

    hs->value[hs->length] = c;
    ++hs->length;
    hs->value[hs->length] = CGUL_HCHAR__NUL;

 out:
    return;
}

void
cgul_hstring__insert(cgul_exception_t* cex,
                     cgul_hstring_t hs,
                     size_t index,
                     const cgul_hchar_t* value,
                     size_t value_length)
{
    size_t i = 0;
    size_t i_max = 0;
    size_t src = 0;
    size_t dst = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(hs);
    assert(value);

    /* Verify that "index" is within bounds. */
    if (index > hs->length) {
        *cex = cgul_exception__new("cgul_hstring__insert: "
                                   "index out of bounds: %lu > %lu",
                                   (unsigned long)index,
                                   (unsigned long)hs->length);
        goto out;
    }

    /* Add extra capacity if necessary to accommodate the extra characters. */
    CGUL_HSTRING__INCREASE_CAPACITY_IF_NECESSARY(cex, hs, value_length);
    if (*cex) {
        goto out;
    }

    /* Shift the characters after and including the insertion point
     * down to the new end of the hstring.  Notice that this includes
     * shifting the NUL terminator. */
    src = hs->length;
    dst = hs->length + value_length;
    i_max = hs->length - index;
    for (i = 0 ; i <= i_max ; ++i, --src, --dst) {
        hs->value[dst] = hs->value[src];
    }

    /* Insert the characters from "value". */
    for (i = 0 ; i < value_length ; ++i) {
        hs->value[index + i] = value[i];
    }

    /* Calculate the new length. */
    hs->length += value_length;

    /* Null-terminate the wide-character string. */
    hs->value[hs->length] = CGUL_HCHAR__NUL;

 out:
    return;
}

void
cgul_hstring__insert_char(cgul_exception_t* cex,
                          cgul_hstring_t hs,
                          size_t index,
                          cgul_hchar_t value)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(hs);
    assert(value);

    /* Have cgul_hstring__insert() do the work. */
    cgul_hstring__insert(cex, hs, index, &value, 1);

 out:
    return;
}

void
cgul_hstring__pad_right(cgul_exception_t* cex,
                        cgul_hstring_t hs,
                        size_t width,
                        cgul_hchar_t pad)
{
    size_t i = 0;
    size_t j = 0;
    size_t pad_size = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(hs);
    assert(hs->value);

    if (hs->length < width) {

        /* Calculate the size of the padding. */
        pad_size = width - hs->length;

        /* Add capacity if necessary. */
        CGUL_HSTRING__INCREASE_CAPACITY_IF_NECESSARY(cex, hs, pad_size);
        if (*cex) {
            goto out;
        }

        /* Shift the string right. */
        for (i = hs->length, j = width ; i > 0 ; --i, --j) {
            hs->value[j - 1] = hs->value[i - 1];
        }

        /* Insert the padding and adjust hs->length along the way. */
        for (i = 0 ; i < pad_size ; ++i) {
            hs->value[i] = pad;
        }
        hs->length = width;
        hs->value[hs->length] = CGUL_HCHAR__NUL;
    }

 out:
    return;
}

void
cgul_hstring__pad_left(cgul_exception_t* cex,
                       cgul_hstring_t hs,
                       size_t width,
                       cgul_hchar_t pad)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(hs);
    assert(hs->value);

    if (hs->length < width) {
        size_t i = 0;
        size_t pad_size = width - hs->length;

        CGUL_HSTRING__INCREASE_CAPACITY_IF_NECESSARY(cex, hs, pad_size);
        if (*cex) {
            goto out;
        }

        /* Insert the padding and adjust hs->length along the way. */
        for ( i = 0 ; i < pad_size ; ++i ) {
            hs->value[hs->length++] = pad;
        }
        hs->value[hs->length] = CGUL_HCHAR__NUL;
    }

 out:
    return;
}

void
cgul_hstring__trim(cgul_exception_t* cex,
                   cgul_hstring_t hs)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(hs);

    /* Trim the string.  Because padding right does not require a
     * memmove(), it is done first in the hope that it will reduce the
     * amount of work memmove() has to do for trim_left(). */
    cgul_hstring__trim_right(cex, hs);
    cgul_hstring__trim_left(cex, hs);

 out:
    return;
}

void
cgul_hstring__trim_left(cgul_exception_t* cex,
                        cgul_hstring_t hs)
{
    size_t i = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(hs);

    /* Find the first character in hs that is not white space. */
    for (i = 0 ; i < hs->length ; ++i) {
        if (!cgul_hchar__isspace(cex, hs->value[i])) {
            break;
        }
    }

    /* Nothing to do if there is no leading white space. */
    if (i == 0) {
        goto out;
    }

    /* Trim the string. */
    memmove(hs->value, &hs->value[i], (hs->length - i) * sizeof(cgul_hchar_t));
    hs->length -= i;
    hs->value[hs->length] = CGUL_HCHAR__NUL;

    /* Decrease capacity if necessary. */
    cgul_hstring__decrease_capacity_if_necessary(cex, hs);

 out:
    return;
}

void
cgul_hstring__trim_right(cgul_exception_t* cex,
                         cgul_hstring_t hs)
{
    size_t i = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(hs);

    /* Iterate in reverse order over the characters in hs.  Note that
     * "i" is always one beyond the character we are checking making
     * it easier to test to make sure we do not read off the front of
     * the string. */
    for (i = hs->length ; i > 0 ; --i) {
        if (!cgul_hchar__isspace(cex, hs->value[i - 1])) {
            break;
        }
    }

    /* Nothing to do if there is no trailing white space. */
    if (i == hs->length) {
        goto out;
    }

    /* Trim the string. */
    hs->length = i;
    hs->value[hs->length] = CGUL_HCHAR__NUL;

    /* Decrease capacity if necessary. */
    cgul_hstring__decrease_capacity_if_necessary(cex, hs);

 out:
    return;
}

void
cgul_hstring__reverse(cgul_exception_t* cex,
                      cgul_hstring_t hs)
{
    size_t i = 0;
    size_t j = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(hs);
    assert(hs->value);

    /* If the string is empty or has only one charcter, there is
     * nothing to reverse. */
    if (hs->length <= 1) {
        goto out;
    }

    for (j = hs->length - 1 ; i < j ; ++i, --j) {
        cgul_hchar_t tmp = hs->value[i];
        hs->value[i] = hs->value[j];
        hs->value[j] = tmp;
    }

 out:
    return;
}

size_t
cgul_hstring__get_length(cgul_exception_t* cex,
                         cgul_hstring_t hs)
{
    size_t rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(hs);

    rv = hs->length;

 out:
    return rv;
}

size_t
cgul_hstring__set_length(cgul_exception_t* cex,
                         cgul_hstring_t hs,
                         size_t length)
{
    size_t rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(hs);

    /* Throw an exception if the user tries to increase the length. */
    if (length > hs->length) {
        *cex = cgul_exception__new("cgul_hstring__set_length: "
                                   "attempt to increase length");
        goto out;
    }

    /* Return the previous length. */
    rv = hs->length;

    /* Truncate value. */
    hs->value[length] = CGUL_HCHAR__NUL;

    /* Set the new length. */
    hs->length = length;

    /* Decrease capacity if necessary. */
    cgul_hstring__decrease_capacity_if_necessary(cex, hs);

 out:
    return rv;
}

void
cgul_hstring__externally_adjusted(cgul_exception_t* cex,
                                  cgul_hstring_t hs)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(hs);
    assert(hs->value);

    hs->length = cgul_hchar__hcslen(cex, hs->value);
    cgul_hstring__decrease_capacity_if_necessary(cex, hs);

 out:
    return;
}

void
cgul_hstring__trim_capacity(cgul_exception_t* cex,
                            cgul_hstring_t hs)
{
    if (*cex) {
        goto out;
    }

    assert(hs);

    /* Use realloc() to trim the buffer. */
    hs->value =
        (cgul_hchar_t*)cgul_libc__realloc(
                           cex,
                           hs->value,
                           sizeof(cgul_hchar_t) * (hs->length + 1));
    if (*cex) {
        goto out;
    }

    /* Remember the new capacity. */
    hs->capacity = hs->length + 1;

 out:
    return;
}

size_t
cgul_hstring__get_minimum_capacity(cgul_exception_t* cex,
                                   cgul_hstring_t hs)
{
    size_t rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(hs);

    /* Return the minimum capacity to the user. */
    rv = hs->minimum_capacity;

 out:
    return rv;
}

void
cgul_hstring__set_minimum_capacity(cgul_exception_t* cex,
                                   cgul_hstring_t hs,
                                   size_t minimum_capacity)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(hs);

    /* Set the new minimum_capacity. */
    if (minimum_capacity == 0) {
        hs->minimum_capacity = CGUL_HSTRING__DEFAULT_MINIMUM_CAPACITY;
    } else {
        hs->minimum_capacity = minimum_capacity;
    }

    /* Realloc the underlying buffer (if necessary). */
    cgul_hstring__set_capacity(cex, hs, hs->minimum_capacity);
    if (*cex) {
        goto out;
    }

 out:
    return;
}

int
cgul_hstring__compare(cgul_exception_t* cex,
                      cgul_hstring_t hs1,
                      cgul_hstring_t hs2)
{
    int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(hs1);
    assert(hs2);

    /* Compare the strings. */
    rv = cgul_hchar__hcscmp(hs1->value, hs2->value);

 out:
    return rv;
}

void
cgul_hstring__swap(cgul_exception_t* cex,
                   cgul_hstring_t hs1,
                   cgul_hstring_t hs2)
{
    struct cgul_hstring tmp;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(hs1);
    assert(hs2);

    /* Nothing to swap if they point to the same objects. */
    if (hs1 == hs2) {
        goto out;
    }

    memcpy((void*)&tmp, (void*)hs1,  sizeof(tmp));
    memcpy((void*)hs1,  (void*)hs2,  sizeof(tmp));
    memcpy((void*)hs2,  (void*)&tmp, sizeof(tmp));

 out:
    return;
}
