#include "cgul_libc_p.h"
#include "cgul_defensive.h"
#include "cgul_exception.h"
#include "cgul_string.h"

/* Size of the buffer used to copy bytes from one FILE to another. */
#define CGUL_LIBC__COPY_FILE_BUFSIZE (16384)

/**
 * This boolean is used to simulate out-of-memory condition.  When it
 * is set, memory allocation is skipped causing an allocation failure
 * to be simulated.
 */
static int cgul_libc__out_of_memory = 0;

void
cgul_libc__set_out_of_memory(cgul_exception_t* cex,
                             int out_of_memory)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    cgul_libc__out_of_memory = out_of_memory;

 out:
    return;
}


void*
cgul_libc__calloc(cgul_exception_t* cex,
                  size_t element_count,
                  size_t element_size)
{
    void* rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    if (!cgul_libc__out_of_memory) {
        rv = calloc(element_count, element_size);
    }

    if (rv == NULL) {
        *cex = cgul_exception__get_out_of_memory_exception();
        goto out;
    }

 out:
    return rv;
}

void*
cgul_libc__malloc(cgul_exception_t* cex, size_t new_size)
{
    void* rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    if (!cgul_libc__out_of_memory) {
        rv = malloc(new_size);
    }

    if (rv == NULL) {
        *cex = cgul_exception__get_out_of_memory_exception();
        goto out;
    }

 out:
    return rv;
}


void*
cgul_libc__realloc(cgul_exception_t* cex, void* orig_pointer, size_t new_size)
{
    void* rv = orig_pointer;
    void* tmp = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    if (!cgul_libc__out_of_memory) {
        tmp = realloc(orig_pointer, new_size);
    }

    if (tmp == NULL) {
        *cex = cgul_exception__get_out_of_memory_exception();
        goto out;
    }

    rv = tmp;

 out:
    return rv;
}

void*
cgul_libc__memset(cgul_exception_t* cex, void *buf, int c, size_t n)
{
    void* rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(buf);

    rv = memset(buf, c, n);

 out:
    return rv;
}

void*
cgul_libc__memcpy(cgul_exception_t* cex,
                  void* dst,
                  const void* src,
                  size_t n)
{
    void* rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(dst);
    assert(src);

    rv = memcpy(dst, src, n);

 out:
    return rv;
}

void*
cgul_libc__memmove(cgul_exception_t* cex,
                   void* dst,
                   const void* src,
                   size_t n)
{
    void* rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(dst);
    assert(src);

    rv = memmove(dst, src, n);

 out:
    return rv;
}

FILE*
cgul_libc__fopen(cgul_exception_t* cex,
                 const char* filename,
                 const char* mode)
{
    FILE* rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(filename);
    assert(mode);

    rv = fopen(filename, mode);
    if (rv == NULL) {
        *cex = cgul_exception__new("cgul_libc__fopen: fopen: %s: %s",
                                   filename,
                                   strerror(errno));
        goto out;
    }

 out:
    return rv;
}

FILE*
cgul_libc__freopen(cgul_exception_t* cex,
                   const char* filename,
                   const char* mode,
                   FILE* f)
{
    FILE* rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(filename);
    assert(mode);
    assert(f);

    rv = freopen(filename, mode, f);
    if (rv == NULL) {
        *cex = cgul_exception__new("cgul_libc__freopen: freopen: %s: %s",
                                   filename,
                                   strerror(errno));
    }

 out:
    return rv;
}

void
cgul_libc__fclose(cgul_exception_t* cex, FILE *f)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    if (fclose(f) == EOF) {
        *cex = cgul_exception__new("cgul_libc__fclose: fclose: %s",
                                   strerror(errno));
        goto out;
    }

 out:
    return;
}

void
cgul_libc__fflush(cgul_exception_t* cex,
                  FILE* f)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    if (fflush(f) == EOF) {
        *cex = cgul_exception__new("cgul_libc__fflush: fflush: %s",
                                   strerror(errno));
        goto out;
    }

 out:
    return;
}

size_t
cgul_libc__fread(cgul_exception_t* cex,
                 void* buf,
                 size_t element_size,
                 size_t element_count,
                 FILE* f)
{
    size_t rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(buf);
    assert(f);

    rv = fread(buf, element_size, element_count, f);
    if ((rv == 0) && ferror(f)) {
        *cex = cgul_exception__new("cgul_libc__fread: fread: %s",
                                   strerror(errno));
        goto out;
    }

 out:
    return rv;
}

size_t
cgul_libc__fread_block(cgul_exception_t* cex,
                       void* block,
                       size_t block_size,
                       FILE* fin)
{
    size_t rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(block);
    assert(fin);

    /* Sanity check. */
    if (block_size <= 0) {
        goto out;
    }

    /* Read as much of the block as possible restarting if interrupted. */
    errno = 0;
    while ((rv = fread(block, 1, block_size, fin)) == 0) {

#ifdef EINTR
        /* Restart if interrupted. */
        if (errno == EINTR) {
            clearerr(fin);
            errno = 0;
            continue;
        }
#endif

        /* Check for errors. */
        if (ferror(fin)) {
            *cex = cgul_exception__new("cgul_libc__fread_block: fread: %m");
            goto out;
        }

        /* Success. */
        break;
    }

 out:
    return rv;
}

size_t
cgul_libc__fwrite(cgul_exception_t* cex,
                  const void* buf,
                  size_t element_size,
                  size_t element_count,
                  FILE* f)
{
    size_t rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(buf);
    assert(f);

    rv = fwrite(buf, element_size, element_count, f);
    if ((rv == 0) && ferror(f)) {
        *cex = cgul_exception__new("cgul_libc__fwrite: fwrite: %s",
                                   strerror(errno));
        goto out;
    }

 out:
    return rv;
}

size_t
cgul_libc__fwrite_block(cgul_exception_t* cex,
                        const void* block,
                        size_t block_size,
                        FILE* fout)
{
    size_t rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(block);
    assert(fout);

    /* Sanity check. */
    if (block_size <= 0) {
        goto out;
    }

    /* Write as much of the block as possible restarting if interrupted. */
    errno = 0;
    while ((rv = fwrite(block, 1, block_size, fout)) == 0) {

#ifdef EINTR
        /* Restart if interrupted. */
        if (errno == EINTR) {
            clearerr(fout);
            errno = 0;
            continue;
        }
#endif

        /* Check for errors. */
        if (ferror(fout)) {
            *cex = cgul_exception__new("cgul_libc__fwrite_block: fwrite: %m");
            goto out;
        }

        /* Success. */
        break;
    }

 out:
    return rv;
}

void
cgul_libc__load_file(cgul_exception_t* cex,
                     FILE* fin,
                     char** ptr,
                     size_t* length)
{
    size_t rcount = 0;
    size_t block_size = 16384;
    char* block = NULL;
    cgul_string_t s = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(fin);
    assert(ptr);

    /* Allocte. */
    s = cgul_string__new(cex);
    block = (char*)cgul_libc__malloc(cex, block_size);
    if (*cex) {
        goto out;
    }

    /* Iterate over each block in the file. */
    for (;;) {

        /* Read the next block. */
        rcount = cgul_libc__fread_block(cex, block, block_size, fin);
        if (*cex) {
            goto out;
        }

        /* EOF */
        if (rcount == 0) {
            break;
        }

        /* Copy the block to memory. */
        cgul_string__append_block(cex, s, block, rcount);
        if (*cex) {
            goto out;
        }
    }

    /* Success. */
    if (length) {
        *length = cgul_string__get_length(cex, s);
    }
    if (ptr) {
        *ptr = cgul_string__take_value(cex, s);
    }

 out:

    /* Clean up. */
    if (block) {
        free(block);
    }
    if (s) {
        cgul_string__delete(s);
    }

    return;
}

void
cgul_libc__save_file(cgul_exception_t* cex,
                     FILE* fout,
                     const char* block,
                     size_t block_size)
{
    size_t wcount = 0;
    size_t wtotal = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(block);
    assert(fout);

    while (wtotal < block_size) {

        /* Write as much of the block as possible. */
        wcount = cgul_libc__fwrite_block(cex,
                                         block + wtotal,
                                         block_size - wtotal,
                                         fout);
        if (*cex) {
            goto out;
        }

        wtotal += wcount;
    }

 out:
    return;
}

void
cgul_libc__save_fname(cgul_exception_t* cex,
                      const char* fname,
                      const char* block,
                      size_t block_size)
{
    FILE* fout = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(fname);
    assert(block);

    /* Open the output file. */
    fout = cgul_libc__fopen(cex, fname, "wb");
    if (*cex) {
        goto out;
    }

    /* Save the block to fout. */
    cgul_libc__save_file(cex, fout, block, block_size);
    if (*cex) {
        goto out;
    }

 out:

    /* Clean up. */
    if (fout) {
        fclose(fout);
    }

    return;
}

void
cgul_libc__load_fname(cgul_exception_t* cex,
                      const char* fname,
                      char** ptr,
                      size_t* length)
{
    FILE* fin = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(fname);

    /* Open the file. */
    fin = cgul_libc__fopen(cex, fname, "rb");
    if (*cex) {
        goto out;
    }

    /* Parse the file. */
    cgul_libc__load_file(cex, fin, ptr, length);
    if (*cex) {
        goto out;
    }

 out:

    if (fin) {
        fclose(fin);
    }

    return;
}

void
cgul_libc__copy_file(cgul_exception_t* cex,
                     FILE* fin,
                     FILE* fout)
{
    size_t rcount = 0;
    size_t wcount = 0;
    size_t write_rv = 0;
    char* buf = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(fin);
    assert(fout);

    /* Better than putting a large buffer on the stack. */
    buf = (char*)cgul_libc__malloc(cex, CGUL_LIBC__COPY_FILE_BUFSIZE);
    if (*cex) {
        goto out;
    }

    for (;;) {

        /* Read a block. */
        rcount = cgul_libc__fread_block(cex,
                                        buf,
                                        CGUL_LIBC__COPY_FILE_BUFSIZE,
                                        fin);
        if (*cex) {
            goto out;
        }

        /* EOF */
        if (rcount == 0) {
            break;
        }

        /* Write a block. */
        wcount = 0;
        while (wcount < rcount) {
            write_rv = cgul_libc__fwrite_block(cex,
                                               buf + wcount,
                                               rcount - wcount,
                                               fout);
            if (*cex) {
                goto out;
            }
            wcount += write_rv;
        }
    }

 out:

    /* Clean up. */
    if (buf) {
        free(buf);
    }
}

long
cgul_libc__ftell(cgul_exception_t* cex, FILE* f)
{
    long rv = -1;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(f);

    rv = ftell(f);
    if (rv == -1) {
        *cex = cgul_exception__new("cgul_libc__ftell: ftell: %s",
                                   strerror(errno));
        goto out;
    }

 out:
    return rv;
}

void
cgul_libc__fseek(cgul_exception_t* cex,
                 FILE* f,
                 long offset,
                 int whence)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(f);

    if (fseek(f, offset, whence) == -1) {
        *cex = cgul_exception__new("cgul_libc__fseek: fseek: %s",
                                   strerror(errno));
        goto out;
    }

 out:
    return;
}

void
cgul_libc__rewind(cgul_exception_t* cex, FILE* f)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(f);

    errno = 0;
    rewind(f);
    if (errno) {
        *cex = cgul_exception__new("cgul_libc__rewind: rewind: %s",
                                   strerror(errno));
        goto out;
    }

 out:
    return;
}

void
cgul_libc__fgetpos(cgul_exception_t* cex,
                   FILE* f,
                   fpos_t* pos)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(f);
    assert(pos);

    if (fgetpos(f, pos) == -1) {
        *cex = cgul_exception__new("cgul_libc__fgetpos: fgetpos: %s",
                                   strerror(errno));
        goto out;
    }

 out:
    return;
}

void
cgul_libc__fsetpos(cgul_exception_t* cex,
                   FILE* f,
                   fpos_t* pos)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(f);
    assert(pos);

    if (fsetpos(f, pos) == -1) {
        *cex = cgul_exception__new("cgul_libc__fsetpos: fsetpos: %s",
                                   strerror(errno));
        goto out;
    }

 out:
    return;
}

FILE*
cgul_libc__tmpfile(cgul_exception_t* cex)
{
    FILE* rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    rv = tmpfile();
    if (rv == NULL) {
        *cex = cgul_exception__new("cgul_libc__tmpfile: tmpfile: %s",
                                   strerror(errno));
        goto out;
    }

 out:
    return rv;
}

extern int
cgul_libc__printf(cgul_exception_t* cex,
                  const char* format,
                  ...)
{
    int rv = -1;
    va_list args;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(format);

    va_start(args, format);
    rv = cgul_libc__vprintf(cex, format, args);
    va_end(args);

 out:
    return rv;
}

int
cgul_libc__vprintf(cgul_exception_t* cex,
                   const char* format,
                   va_list args)
{
    int rv = -1;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(format);

    rv = vprintf(format, args);
    if (rv < 0) {
        *cex = cgul_exception__new("cgul_libc__vprintf: vprintf: %s",
                                   strerror(errno));
        rv = -1;
        goto out;
    }

 out:
    return rv;
}

extern int
cgul_libc__fprintf(cgul_exception_t* cex,
                   FILE* f,
                   const char* format,
                   ...)
{
    int rv = -1;
    va_list args;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(f);
    assert(format);

    va_start(args, format);
    rv = cgul_libc__vfprintf(cex, f, format, args);
    va_end(args);

 out:
    return rv;
}

int
cgul_libc__vfprintf(cgul_exception_t* cex,
                    FILE* f,
                    const char* format,
                    va_list args)
{
    int rv = -1;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(f);
    assert(format);

    rv = vfprintf(f, format, args);
    if (rv < 0) {
        *cex = cgul_exception__new("cgul_libc__vfprintf: vfprintf: %s",
                                   strerror(errno));
        rv = -1;
        goto out;
    }

 out:
    return rv;
}

int
cgul_libc__fgetc(cgul_exception_t* cex, FILE* f)
{
    int rv = EOF;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(f);

    rv = fgetc(f);

    if (rv == EOF) {
        if (ferror(f)) {
            *cex = cgul_exception__new("cgul_libc__fgetc: "
                                       "fgetc: %s",
                                       strerror(errno));
            goto out;
        }
    }

 out:
    return rv;
}

void
cgul_libc__fputc(cgul_exception_t* cex,
                 int c,
                 FILE* f)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(f);

    if (fputc(c, f) == EOF) {
        *cex = cgul_exception__new("cgul_libc__fputc: "
                                   "fputc: %s",
                                   strerror(errno));
        goto out;
    }

 out:
    return;
}

char*
cgul_libc__fgets(cgul_exception_t* cex,
                 char* buf,
                 int buf_size,
                 FILE* f)
{
    char* rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(buf);
    assert(f);
    assert(buf_size > 0);

    rv = fgets(buf, buf_size, f);
    if (rv == NULL) {
        if (ferror(f)) {
            *cex = cgul_exception__new("cgul_libc__fgets: fgets: %s",
                                       strerror(errno));
            goto out;
        }
    }

 out:
    return rv;
}

void
cgul_libc__fputs(cgul_exception_t* cex,
                 const char* s,
                 FILE* f)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(s);
    assert(f);

    if (fputs(s, f) == EOF) {
        *cex = cgul_exception__new("cgul_libc__fputs: fputs: %s",
                                   strerror(errno));
        goto out;
    }

 out:
    return;
}

void
cgul_libc__puts(cgul_exception_t* cex,
                 const char* s)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(s);

    if (puts(s) == EOF) {
        *cex = cgul_exception__new("cgul_libc__puts: puts: %s",
                                   strerror(errno));
        goto out;
    }

 out:
    return;
}

size_t
cgul_libc__strlen(cgul_exception_t* cex, const char* s)
{
    size_t rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(s);

    rv = strlen(s);

 out:
    return rv;
}

char*
cgul_libc__strcpy(cgul_exception_t* cex, char* dst, const char* src)
{
    char* rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(dst);
    assert(src);

    rv = strcpy(dst, src);

 out:
    return rv;
}

char*
cgul_libc__strncpy(cgul_exception_t* cex,
                   char* dst,
                   const char* src,
                   size_t dst_size)
{
    char* rv = NULL;
    const char* i = NULL;
    char* j = NULL;
    size_t count = 0;
    size_t count_max = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(dst);
    assert(src);

    rv = dst;

    if (dst_size == 0) {
        goto out;
    }

    count_max = dst_size - 1;

    for (i = src, j = dst ;
         (*i != '\0') && (count < count_max) ;
         ++i, ++j, ++count)
    {
        *j = *i;
    }

    *j = '\0';

 out:
    return rv;
}

char*
cgul_libc__strcat(cgul_exception_t* cex, char* dst, const char *src)
{
    char* rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(dst);
    assert(src);

    rv = strcat(dst, src);

 out:
    return rv;
}

int
cgul_libc__strcmp(cgul_exception_t* cex, const char* s1, const char* s2)
{
    int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(s1);
    assert(s2);

    rv = strcmp(s1, s2);

 out:
    return rv;
}

char*
cgul_libc__strdup(cgul_exception_t* cex, const char* s)
{
    char* rv = NULL;
    size_t len = (size_t)-1;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(s);

    len = cgul_libc__strlen(cex, s);
    rv = (char*)cgul_libc__malloc(cex, len + 1);
    cgul_libc__strcpy(cex, rv, s);

 out:
    return rv;
}

char*
cgul_libc__asprintf(cgul_exception_t* cex,
                    const char* format,
                    ...)
{
    char* rv = NULL;
    va_list args;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(format);

    va_start(args, format);
    rv = cgul_libc__vasprintf(cex, format, args);
    va_end(args);

 out:
    return rv;
}

char*
cgul_libc__vasprintf(cgul_exception_t* cex,
                     const char* format,
                     va_list args)
{
    char* rv = NULL;
    cgul_string_t s = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    s = cgul_string__new(cex);
    cgul_string__append_vsprintf(cex, s, format, args);
    rv = cgul_string__take_value(cex, s);

 out:

    if (s) {
        cgul_string__delete(s);
    }

    return rv;
}

int
cgul_libc__snprintf(cgul_exception_t* cex,
                    char* dst,
                    long dst_size,
                    const char* format,
                    ...)
{
    int rv = -1;
    va_list args;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(dst || (dst_size == 0));
    assert(format);

    va_start(args, format);
    rv = cgul_libc__vsnprintf(cex, dst, dst_size, format, args);
    va_end(args);

 out:
    return rv;
}

int
cgul_libc__vsnprintf(cgul_exception_t* cex,
                     char* dst,
                     long dst_size,
                     const char* format,
                     va_list args)
{
    int rv = -1;
    cgul_string_t s = NULL;
    char* src = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(dst || (dst_size == 0));
    assert(format);

    /* Expand "format" and put the result in "s". */
    s = cgul_string__new(cex);
    rv = cgul_string__append_vsprintf(cex, s, format, args);
    if (*cex) {
        goto out;
    }

    /* Copy no more than "dst_size" - 1 bytes to "dst". */
    if (dst && (dst_size > 0)) {
        src = cgul_string__get_value(cex, s);
        cgul_libc__strncpy(cex, dst, src, dst_size);
    }

 out:

    if (s) {
        cgul_string__delete(s);
    }

    return rv;
}

char*
cgul_libc__basename(cgul_exception_t* cex,
                    const char* name)
{
    char c = '\0';
    size_t i = 0;
    size_t name_length = 0;
    char* rv = NULL;
    cgul_string_t s = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    if (!name) {
        goto out;
    }

    /* Allocate s. */
    s = cgul_string__new(cex);
    if (*cex) {
        goto out;
    }

    /* Get the length of "name". */
    name_length = strlen(name);

    /* Skip the trailing path separators.  For example, the basename
       of "foo/bar//" should be "bar". */
    for (i = name_length ; i > 0 ; --i) {
        c = name[i - 1];
        if ((c != '/') && (c != '\\')) {
            break;
        }
    }

    /* Handle special case where the basename of "/" is "/". */
    if ((name_length > 0) && (i == 0)) {
        if (name[0] == '/') {
            rv = cgul_libc__strdup(cex, "/");
            goto out;
        }
        if (name[0] == '\\') {
            rv = cgul_libc__strdup(cex, "\\");
            goto out;
        }
    }

    /* Iterate in reverse order until we find the next path separator
     * inserting characters into s as we go. */
    for ( ; i > 0 ; --i) {
        c = name[i - 1];
        if ((c == '/') || (c == '\\')) {
            break;
        }
        /* Insert the characters in reverse order. */
        cgul_string__append_char(cex, s, c);
    }

    /* Check for exceptions. */
    if (*cex) {
        goto out;
    }

    /* Reverse the string to compensate for inserting in reverse order. */
    cgul_string__reverse(cex, s);

    /* Success. */
    rv = cgul_string__take_value(cex, s);

 out:

    /* Clean up. */
    if (s) {
        cgul_string__delete(s);
    }

    return rv;
}

char*
cgul_libc__dirname(cgul_exception_t* cex,
                   const char* name)
{
    char c = '\0';
    size_t i = 0;
    size_t name_length = 0;
    char* rv = NULL;
    cgul_string_t s = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    if (!name) {
        goto out;
    }

    /* Get the length of "name". */
    name_length = strlen(name);

    /* Handle special case where the dirname of "/" is "/". */
    if (name_length == 1) {
        if (name[0] == '/') {
            rv = cgul_libc__strdup(cex, "/");
            goto out;
        }
        if (name[0] == '\\') {
            rv = cgul_libc__strdup(cex, "\\");
            goto out;
        }
    }

    /* Handle special case where the dirname of ".." is "." */
    if (name_length == 2) {
        if ((name[0] == '.') && (name[1] == '.')) {
            rv = cgul_libc__strdup(cex, ".");
            goto out;
        }
    }

    /* Skip the trailing path separators.  For example, the dirname
     * of "foo/bar//" should be "foo". */
    for (i = name_length ; i > 0 ; --i) {
        c = name[i - 1];
        if ((c != '/') && (c != '\\')) {
            break;
        }
    }

    /* Handle the special case where the dirname of "//" is "/" and
     * "///" is "/" etc. */
    if (i == 0) {
        if (name[0] == '/') {
            rv = cgul_libc__strdup(cex, "/");
            goto out;
        }
        if (name[0] == '\\') {
            rv = cgul_libc__strdup(cex, "\\");
            goto out;
        }
    }

    /* Skip the basename part of the path. */
    for ( ; i > 0 ; --i) {
        c = name[i - 1];
        if ((c == '/') || (c == '\\')) {
            break;
        }
    }

    /* Handle the special case where "name" is comprised of just a
     * basename.  For example, the dirname of "foo" should be "." */
    if (i == 0) {
        rv = cgul_libc__strdup(cex, ".");
        goto out;
    }

    /* Skip any extra path separators between the basename and the
     * dirname.  For example, the dirname of "/foo//bar" should be
     * "/foo". */
    for ( ; i > 0 ; --i) {
        c = name[i - 1];
        if ((c != '/') && (c != '\\')) {
            break;
        }
    }

    /* Handle the special case of an absolute path where there are no
     * intervening directory components between the absolute path
     * separator and the basename.  For example, the dirname of
     * "////foo" should be "/", but the loop above consumes the "/"
     * characters. */
    if (i == 0) {
        i = 1;
    }

    /* Allocate s. */
    s = cgul_string__new(cex);
    if (*cex) {
        goto out;
    }

    /* What remains is the dirname. */
    for ( ; i > 0 ; --i) {
        cgul_string__append_char(cex, s, name[i - 1]);
    }

    /* Check for exceptions. */
    if (*cex) {
        goto out;
    }

    /* Reverse the string to compensate for inserting in reverse order. */
    cgul_string__reverse(cex, s);

    /* Success. */
    rv = cgul_string__take_value(cex, s);

 out:

    /* Clean up. */
    if (s) {
        cgul_string__delete(s);
    }

    return rv;
}

char*
cgul_libc__extension(cgul_exception_t* cex,
                     const char* name)
{
    char c = '\0';
    size_t i = 0;
    size_t i_stop = 0;
    size_t name_length = 0;
    cgul_string_t s = NULL;
    char* rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    if (!name) {
        goto out;
    }

    /* Allocate. */
    s = cgul_string__new(cex);
    if (*cex) {
        goto out;
    }

    /* Get the length of "name". */
    name_length = strlen(name);

    /* Assume "name" has an extension.  i_stop is the _inclusive_
     * index where copying stops. */
    i_stop = name_length - 1;

    /* Skip trailing path separators.  For example the base name of
     * "foo/bar.d/" is "bar.d"; thus, its extension should be ".d". */
    for (i = name_length ; i > 0 ; --i) {
        /* Get the character. */
        c = name[i - 1];
        /* Break if c is not a path separator. */
        if ((c != '/') && (c != '\\')) {
            i_stop = i - 1;
            break;
        }
    }

    /* Search for the dot, but do not search beyond the path separator. */
    for ( ; i > 0 ; --i) {
        /* Get the character. */
        c = name[i - 1];
        /* Break if c is a dot. */
        if (c == '.') {
            break;
        }
        /* Break if c is a path separator. */
        if ((c == '/') || (c == '\\')) {
            i_stop = i - 1;
            break;
        }
    }

    /* Copy the extension to s. */
    if (c == '.') {
        cgul_string__append_block(cex, s, &name[i - 1], i_stop - (i - 1) + 1);
    }

    /* Return the result. */
    rv = cgul_string__take_value(cex, s);

 out:

    /* Clean up. */
    if (s) {
        cgul_string__delete(s);
    }

    return rv;
}

char*
cgul_libc__rootname(cgul_exception_t* cex,
                    const char* name)
{
    char c = '\0';
    size_t i = 0;
    size_t root_length = 0;
    size_t name_length = 0;
    cgul_string_t s = NULL;
    char* rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    if (!name) {
        goto out;
    }

    /* Allocate. */
    s = cgul_string__new(cex);
    if (*cex) {
        goto out;
    }

    /* Get the length of "name". */
    name_length = strlen(name);

    /* Assume there is no extension. */
    root_length = name_length;

    /* Skip trailing path separators.  For example the base name of
     * "foo/bar.d/" is "bar.d"; thus, its root name should be
     * "foo/bar". */
    for (i = name_length ; i > 0 ; --i) {
        /* Get the character. */
        c = name[i - 1];
        /* Break if c is not a path separator. */
        if ((c != '/') && (c != '\\')) {
            root_length = i;
            break;
        }
    }

    /* Search for the dot, but do not search beyond the path separator. */
    for ( ; i > 0 ; --i) {
        /* Get the character. */
        c = name[i - 1];
        /* Break if c is a dot. */
        if (c == '.') {
            root_length = i - 1;
            break;
        }
        /* Break if c is a path separator. */
        if ((c == '/') || (c == '\\')) {
            break;
        }
    }

    /* Copy the root name to s. */
    cgul_string__append_block(cex, s, name, root_length);

    /* Return the result. */
    rv = cgul_string__take_value(cex, s);

 out:

    /* Clean up. */
    if (s) {
        cgul_string__delete(s);
    }

    return rv;
}

int
cgul_libc__strip_utf8_byte_order_mark(cgul_exception_t* cex,
                                      FILE* f)
{
    int rv = 0;
    int c0 = 0;
    int c1 = 0;
    int c2 = 0;
    fpos_t pos;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(f);

    /* Remember the current file position. */
    cgul_libc__fgetpos(cex, f, &pos);
    if (*cex) {
        goto out;
    }

    /* Read the next three characters from the file. */
    c0 = fgetc(f);
    c1 = fgetc(f);
    c2 = fgetc(f);

    /* Look for the UTF-8 Byte-Order Mark.  Note that this indirectly
     * checks for EOF. */
    if ((c0 == 0xef) && (c1 == 0xbb) && (c2 == 0xbf)) {
        /* Success. */
        rv = 1;
        goto out;
    }

    /* Clear any errors that may have occurred as a result of our
     * reading from the file. */
    clearerr(f);

    /* There was no byte-order mark so we must restore the
     * original position of the file. */
    cgul_libc__fsetpos(cex, f, &pos);
    if (*cex) {
        goto out;
    }

 out:

    return rv;
}
