#include "cgul_line_wrapper.h"
#include "cgul_unicode.h"
#include "cgul_wstring.h"

/**
 * This function sets <code>leading_white_space</code>,
 * <code>next_word</code>, and <code>trailing_white_space</code> to
 * the leading white space, next word, and trailing white space
 * starting at <code>wlines[i]</code>. <p>
 *
 * This function doesn't strictly return the next word.  It will split
 * a word if it detects a hyphen. <p>
 *
 * White space is defined to be only space and tab characters. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] wlines wide-character, multi-line string
 * @param[in] wlines_length length of <code>wlines</code>
 * @param[in] i index into wlines
 * @param[out] leading_white_space leading white space
 * @param[out] next_word next word as a UTF-8 string
 * @param[out] trailing_white_space trailing white space
 */
static void
cgul_line_wrapper__get_next_word(cgul_exception_t* cex,
                                 const cgul_wchar_t* wlines,
                                 size_t wlines_length,
                                 size_t i,
                                 cgul_wstring_t leading_white_space,
                                 cgul_wstring_t next_word,
                                 cgul_wstring_t trailing_white_space)
{
    cgul_wchar_t wc = CGUL_WCHAR__NUL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(wlines);
    assert(leading_white_space);
    assert(trailing_white_space);

    /*
     * Clear the return strings.
     */

    /* leading_white_space */
    cgul_wstring__set_value(
        cex, leading_white_space, CGUL_WCHAR__EMPTY_STRING);

    /* next_word */
    cgul_wstring__set_value(
        cex, next_word, CGUL_WCHAR__EMPTY_STRING);

    /* trailing_white_space */
    cgul_wstring__set_value(
        cex, trailing_white_space, CGUL_WCHAR__EMPTY_STRING);

    if (*cex) {
        goto out;
    }

    /*
     * Leading White Space
     */

    /* Iterate over the wlines copying the leading white space to
     * leading_white_space. */
    for ( ; i < wlines_length ; ++i) {

        /* Get the next wide character. */
        wc = wlines[i];

        /* Copy white space. */
        if ((wc == (cgul_wchar_t)' ') || (wc == (cgul_wchar_t)'\t')) {
            cgul_wstring__append_char(cex, leading_white_space, wc);
            continue;
        }
        break;
    }

    if (*cex) {
        goto out;
    }

    /*
     * Next Word
     */

    /* Iterate over the wlines copying the next word to next_word. */
    for ( ; i < wlines_length ; ++i) {

        /* Get the next wide character. */
        wc = wlines[i];

        /* If this is white space or a new line, we have reached the
         * end of the word. */
        if ((wc == L' ') || (wc == L'\t') || (wc == L'\n')) {
            break;
        }

        /* Copy this character to next_word. */
        cgul_wstring__append_char(cex, next_word, wc);

        /* If this is an otherwise suitable place to break the line,
         * stop copying characters to next_word. */
        if (wc == L'-') {
            ++i;
            break;
        }

        /* If this is a CJK ideograph, we can also break the line here. */
        if (   ((wlines[i] >= 0x3400)  && (wlines[i] <= 0x4dbf))
            || ((wlines[i] >= 0x4e00)  && (wlines[i] <= 0x9fcf))
            || ((wlines[i] >= 0xf900)  && (wlines[i] <= 0xfaff))
            || ((wlines[i] >= 0x20000) && (wlines[i] <= 0x2a6df))
            || ((wlines[i] >= 0x2f800) && (wlines[i] <= 0x2fa1f)))
        {
            break;
        }
    }

    if (*cex) {
        goto out;
    }

    /*
     * Trailing White Space
     */

    /* Iterate over the wlines copying the trailing white space to
     * trailing_white_space. */
    for ( ; i < wlines_length ; ++i) {
        /* Get the next wide character. */
        wc = wlines[i];
        /* Copy white space. */
        if ((wc == (cgul_wchar_t)' ') || (wc == (cgul_wchar_t)'\t')) {
            cgul_wstring__append_char(cex, trailing_white_space, wc);
            continue;
        }
        break;
    }

 out:
    return;
}

void
cgul_line_wrapper__wrap_utf8_at_column(cgul_exception_t* cex,
                                       cgul_string_t lines,
                                       unsigned int column)
{
    cgul_wstring_t utf32 = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(lines);

    /* Create a temporary UTF-32 string that we can pass into
     * cgul_line_wrapper__wrap_utf32_at_width() so it can do the real
     * work. */
    utf32 = cgul_wstring__new(cex);

    /* Generate the UTF-32 version of "lines". */
    cgul_unicode__mbstowcs(cex,
                           cgul_string__get_value(cex, lines),
                           cgul_string__get_length(cex, lines),
                           0, /* skip_bom */
                           utf32);

    /* Wrap the lines. */
    cgul_line_wrapper__wrap_utf32_at_column(cex, utf32, column);
    if (*cex) {
        goto out;
    }

    /* Reset "lines" because cgul_unicode__wcstombs() appends its result. */
    cgul_string__set_value(cex, lines, "");

    /* Convert the result of line wrapping back to UTF-8. */
    cgul_unicode__wcstombs(cex,
                           cgul_wstring__get_value(cex, utf32),
                           cgul_wstring__get_length(cex, utf32),
                           0, /* skip_bom */
                           lines);

 out:

    /* Clean up. */
    if (utf32) {
        cgul_wstring__delete(utf32);
    }

    return;
}

void
cgul_line_wrapper__wrap_utf32_at_column(cgul_exception_t* cex,
                                        cgul_wstring_t wlines_string,
                                        unsigned int column)
{
    size_t i = 0;
    size_t wlines_length = 0;
    const cgul_wchar_t* wlines = NULL;
    cgul_wstring_t leading_white_space = NULL;
    cgul_wstring_t next_word = NULL;
    cgul_wstring_t trailing_white_space = NULL;
    cgul_wstring_t test_string = NULL;
    cgul_wstring_t wrapped_lines = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(wlines_string);

    /* Allocate temporary variables. */
    leading_white_space = cgul_wstring__new(cex);
    next_word = cgul_wstring__new(cex);
    trailing_white_space = cgul_wstring__new(cex);
    test_string = cgul_wstring__new(cex);
    wrapped_lines = cgul_wstring__new(cex);

    /* Unpack wlines_string. */
    wlines = cgul_wstring__get_value(cex, wlines_string);
    wlines_length = cgul_wstring__get_length(cex, wlines_string);

    /* Iterate over wlines. */
    while (i < wlines_length) {

        /*
         * Start a new line.
         */

        /* Get the first word on the line.  We have to treat the
         * first word specially because it is the only time we
         * allow a word to extend beyond the edge of the box.
         * Also, we need to discard all leading white space before
         * the first word so that all the lines align properly. */
        cgul_line_wrapper__get_next_word(
            cex,
            wlines, wlines_length, i,
            leading_white_space,
            next_word,
            trailing_white_space);

        /* If the first word on the line exceeds the box width, as
         * a special case, add it to the return value with a line
         * break to follow.  The alternative would be to break the
         * word in the middle. */
        if (cgul_wstring__get_length(cex, next_word) >= column) {

            /* Add the word to the result. */
            cgul_wstring__append(
                cex,
                wrapped_lines,
                cgul_wstring__get_value(cex, next_word));

            /* Fast-forward past any trailing white space to see
             * if a '\0' or '\n' character follows the first word. */
            i += cgul_wstring__get_length(cex, leading_white_space)
                + cgul_wstring__get_length(cex, next_word)
                + cgul_wstring__get_length(cex, trailing_white_space);

            /* If the string ends without a trailing new-line
             * character, keep it that way.  Otherwise, make sure we
             * insert the line break. */
            if ((i < wlines_length) && (wlines[i] != CGUL_WCHAR__NUL)) {
                cgul_wstring__append_char(
                    cex,
                    wrapped_lines,
                    (cgul_wchar_t)'\n');
            }

            /* Consume the corresponding line break in wlines to
             * avoid a double line space. */
            if ((i < wlines_length) && (wlines[i] == (cgul_wchar_t)'\n')) {
                ++i;
            }

            /* Start the next line. */
            continue;
        }

        /* Ok, the first word fits on the line.  Append it to the
         * return value and reset the test string using it. */
        cgul_wstring__append(
            cex,
            wrapped_lines,
            cgul_wstring__get_value(cex, next_word));

        cgul_wstring__set_value(
            cex,
            test_string,
            cgul_wstring__get_value(cex, next_word));

        if (*cex) {
            goto out;
        }

        /* Fast forward past the leading white space and past the
         * first word.  We do not fast forward past the trailing
         * white space of the first word because it will become
         * the leading white space of the second word which means
         * it is significant for determining the length of the
         * line. */
        i += cgul_wstring__get_length(cex, leading_white_space)
            + cgul_wstring__get_length(cex, next_word);

        /*
         * Processing the remainder of the line.
         */

        /* Iterate over all the other words in this line until the
         * line is too long for the box. */
        while (i < wlines_length) {

            /* Get the next word including leading and trailing
             * white space. */
            cgul_line_wrapper__get_next_word(
                cex,
                wlines, wlines_length, i,
                leading_white_space,
                next_word,
                trailing_white_space);

            if (*cex) {
                goto out;
            }

            /* If the next word is empty, we are at the end of the
             * line. */
            if (cgul_wstring__get_length(cex, next_word) == 0) {

                /* Fast forward to the end of the line. */
                i += cgul_wstring__get_length(cex, leading_white_space)
                    + cgul_wstring__get_length(cex, next_word)
                    + cgul_wstring__get_length(cex, trailing_white_space);

                /* If the string ends without a trailing new-line
                 * character, keep it that way.  Otherwise, make sure we
                 * insert the line break. */
                if ((i < wlines_length) && (wlines[i] != CGUL_WCHAR__NUL)) {
                    cgul_wstring__append_char(cex,
                                              wrapped_lines,
                                              (cgul_wchar_t)'\n');
                }

                /* Skip past the corresponding line break in
                 * wlines to avoid a double line space. */
                if ((i < wlines_length) && (wlines[i] == (cgul_wchar_t)'\n')) {
                    ++i;
                }

                /* The start of the next line is handled by the
                 * outer loop.  So, break out of the inner
                 * loop. */
                break;
            }

            /*
             * The next word wasn't empty so we are not at the end
             * of the line.  Test to see if it fits on the
             * line.
             */

            /* Add the next word (including leading white space)
             * to the test string.  Note that trailing white space
             * is not significant when determining the length of
             * the line so we don't add it to the test_string. */
            cgul_wstring__append(
                cex,
                test_string,
                cgul_wstring__get_value(cex, leading_white_space));

            cgul_wstring__append(
                cex,
                test_string,
                cgul_wstring__get_value(cex, next_word));

            /* If the test string still fits in the box, permanently
             * add the next word (including leading white space) to
             * the return value and start working on the next word on
             * this line by continuing at the top of the inner
             * loop. */
            if (cgul_wstring__get_length(cex, test_string) <= column) {

                /* Add the leading white space to the result. */
                cgul_wstring__append(
                    cex,
                    wrapped_lines,
                    cgul_wstring__get_value(cex, leading_white_space));

                /* Add the next word to the result. */
                cgul_wstring__append(
                    cex,
                    wrapped_lines,
                    cgul_wstring__get_value(cex, next_word));

                /* Fast forward the iterator to just pass the leading
                 * white space and the next word. */
                i += cgul_wstring__get_length(cex, leading_white_space)
                    + cgul_wstring__get_length(cex, next_word);

                continue;
            }

            /* We are here because next_word caused the
             * test_string to be too long.  So, we simply append a
             * line break to the return value without copying
             * next_word to it.  This effectively strips out
             * trailing white space.  We then break out of the
             * inner loop and return to the outer loop.  It is
             * IMPORTANT to understand that "i" still points to
             * the word we just tested.  Thus, the outer loop will
             * retest the same word, but the outer loop does so
             * assuming the word starts a new line. */
            cgul_wstring__append_char(cex, wrapped_lines, (cgul_wchar_t)'\n');
            break;
        }
    }

    /*
     * Success
     */

    /* Return the wrapped lines to the user in wlines_string. */
    cgul_wstring__swap(cex, wrapped_lines, wlines_string);

 out:

    if (wrapped_lines) {
        cgul_wstring__delete(wrapped_lines);
    }

    if (test_string) {
        cgul_wstring__delete(test_string);
    }

    if (trailing_white_space) {
        cgul_wstring__delete(trailing_white_space);
    }

    if (next_word) {
        cgul_wstring__delete(next_word);
    }

    if (leading_white_space) {
        cgul_wstring__delete(leading_white_space);
    }

    return;
}

extern void
cgul_line_wrapper__wrap_utf8_at_width(cgul_exception_t* cex,
                                      cgul_string_t lines,
                                      cgul_bdf_t bdf,
                                      int substitute,
                                      unsigned int max_line_width)
{
    cgul_wstring_t utf32 = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(lines);
    assert(bdf);

    /* Create a temporary UTF-32 string that we can pass into
     * cgul_line_wrapper__wrap_utf32_at_width() so it can do the real
     * work. */
    utf32 = cgul_wstring__new(cex);

    /* Generate the UTF-32 version of "lines". */
    cgul_unicode__mbstowcs(cex,
                           cgul_string__get_value(cex, lines),
                           cgul_string__get_length(cex, lines),
                           0, /* skip_bom */
                           utf32);

    /* Wrap the lines. */
    cgul_line_wrapper__wrap_utf32_at_width(cex, utf32, bdf,
                                           substitute, max_line_width);
    if (*cex) {
        goto out;
    }

    /* Reset "lines" because cgul_unicode__wcstombs() appends its result. */
    cgul_string__set_value(cex, lines, "");

    /* Convert the result of line wrapping back to UTF-8. */
    cgul_unicode__wcstombs(cex,
                           cgul_wstring__get_value(cex, utf32),
                           cgul_wstring__get_length(cex, utf32),
                           0, /* skip_bom */
                           lines);

 out:

    /* Clean up. */
    if (utf32) {
        cgul_wstring__delete(utf32);
    }

    return;
}

extern void
cgul_line_wrapper__wrap_utf32_at_width(cgul_exception_t* cex,
                                       cgul_wstring_t wlines_string,
                                       cgul_bdf_t bdf,
                                       int substitute,
                                       unsigned int max_line_width)
{
    unsigned int width = 0;
    size_t i = 0;
    size_t wlines_length = 0;
    const cgul_wchar_t* wlines = NULL;
    cgul_wstring_t leading_white_space = NULL;
    cgul_wstring_t next_word = NULL;
    cgul_wstring_t trailing_white_space = NULL;
    cgul_wstring_t test_string = NULL;
    cgul_wstring_t wrapped_lines = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(wlines_string);
    assert(bdf);

    /* Allocate temporary variables. */
    leading_white_space = cgul_wstring__new(cex);
    next_word = cgul_wstring__new(cex);
    trailing_white_space = cgul_wstring__new(cex);
    test_string = cgul_wstring__new(cex);
    wrapped_lines = cgul_wstring__new(cex);

    /* Unpack wlines_string. */
    wlines = cgul_wstring__get_value(cex, wlines_string);
    wlines_length = cgul_wstring__get_length(cex, wlines_string);

    /* Iterate over wlines. */
    while (i < wlines_length) {

        /*
         * Start a new line.
         */

        /* Get the first word on the line.  We have to treat the
         * first word specially because it is the only time we
         * allow a word to extend beyond the edge of the box.
         * Also, we need to discard all leading white space before
         * the first word so that all the lines align properly. */
        cgul_line_wrapper__get_next_word(
            cex,
            wlines, wlines_length, i,
            leading_white_space,
            next_word,
            trailing_white_space);

        /* Get the width of just the first word. */
        cgul_bdf__get_utf32_extents(
            cex,
            bdf,
            cgul_wstring__get_value(cex, next_word),
            cgul_wstring__get_length(cex, next_word),
            substitute,
            &width,
            NULL);

        if (*cex) {
            goto out;
        }

        /* If the first word on the line exceeds the box width, as
         * a special case, add it to the return value with a line
         * break to follow.  The alternative would be to break the
         * word in the middle. */
        if (width >= max_line_width) {

            /* Add the word to the result. */
            cgul_wstring__append(
                cex,
                wrapped_lines,
                cgul_wstring__get_value(cex, next_word));

            /* Fast-forward past any trailing white space to see
             * if a '\0' or '\n' character follows the first word. */
            i += cgul_wstring__get_length(cex, leading_white_space)
                + cgul_wstring__get_length(cex, next_word)
                + cgul_wstring__get_length(cex, trailing_white_space);

            /* If the string ends without a trailing new-line
             * character, keep it that way.  Otherwise, make sure we
             * insert the line break. */
            if ((i < wlines_length) && (wlines[i] != CGUL_WCHAR__NUL)) {
                cgul_wstring__append_char(
                    cex,
                    wrapped_lines,
                    (cgul_wchar_t)'\n');
            }

            /* Consume the corresponding line break in wlines to
             * avoid a double line space. */
            if ((i < wlines_length) && (wlines[i] == (cgul_wchar_t)'\n')) {
                ++i;
            }

            /* Start the next line. */
            continue;
        }

        /* Ok, the first word fits on the line.  Append it to the
         * return value and reset the test string using it. */
        cgul_wstring__append(
            cex,
            wrapped_lines,
            cgul_wstring__get_value(cex, next_word));

        cgul_wstring__set_value(
            cex,
            test_string,
            cgul_wstring__get_value(cex, next_word));

        if (*cex) {
            goto out;
        }

        /* Fast forward past the leading white space and past the
         * first word.  We do not fast forward past the trailing
         * white space of the first word because it will become
         * the leading white space of the second word which means
         * it is significant for determining the length of the
         * line. */
        i += cgul_wstring__get_length(cex, leading_white_space)
            + cgul_wstring__get_length(cex, next_word);

        /*
         * Processing the remainder of the line.
         */

        /* Iterate over all the other words in this line until the
         * line is too long for the box. */
        while (i < wlines_length) {

            /* Get the next word including leading and trailing
             * white space. */
            cgul_line_wrapper__get_next_word(
                cex,
                wlines, wlines_length, i,
                leading_white_space,
                next_word,
                trailing_white_space);

            if (*cex) {
                goto out;
            }

            /* If the next word is empty, we are at the end of the
             * line. */
            if (cgul_wstring__get_length(cex, next_word) == 0) {

                /* Fast forward to the end of the line. */
                i += cgul_wstring__get_length(cex, leading_white_space)
                    + cgul_wstring__get_length(cex, next_word)
                    + cgul_wstring__get_length(cex, trailing_white_space);

                /* If the string ends without a trailing new-line
                 * character, keep it that way.  Otherwise, make sure we
                 * insert the line break. */
                if ((i < wlines_length) && (wlines[i] != CGUL_WCHAR__NUL)) {
                    cgul_wstring__append_char(
                                               cex,
                                               wrapped_lines,
                                               (cgul_wchar_t)'\n');
                }

                /* Skip past the corresponding line break in
                 * wlines to avoid a double line space. */
                if ((i < wlines_length) && (wlines[i] == (cgul_wchar_t)'\n')) {
                    ++i;
                }

                /* The start of the next line is handled by the
                 * outer loop.  So, break out of the inner
                 * loop. */
                break;
            }

            /*
             * The next word wasn't empty so we are not at the end
             * of the line.  Test to see if it fits on the
             * line.
             */

            /* Add the next word (including leading white space)
             * to the test string.  Note that trailing white space
             * is not significant when determining the length of
             * the line so we don't add it to the test_string. */
            cgul_wstring__append(
                cex,
                test_string,
                cgul_wstring__get_value(cex, leading_white_space));

            cgul_wstring__append(
                cex,
                test_string,
                cgul_wstring__get_value(cex, next_word));

            /* Calculate the width of the test string. */
            cgul_bdf__get_utf32_extents(
                cex,
                bdf,
                cgul_wstring__get_value(cex, test_string),
                cgul_wstring__get_length(cex, test_string),
                substitute,
                &width,
                NULL);

            if (*cex) {
                goto out;
            }

            /* If the test string still fits in the box, permanently
             * add the next word (including leading white space) to
             * the return value and start working on the next word on
             * this line by continuing at the top of the inner
             * loop. */
            if (width <= max_line_width) {

                /* Add the leading white space to the result. */
                cgul_wstring__append(
                    cex,
                    wrapped_lines,
                    cgul_wstring__get_value(cex, leading_white_space));

                /* Add the next word to the result. */
                cgul_wstring__append(
                    cex,
                    wrapped_lines,
                    cgul_wstring__get_value(cex, next_word));

                /* Fast forward the iterator to just pass the leading
                 * white space and the next word. */
                i += cgul_wstring__get_length(cex, leading_white_space)
                    + cgul_wstring__get_length(cex, next_word);

                continue;
            }

            /* We are here because next_word caused the
             * test_string to be too long.  So, we simply append a
             * line break to the return value without copying
             * next_word to it.  This effectively strips out
             * trailing white space.  We then break out of the
             * inner loop and return to the outer loop.  It is
             * IMPORTANT to understand that "i" still points to
             * the word we just tested.  Thus, the outer loop will
             * retest the same word, but the outer loop does so
             * assuming the word starts a new line. */
            cgul_wstring__append_char(cex, wrapped_lines, (cgul_wchar_t)'\n');
            break;
        }
    }

    /*
     * Success
     */

    /* Return the wrapped lines to the user in wlines_string. */
    cgul_wstring__swap(cex, wrapped_lines, wlines_string);

 out:

    if (wrapped_lines) {
        cgul_wstring__delete(wrapped_lines);
    }

    if (test_string) {
        cgul_wstring__delete(test_string);
    }

    if (trailing_white_space) {
        cgul_wstring__delete(trailing_white_space);
    }

    if (next_word) {
        cgul_wstring__delete(next_word);
    }

    if (leading_white_space) {
        cgul_wstring__delete(leading_white_space);
    }

    return;
}
