#ifndef CGUL_MICROXML_P_H
#define CGUL_MICROXML_P_H

#include "cgul_microxml.h"
#include "cgul_cache.h"
#include "cgul_vector.h"
#include "cgul_wstring.h"

CGUL_BEGIN_C

/**
 * States for the state machine used by
 * <code>cgul_microxml__read_next_token()</code> to pick up where it
 * left off.  With a push parser, this isn't a problem because push
 * parsers runs to completion, but a pull parser (like the one
 * implemented here) needs to be able to stop everything in order to
 * return the next token to the user.  Then on the next call, it needs
 * to be able to pick up where it left off.
 */
typedef enum cgul_microxml__read_state {

    /**
     * Invalid state.
     */
    CGUL_MICROXML__READ_STATE__INVALID,

    /**
     * Document start state.
     */
    CGUL_MICROXML__READ_STATE__DOCUMENT_START,

    /**
     * Read the prolog (before the root element) which can contain
     * white space or comments.
     */
    CGUL_MICROXML__READ_STATE__PROLOG,

    /**
     * Main loop state.
     */
    CGUL_MICROXML__READ_STATE__MAIN_LOOP,

    /**
     * Inject element end token.
     */
    CGUL_MICROXML__READ_STATE__INJECT_ELEMENT_END_TOKEN,

    /**
     * Read the epilog (after the root element) which can contain
     * white space or comments.
     */
    CGUL_MICROXML__READ_STATE__EPILOG,

    /**
     * Document end state.
     */
    CGUL_MICROXML__READ_STATE__DOCUMENT_END,

    /**
     * Final state.
     */
    CGUL_MICROXML__READ_STATE__FINAL,

} cgul_microxml__read_state_t;

/**
 * Enumerate the MicroXML token types.
 */
typedef enum cgul_microxml__token {

    /*
     * REMEMBER: Keep the following methods up to date with any
     * changes made to this structure:
     *
     *     cgul_microxml__token_to_string()
     *     cgul_microxml__token_from_string()
     */

    /**
     * Token is invalid indicating no token has been read yet or an
     * error occurred.
     */
    CGUL_MICROXML__TOKEN_INVALID = 0,

    /**
     * Document start token indicating that tokenization is starting.
     * This is always the first token.
     */
    CGUL_MICROXML__TOKEN_DOCUMENT_START,
    
    /**
     * Document end token indicating that tokenization has ended.
     * This is always the last token if no errors occur.  Attempting
     * to get the next token after this one results in an error.
     */
    CGUL_MICROXML__TOKEN_DOCUMENT_END,
    
    /**
     * Element start token indicating an start tag and attributes have
     * been read.  Use <code>cgul_microxml__get_element_name()</code>
     * and <code>cgul_microxml__get_attributes()</code> to get the
     * name and attributes associated wit the start tag.  Use
     * <code>cgul_microxml__is_empty_element()</code> to determine if
     * the empty-token syntax was used to close the tag.
     */
    CGUL_MICROXML__TOKEN_ELEMENT_START,

    /**
     * Element end token indicating an closing tag has been read or,
     * if an empty element is parsed, this token will be injected into
     * the stream before reading the next token.  This way the
     * following two XML samples that both indicate an empty element
     * generate the same token stream: <p>
     *
     * <pre>
     *     &lt;foo&gt;&lt;/foo&gt;
     *     &lt;foo/&gt;
     * </pre>
     *
     * Use <code>cgul_microxml__get_element_name()</code> to get the
     * name associated with the token. <p>
     */
    CGUL_MICROXML__TOKEN_ELEMENT_END,

    /**
     * Text token indicating a character list has been read.  Use
     * <code>cgul_microxml__get_text()</code> get the text
     * associated with the token.
     */
    CGUL_MICROXML__TOKEN_TEXT,

    /**
     * Comment token indicating a comment has been read.  Use
     * <code>cgul_microxml__get_comment()</code> to get the comment
     * associated with the token.
     */
    CGUL_MICROXML__TOKEN_COMMENT,

} cgul_microxml__token_t;

/**
 * Data for each open element that is pushed onto the
 * <code>open_elements</code> stack.
 */
typedef struct open_element {

    /**
     * Name of the element as an interned symbol (which is owned by
     * the symbol table "symbols").
     */
    const char* element_symbol;

    /**
     * Starting line number for the element.
     */
    size_t line_start;

    /**
     * Starting column number for the element.
     */
    size_t column_start;

}* open_element_t;

/** Data encapsulated by the <code>cgul_microxml</code> class. */
struct cgul_microxml {

    /**
     * Read state to control where
     * <code>cgul_microxml__read_next_token()</code> picks up next.
     */
    cgul_microxml__read_state_t read_state;

    /**
     * The current MicroXML token.  It is updated by calls to
     * <code>cgul_microxml__read_next_token()</code>.
     */
    cgul_microxml__token_t current_token;

    /**
     * The starting line for the current token.
     */
    size_t token_line_start;

    /**
     * The starting column for the current token.
     */
    size_t token_column_start;

    /**
     * The current character being parsed which is valid only after
     * the first read and before end of file (EOF) is reached.  The
     * early state machine transitions will make sure the first read
     * occurs before <code>current_char</code> is accessed, and the
     * state machine will also transition directly to the
     * <code>DOCUMENT_END</code> state if it detects EOF for the end
     * tag of the root element.
     */
    cgul_wchar_t current_char;

    /**
     * Symbol table.  This is used to conserve space by interning
     * symbols that are used repeatedly throughout an XML file.
     * Element names and attribute names will be interned.  This
     * allows a DOM to be created without have to duplicate all the
     * repeated element and attribute names.
     */
    cgul_hash_t symbols;

    /**
     * Cache of temporary <code>cgul_string</code> objects used to
     * avoid thrashing the allocator.
     */
    cgul_cache_t strings_cache;

    /**
     * Cache of <code>cgul_hash</code> objects used to avoid thrashing
     * the allocator.  The "restructor" for the hashes_cache is set to
     * <code>NULL</code> so you must remember to call
     * <code>cgul_microxml__clear_element_attributes()</code> before
     * putting an attributes hash back on this cache.
     */
    cgul_cache_t hashes_cache;

    /**
     * The maximum length to which the <code>entity_utf8</code> and
     * <code>entity_utf32</code> entities are allowed to grow.  If set
     * to <code>0</code>, the length is limited only by available
     * memory.
     *
     * @see entity_utf8
     * @see entity_utf32
     */
    size_t entity_length_max;

    /**
     * Temporary string used to build up entities as UTF-8 strings.
     * It is better to allocate this just once instead of once for
     * each character reference.  You have to be careful to not use
     * this string in a recursive context.
     *
     * @see entity_length_max
     */
    cgul_string_t entity_utf8;

    /**
     * Temporary string used to build up entities as UTF-32 strings.
     * It is better to allocate this just once instead of once for
     * each character reference.  You have to be careful to not use
     * this string in a recursive context.
     *
     * @see entity_length_max
     */
    cgul_wstring_t entity_utf32;

    /**
     * Cache of <code>open_element</code> objects to avoid thrashing
     * the allocator.
     */
    cgul_cache_t open_elements_cache;

    /**
     * Stack of open elements.  Each element on the stack is a C-style
     * string holding the name of an open element.  This stack is used
     * to make sure the XML is well-formed.
     */
    cgul_vector_t open_elements;

    /**
     * The result of interning <code>element_name</code> for the
     * current element.
     *
     * @see element_name
     */
    const char* element_symbol;

    /**
     * Attributes associated with the token.  This is for specifying
     * the attributes associated with a start tag.
     */
    cgul_hash_t element_attributes;

    /**
     * Whether a start tag was closed using the empty-element syntax
     * (e.g., <foo/>).  Do *not* clear this member between calls to
     * <code>cgul_microxml__read_next_token()</code> because it needs
     * to be valid for both the actual empty element start tag *and*
     * its artificial end tag.
     */
    int is_empty_element;

    /**
     * Text associated with a token.  This is for specifying textual
     * content.
     *
     * @see text_length_max
     */
    cgul_string_t text;

    /**
     * The maximum text length for <code>text</code> in Unicode
     * characters.  If set to <code>0</code>, the size of
     * <code>text</code> is limited only by available memory which is
     * the default.
     *
     * @see text
     */
    size_t text_length_max;

    /**
     * Whether the text is pure white space.  This flag is only valid
     * if <code>current_token</code> is set to
     * <code>CGUL_MICROXML__TOKEN_TEXT</code>.
     */
    int is_white_space;

    /**
     * Whether to tokenize pure white space.
     */
    int skip_white_space;

    /**
     * Comment associated with a token.  This is for specifying the
     * text of a comment.
     *
     * @see comment_length_max
     */
    cgul_string_t comment;

    /**
     * Maximum length to which <code>comment</code> is allowed to
     * grow.  If set to <code>0</code>, the length is limited only by
     * available memory.
     *
     * @see comment
     */
    size_t comment_length_max;

    /**
     * Whether to tokenize comments.
     */
    int skip_comments;

    /**
     * Whether to skip prolog and epilog tokens.  These are the white
     * space and comment tokens that surround the root element.  By
     * default, these tokens are skipped.
     */
    int skip_prolog_or_epilog_tokens;

    /**
     * Stream reader used to read characters from the block reader.
     */
    cgul_stream_reader_t sr;

    /**
     * Whether we own <code>sr</code>.
     */
    int we_own_sr;

    /**
     * Block reader used to read blocks from a file or from memory.
     * This is strictly an internal data structure.  The client passes
     * us stream readers, not block readers.  It is used when the
     * client uses the "easy" API and just gives us the name of the
     * file to parse, a <code>FILE*</code>, or a block of memory.  In
     * this case, we need to internally allocate a block reader.
     */
    cgul_block_reader_t br;

};

/**
 * Return the string representation of the token <code>code</code>.
 * If an error occurs, <code>NULL</code> is returned, and an exception
 * is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] token token
 * @return string representation of <code>token</code>
 */
extern CGUL_EXPORT const char*
cgul_microxml__token_to_string(cgul_exception_t* cex,
                               cgul_microxml__token_t token);

/**
 * Convert the string <code>s</code> to a
 * <code>cgul_microxml__token</code> and return the result.  If an
 * error occurs, <code>CGUL_MICROXML__INVALID_TOKEN</code> is
 * returned, and an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param s string
 */
extern CGUL_EXPORT cgul_microxml__token_t
cgul_microxml__token_from_string(cgul_exception_t* cex,
                                 const char* s);

/**
 * Return a new, uninitialized <code>open_element</code> structure.
 * If an error occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 */
extern CGUL_EXPORT open_element_t
cgul_microxml__open_element__new(cgul_exception_t* cex);

/**
 * Delete the open element structure <code>open_element</code> freeing
 * all internally allocated resources.
 *
 * @param]in] open_element open element
 */
extern CGUL_EXPORT void
cgul_microxml__open_element__delete(open_element_t open_element);

/**
 * Reinitialize the open element structure <code>open_element</code>.
 *
 * @param]in] open_element open element
 */
extern CGUL_EXPORT void
cgul_microxml__open_element__reinitialize(open_element_t open_element);

CGUL_END_C

#endif  /*  CGUL_MICROXML_P_H  */
