#ifndef CGUL_MIME_P_H
#define CGUL_MIME_P_H

#include "cgul_common.h"
#include "cgul_mime.h"
#include "cgul_string.h"
#include "cgul_substring.h"

CGUL_BEGIN_C

/**
 * Context for the <code>cgul_mime__parse()</code> function
 * which is broken down into many smaller functions all of which are
 * tightly coupled using the same context.
 * 
 */
struct cgul_mime__parse__context {

    /**
     * Whether we own <code>buf</code>.  This is needed because when
     * parsing a MIME file in memory, it is much more efficient to
     * directly parse the user-supplied memory rather than copying
     * blocks from the user-supplied memory into the same small buffer
     * owned by the parser.
     */
    int own_buf;

    /**
     * Buffer used to read in one block at a time from the MIME file.
     * <code>free()</code> must be called on this pointer when the
     * client has finished with the context if and only if
     * <code>own_buf</code> is true.
     */
    char* buf;

    /**
     * Total size of <code>buf</code> in bytes.
     */
    size_t buf_size;

    /**
     * Index into <code>buf</code> that points to the current byte
     * being processed.
     */
    size_t i;

    /**
     * Active count.  This is the number of bytes that are active in
     * <code>buf</code>.  <code>acount</code> should always be less
     * than or equal to buf_size, and the the code should never
     * attempt to access bytes beyond <code>buf[acount - 1]</code>.
     *
     * @see rcount
     */
    size_t acount;

    /**
     * Read count.  Number of bytes that were read into
     * <code>buf</code> from the MIME file.  This is not necessarily
     * the same as <code>acount</code> because the code needs to
     * stitch together the blocks it reads from the MIME file.  So,
     * when a block needs to be read from the MIME file, the code
     * copies the unused bytes from the end of <code>buf</code> to the
     * beginning of <code>buf</code> before reading more bytes from
     * the MIME file.  <code>rcount</code> is the number of bytes that
     * have been read from file; whereas, <code>acount</code> is
     * <code>rcount</code> plus the number of bytes that were copied
     * from the end of <code>buf</code> to the beginning.
     *
     * @see acount
     */
    size_t rcount;

    /**
     * MIME boundary string.  The user passes in the MIME boundary
     * string with all but the leading and trailing "--" prefix and
     * suffix.  The code will append a leading EOL sequence followed
     * by the "--" prefix.  It is tempting to have the code also
     * append a trailing EOL sequence, but the MIME standard allows
     * for optional whitespace between the boundary string and the
     * trailing EOL sequence.
     */
    cgul_substring_t b;

    /**
     * Final MIME boundary string which is the same as <code>b</code>
     * except it has a "--" suffix.
     */
    cgul_substring_t b_final;

    /**
     * String that holds a pair of EOL sequences.  For example, if the
     * user passes in <code>"\r\n"</code> as the EOL sequence,
     * <code>two_eols</code> will be set to <code>"\r\n\r\n"</code>.
     * This is needed when searching for the two EOLs that separate
     * the headers from the bodies.
     */
    cgul_substring_t two_eols;

    /**
     * The EOL sequence.  This is passed in by the user and then
     * copied into this structure for use by sub-procedures.  This is
     * just a copy of the user's pointer so do not call
     * <code>free()</code> on this pointer.
     */
    const char* eol;

    /**
     * Length in bytes of the EOL sequence.
     */
    size_t eol_length;

    /**
     * Count of MIME parts that have been fully processed.  This value
     * is not incremented until the end of the body is found so in
     * many cases you need to use <code>part + 1</code>.
     */
    unsigned long int part;

    /**
     * Callback registered by the client to produce a new block from
     * the original MIME source.
     */
    cgul_mime__produce_t produce;

    /**
     * Client data for the <code>produce</code> callback.
     */
    void* produce_data;

    /**
     * Callback registered by the client to consume each block as it
     * is parsed.
     */
    cgul_mime__consume_t consume;

    /**
     * Client data for the <code>consume</code> callback.
     */
    void* consume_data;

    /**
     * The client can cancel further processing by having the
     * <code>consume</code> callback return false.  Note that the
     * polarity for <code>done</code> is the opposite of what
     * <code>consume</code> returns.
     */
    int done;
};

/** Pointer to <code>cgul_mime__parse__context</code> structure. */
typedef struct cgul_mime__parse__context*
cgul_mime__parse__context_t;

/**
 * The <code>cgul_mime__split_file()</code> and
 * <code>cgul_mime__split_memory()</code> functions are written in
 * terms of <code>cgul_mime__parse_file()</code> and
 * <code>cgul_mime__parse_memory()</code>.  The structure below is the
 * client data for the callback registered with
 * <code>cgul_mime__parse()</code>.
 */
struct cgul_mime__split__context {

    /**
     * The directory name where the header and body files for each
     * MIME part will be written.  This is just a copy of the user's
     * pointer so do not call <code>free()</code> on this pointer.
     */
    const char* dname;

    /**
     * The name for the preamble file.  This is just a copy of the
     * user's pointer so do not call <code>free()</code> on this
     * pointer.
     */
    const char* pname;

    /**
     * The base name for the header files.  This is just a copy of the
     * user's pointer so do not call <code>free()</code> on this
     * pointer.
     */
    const char* hname;

    /**
     * The base name for the body files.  This is just a copy of the
     * user's pointer so do not call <code>free()</code> on this
     * pointer.
     */
    const char* bname;

    /**
     * The name for the epilogue file.  This is just a copy of the
     * user's pointer so do not call <code>free()</code> on this
     * pointer.
     */
    const char* ename;

    /**
     * The previous section.
     */
    cgul_mime__section_t prev_section;

    /**
     * Output file.
     */
    FILE* fout;

    /**
     * Temporary string used for generating file names.
     */
    cgul_string_t tmp;

};

/** Pointer to <code>cgul_mime__split_file__context</code> structure. */
typedef struct cgul_mime__split__context*
cgul_mime__split__context_t;

CGUL_END_C

#endif  /*  CGUL_MIME_P_H  */
