#ifndef CGUL_RBTREE_P_H
#define CGUL_RBTREE_P_H

#include "cgul_common.h"
#include "cgul_cache.h"
#include "cgul_rbtree.h"
#include "cgul_rbtree_node_p.h"

CGUL_BEGIN_C

/**
 * Data encapsulated by the <code>cgul_rbtree</code> class.
 */
struct cgul_rbtree {

    /**
     * Whether this <code>cgul_rbtree</code> object is indexed.  If it
     * is indexed, it means <code>cgul_rbtree__find_at()</code> and
     * <code>cgul_rbtree__remove_at()</code> are available and will
     * not throw an exception if called.  We go to all the trouble of
     * having two different classes that can be used as nodes in our
     * <code>cgul_rbtree</code> because the indexing feature costs two
     * extra <code>unsigned long</code> values per node.  It also
     * slows down inserts and deletes by about 15%.
     */
    int is_indexed;

    /**
     * Size of the tree.  This is the count of the number of nodes
     * (or key/value pairs) held by the tree.
     */
    unsigned long int size;

    /**
     * Root node of the tree.
     */
    cgul_rbtree_node_t root;

    /**
     * Front node of the tree.  This is the node that is currently
     * first in sorted order.
     */
    cgul_rbtree_node_t front;

    /**
     * Back node of the tree.  This is the node that is currently last
     * in sorted order.
     */
    cgul_rbtree_node_t back;

    /**
     *  The oldest node in the tree.  This is the first node that was
     *  inserted.
     */
    cgul_rbtree_node_t oldest;

    /**
     *  The youngest node in the tree.  This is the last node that was
     *  inserted.
     */
    cgul_rbtree_node_t youngest;

    /**
     * Comparison function used to determine sorted order when
     * inserting.
     */
    cgul_rbtree__compare_t compare;

    /**
     * For efficiency, the <code>cgul_rbtree</code> can keep a cache
     * of nodes that can be reused.  In many situations, this can
     * greatly reduce the number of calls to <code>malloc()</code>
     * which can greatly improve performance and reduce memory
     * fragmentation.
     */
    cgul_cache_t cache;

};

/**
 * Data needed for <code>cgul_rbtree__clear_cb()</code>.
 */
typedef struct cgul_rbtree__clear_cb_data {

    /** The <code>cgul_cache</code> the client is using for caching keys. */
    cgul_cache_t keys_cache;

    /** The <code>cgul_cache</code> the client is using for caching values. */
    cgul_cache_t values_cache;

    /** Return value indicating whether any nodes have been removed. */
    int rv;

} *cgul_rbtree__clear_cb_data_t;

CGUL_END_C

#endif  /*  CGUL_RBTREE_P_H  */
