#ifndef CGUL_STRING_P_H
#define CGUL_STRING_P_H

#include "cgul_common.h"
#include "cgul_string.h"

CGUL_BEGIN_C

/**
 * Data encapsulated by the <code>cgul_string</code> class.
 */
struct cgul_string {

    /**
     * The user's string. <p>
     *
     * This parameter must be the first parameter in the structure
     * because it allows the user to cast <code>cgul_string_t</code>
     * to a <code>char*</code>.  For example: <p>
     *
     * <pre>
     *     cgul_string_t s = cgul_string__new_from_cstring(&local, "hello");
     *     printf("s = %s\n", *(char**)s);
     * </pre>
     *
     * (See tests/main_string_as_cstring.c for a full example.)
     */
    char* value;

    /** Case-insensitive copy of the user's string.  */
    char* ivalue;

    /**
     * Whether we can use the cached copy of "ivalue" or whether we
     * need to generate a new one.
     */
    int is_ivalue_valid;

    /** Length of the user's string. */
    size_t length;

    /**
     * Current capacity of this cgul_string object.  The capacity is
     * the number of bytes the string can hold including the trailing
     * NUL terminator.  For example, the capacity must be at least 8
     * in order to store a character having 7 ASCII characters plus a
     * NUL terminator.
     */
    size_t capacity;

    /**
     * Minimum capacity below which <code>capacity</code> is not
     * allowed to fall.
     */
    size_t minimum_capacity;

};

/**
 * Set the capacity to <code>capacity</code> for the string
 * <code>s</code>.  If <code>capacity</code> is less than the current
 * size, the current size is used instead.  If <code>capacity</code>
 * is zero, the default capacity is used instead.  If memory cannot be
 * allocated, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] s <code>cgul_string</code> instance
 * @param[in] capacity new capacity
 */
extern CGUL_EXPORT void
cgul_string__set_capacity(cgul_exception_t* cex,
                          cgul_string_t s,
                          size_t capacity);

/**
 * If necessary, decrease capacity for the string <code>s</code> such
 * that the capacity is reduced by 1/2 if the size is 1/4th the
 * capacity.
 *
 * @param[in,out] cex c-style exception
 * @param[in] s <code>cgul_string</code> instance
 */
extern CGUL_EXPORT void
cgul_string__decrease_capacity_if_necessary(cgul_exception_t* cex,
                                            cgul_string_t s);

/**
 * If necessary, increase capacity to accommodate appending a string
 * that is <code>extra_size</code> bytes long.  When calculating
 * <code>extra_size</code>, the caller does not need to include the
 * character associated with the NUL terminator because an extra byte
 * will be added automatically.  If more memory cannot be allocated,
 * an exception is thrown, but the underlying string is not altered.
 *
 * @param[in,out] cex c-style exception
 * @param[in] s this string
 * @param[in] extra_size minimum number of extra bytes to add to the capacity
 */
extern CGUL_EXPORT void
cgul_string__increase_capacity_if_necessary(cgul_exception_t* cex,
                                            cgul_string_t s,
                                            size_t extra_size);

CGUL_END_C

#endif  /*  CGUL_STRING_P_H  */
