#include "cgul_vector_p.h"
#include "cgul_defensive.h"
#include "cgul_libc.h"
#include "cgul_merge_sort.h"
#include "cgul_shuffle.h"

/**
 * Default capacity for each newly created <code>cgul_vector</code>.
 */
static const size_t CGUL_VECTOR__DEFAULT_MINIMUM_CAPACITY = 128;

/**
 * Set the capacity to <code>capacity</code> for the vector
 * <code>v</code>.  If <code>capacity</code> is less than the current
 * size, the current size is used instead.  If <code>capacity</code>
 * is zero, the default capacity is used instead.  If memory cannot be
 * allocated, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] v <code>cgul_vector</code> instance
 * @param[in] capacity new capacity to request
 */
static void
cgul_vector__set_capacity(cgul_exception_t* cex,
                          cgul_vector_t v,
                          size_t capacity)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(v);

    /* Sanity check. */
    if (capacity == 0) {
        capacity = v->minimum_capacity;
    }

    /* Make sure the capacity is at least large enough to hold the
     * current number of elements. */
    capacity = CGUL_MAX(capacity, v->size);

#if 0
    fprintf(stderr, "cgul_vector__set_capacity = %lu\n", capacity);
    fflush(stderr);
#endif

    /* realloc */
    v->v = (void**)cgul_libc__realloc(cex, v->v, sizeof(*v->v) * capacity);
    if (*cex) {
        goto out;
    }

    /* Rememeber the new capacity. */
    v->capacity = capacity;

    /* Clear the extra elements. */
    if (capacity > v->size) {
        memset(v->v + v->size,
               0,
               sizeof(*v->v) * (v->capacity - v->size));
    }

 out:
    return;
}

/**
 * If necessary, decrease capacity for the vector <code>v</code> such
 * that the capacity is reduced by 1/2 if the size is 1/4th the
 * capacity.
 *
 * @param[in,out] cex c-style exception
 * @param[in] v <code>cgul_vector</code> instance
 */
static void
cgul_vector__decrease_capacity_if_necessary(cgul_exception_t* cex,
                                            cgul_vector_t v)
{
    size_t actual_new_capacity = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* We only have excess capacity when 1/4 capacity is greater than
     * or equal to the size.  So, do nothing when 1/4 capacity is less
     * than the size. */
    if ((v->capacity >> 2) < v->size) {
        goto out;
    }

    /* Keep halving the capacity to get the best fit.  It should be at
     * least double the size to avoid thrashing the allocator.  We
     * know from the if() above that we can start iterating with 1/2
     * of v->capacity. */
    actual_new_capacity = v->capacity >> 1;
    while (   ((actual_new_capacity >> 1) > 0)
           && ((actual_new_capacity >> 1) >= (v->size << 1)))
    {
        actual_new_capacity >>= 1;
    }

    /* The actual new capacity should never drop below the minimum
     * capacity. */
    actual_new_capacity = CGUL_MAX(actual_new_capacity, v->minimum_capacity);
    assert(actual_new_capacity > 0);

    /* If the capacity is already less than or equal to the actual new
     * capacity, do nothing. */
    if (v->capacity <= actual_new_capacity) {
        goto out;
    }

    /* Sanity check to make sure we are not going to accidentally
     * reduce the capacity below the necessary size. */
    if (actual_new_capacity <= v->size) {
        goto out;
    }

    /* Set the reduced capacity. */
    cgul_vector__set_capacity(cex, v, actual_new_capacity);
    if (*cex) {
        goto out;
    }

 out:
    return;
}

/**
 * Wrapper around <code>cgul_vector__increase_capacity_if_necessary()</code>
 * that performs a quick check to determine if there is already enough
 * excess capacity to accommodate the extra size.  If not, the real
 * method is called.
 *
 * @param[in,out] cex c-style exception
 * @param[in] v <code>cgul_vector</code> instance
 * @param[in] extra_size extra size
 */
#define CGUL_VECTOR__INCREASE_CAPACITY_IF_NECESSARY(cex, v, extra_size) \
do {                                                                    \
    size_t excess_capacity = (v)->capacity - (v)->size;                 \
    size_t local_extra_size = (extra_size);                             \
    if (excess_capacity < local_extra_size) {                           \
        cgul_vector__increase_capacity_if_necessary(                    \
            (cex), (v), local_extra_size);                              \
    }                                                                   \
} while (0)

/**
 * If necessary, increase capacity to <code>capacity</code> for the
 * vector <code>v</code>.
 *
 * @param[in,out] cex c-style exception
 * @param[in] v <code>cgul_vector</code> instance
 * @param[in] extra_size how many extra slots to add to the underlying array
 */
static void
cgul_vector__increase_capacity_if_necessary(cgul_exception_t* cex,
                                            cgul_vector_t v,
                                            size_t extra_size)
{
    size_t needed_capacity = 0;
    size_t target_capacity = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(v);

    /* Sanity check. */
    if (extra_size == 0) {
        goto out;
    }

    /*
     * NOTE:
     * ====
     *
     * It is more efficient (on average) to call this method
     * through the CGUL_VECTOR__INCREASE_CAPACITY_IF_NECESSARY() macro
     * defined above.
     */

    /* Calculate the needed capacity. */
    needed_capacity =
        cgul_defensive__add_unsigned_long(cex, v->size, extra_size);
    if (*cex) {
        goto out;
    }

    /* If we need more capacity, realloc() it.*/
    if (needed_capacity > v->capacity) {

        /* Try to double the capacity to allow room to grow.  The test
         * against zero is used to detect overflow. */
        target_capacity = v->capacity;
        while (   ((target_capacity << 1) > 0)
               && ((target_capacity << 1) <= (needed_capacity << 1)))
        {
            target_capacity <<= 1;
        }

        /* If there is overflow, just use needed_capacity as-is which
         * doesn't allow for room to grow without reallocating but is
         * large enough to hold the new elements.  Also, we know that
         * needed_capacity did not overflow when calculated above
         * because an exception would have been thrown. */
        if (target_capacity < needed_capacity) {
            target_capacity = needed_capacity;
        }

        /* Set the new capacity. */
        cgul_vector__set_capacity(cex, v, target_capacity);
    }

 out:
    return;
}

static void
cgul_vector__initialize(cgul_exception_t* cex,
                        cgul_vector_t v)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Free the old underlying array if any. */
    if (v->v) {
        free(v);
    }

    /* Initialize all the member variables. */
    v->v = NULL;
    v->size = 0;
    v->capacity = 0;

    /* minimum_capacity -- initialized in constructor */

    /* Set the initial capacity. */
    cgul_vector__set_capacity(cex, v, v->minimum_capacity);
    if (*cex) {
        goto out;
    }

 out:
    return;
}

/* Comments are in cgul_vector_p.h. */
cgul_vector_t
cgul_vector__new_with_capacity(cgul_exception_t* cex,
                               size_t capacity)
{
    cgul_vector_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Allocate the cgul_vector object. */
    rv = (cgul_vector_t)cgul_libc__calloc(cex, 1, sizeof(*rv));
    if (*cex) {
        goto out;
    }

    /* minimum_capacity */
    rv->minimum_capacity = capacity;

    /* Initialize. */
    cgul_vector__initialize(cex, rv);
    if (*cex) {
        goto out;
    }

 out:

    /* Clean up partially initialized objects. */
    if (*cex && rv) {
        cgul_vector__delete(rv);
        rv = NULL;
    }

    return rv;
}

cgul_vector_t
cgul_vector__new(cgul_exception_t* cex)
{
    cgul_vector_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    rv = cgul_vector__new_with_capacity(cex,
                                        CGUL_VECTOR__DEFAULT_MINIMUM_CAPACITY);

 out:
    return rv;
}

void
cgul_vector__delete(cgul_vector_t v)
{
    if (v) {
        /* Free the vector if one exits. */
        if (v->v) {
            free(v->v);
        }
        /* Free the object itself. */
        free(v);
    }
}

void
cgul_vector__free_values(cgul_vector_t v)
{
    size_t i = 0;

    if (v) {
        /* Iterate over each value. */
        for (i = 0 ; i < v->size ; ++i) {
            /* Free each value. */
            free(v->v[i]);
        }
    }
}

int
cgul_vector__is_empty(cgul_exception_t* cex,
                      cgul_vector_t v)
{
    int rv = 1;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(v);

    rv = (v->size == 0);

 out:
    return rv;
}

void
cgul_vector__assign(cgul_exception_t* cex,
                    cgul_vector_t v,
                    size_t index,
                    const void* ptr)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(v);

    if (index >= v->size) {
        *cex = cgul_exception__new("cgul_vector__assign: "
                                   "index out of bounds: %zu",
                                   index);
        goto out;
    }

    assert(v->v);

    v->v[index] = (void*)ptr;

 out:
    return;
}

void
cgul_vector__insert(cgul_exception_t* cex,
                    cgul_vector_t v,
                    size_t index,
                    const void* ptr)
{
    void** data = NULL;
    size_t i = 0;
    size_t j = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(v);

    /* We allow inserting at the end. */
    if (index > v->size) {
        *cex = cgul_exception__new("cgul_vector__insert: "
                                   "index out of bounds: %zu",
                                   index);
        goto out;
    }

    assert(v->v);

    CGUL_VECTOR__INCREASE_CAPACITY_IF_NECESSARY(cex, v, 1);
    if (*cex) {
        goto out;
    }

    data = v->v;

    /* Shift the data right 1 slot being careful not to overwrite. */
    i = v->size;
    j = i - 1;
    for( ; i > index ; --i, --j ) {
        data[i] = data[j];
    }

    /* Insert the new element. */
    data[index] = (void*)ptr;

    /* Increase the size. */
    ++(v->size);

 out:
    return;
}

void
cgul_vector__erase(cgul_exception_t* cex,
                   cgul_vector_t v,
                   size_t index)
{
    void** data = NULL;
    size_t i = 0;
    size_t j = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(v);

    if (index >= v->size) {
        *cex = cgul_exception__new("cgul_vector__erase: "
                                   "index out of bounds: %zu",
                                   index);
        goto out;
    }

    assert(v->v);
    data = v->v;

    /* Shift the data left 1 slot. */
    i = index;
    j = i + 1;
    for( ; j < v->size ; ++i, ++j ) {
        data[i] = data[j];
    }

    /* Reduce the size. */
    --(v->size);

    /* Decrease capacity if necessary. */
    cgul_vector__decrease_capacity_if_necessary(cex, v);

 out:
    return;
}

void
cgul_vector__erase_unordered(cgul_exception_t* cex,
                             cgul_vector_t v,
                             size_t index)
{
    void** data = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(v);

    if (index >= v->size) {
        *cex = cgul_exception__new("cgul_vector__erase_unordered: "
                                   "index out of bounds: %zu",
                                   index);
        goto out;
    }

    assert(v->v);
    data = v->v;

    /* Shrink the size by one.  We do this here to minimize the number
     * of times we have to do this subtraction. */
    --(v->size);

    /* Replace the erased element with the back element noting that we
     * have already subtracted 1 from v->size above.  Also note that
     * if we are erasing the back element, the following assignment is
     * harmless. */
    data[index] = data[v->size];

    /* Decrease capacity if necessary. */
    cgul_vector__decrease_capacity_if_necessary(cex, v);

 out:
    return;
}

void*
cgul_vector__at(cgul_exception_t* cex,
                cgul_vector_t v,
                size_t index)
{
    void* rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(v);
    assert(v->v);

    if (index >= v->size) {
        *cex = cgul_exception__new("cgul_vector__at: "
                                   "invalid index: %zu",
                                   index);
        goto out;
    }

    /* Success. */
    rv = v->v[index];

 out:
    return rv;
}

void*
cgul_vector__get_back(cgul_exception_t* cex,
                      cgul_vector_t v)
{
    void* rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(v);
    assert(v->v);

    if( v->size > 0 ) {
        rv = v->v[v->size - 1];
    }

 out:
    return rv;
}

void
cgul_vector__push_back(cgul_exception_t* cex,
                       cgul_vector_t v,
                       const void* ptr)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(v);
    assert(v->v);

    CGUL_VECTOR__INCREASE_CAPACITY_IF_NECESSARY(cex, v, 1);
    if (*cex) {
        goto out;
    }

    v->v[v->size] = (void*)ptr;
    ++(v->size);

 out:
    return;
}

void*
cgul_vector__pop_back(cgul_exception_t* cex,
                      cgul_vector_t v)
{
    void* rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(v);
    assert(v->v);

    if( v->size > 0 ) {
        --(v->size);
        rv = v->v[v->size];
        cgul_vector__decrease_capacity_if_necessary(cex, v);
    }

 out:
    return rv;
}

void
cgul_vector__sort(cgul_exception_t* cex,
                  cgul_vector_t v,
                  cgul_vector__compare_t f)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(v);
    assert(f);

    /* Sort. */
    cgul_merge_sort_pointers(cex,
                             v->v,
                             v->size,
                             (cgul_merge_sort__compare_t)f);

 out:
    return;
}

void
cgul_vector__shuffle(cgul_exception_t* cex,
                     cgul_vector_t v,
                     cgul_vector__next_in_range_t f,
                     void* prng)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(v);
    assert(f);

    /* Shuffle. */
    cgul_shuffle_pointers(cex,
                          v->v,
                          v->size,
                          (cgul_shuffle__next_in_range_t)f,
                          prng);

 out:
    return;
}

void
cgul_vector__clear(cgul_exception_t* cex,
                   cgul_vector_t v)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(v);

    v->size = 0;
    cgul_vector__decrease_capacity_if_necessary(cex, v);

 out:
    return;
}

void**
cgul_vector__get_array(cgul_exception_t* cex,
                       cgul_vector_t v)
{
    void** rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(v);

    rv = v->v;

 out:
    return rv;
}

void**
cgul_vector__take_array(cgul_exception_t* cex,
                        cgul_vector_t v)
{
    void** rv = NULL;
    cgul_vector_t tmp = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(v);

    /* Allocate. */
    tmp = cgul_vector__new(cex);
    if (*cex) {
        goto out;
    }

    /*
     * If you reach here, the rest of this function will complete
     * without throwing an exception.
     */

    /* Swap the newly allocated vector with v thereby resetting v. */
    cgul_vector__swap(cex, v, tmp);

    /* The client is taking ownership of what is now tmp->value. */
    rv = tmp->v;
    tmp->v = NULL;

 out:

    /* Clean up. */
    if (tmp) {
        cgul_vector__delete(tmp);
    }

    return rv;
}

size_t
cgul_vector__get_size(cgul_exception_t* cex,
                      cgul_vector_t v)
{
    size_t rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(v);

    rv = v->size;

 out:
    return rv;
}

void
cgul_vector__trim(cgul_exception_t* cex,
                  cgul_vector_t v)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(v);

    v->v = (void**)cgul_libc__realloc(cex,
                                      v->v,
                                      sizeof(void*) * v->size);
    if (*cex) {
        goto out;
    }

    v->capacity = v->size;

 out:
    return;
}

size_t
cgul_vector__get_minimum_capacity(cgul_exception_t* cex,
                                  cgul_vector_t v)
{
    size_t rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(v);

    /* Return the minimum capacity to the user. */
    rv = v->minimum_capacity;

 out:
    return rv;
}

void
cgul_vector__set_minimum_capacity(cgul_exception_t* cex,
                                  cgul_vector_t v,
                                  size_t minimum_capacity)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(v);

    /* Set the new minimum_capacity. */
    if (minimum_capacity == 0) {
        v->minimum_capacity = CGUL_VECTOR__DEFAULT_MINIMUM_CAPACITY;
    } else {
        v->minimum_capacity = minimum_capacity;
    }

    /* Realloc the underlying buffer (if necessary). */
    cgul_vector__set_capacity(cex, v, v->minimum_capacity);
    if (*cex) {
        goto out;
    }

 out:
    return;
}

void
cgul_vector__swap(cgul_exception_t* cex,
                  cgul_vector_t v1,
                  cgul_vector_t v2)
{
    struct cgul_vector tmp;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(v1);
    assert(v2);

    /* Nothing to swap if they point to the same objects. */
    if (v1 == v2) {
        goto out;
    }

    /* A shallow copy, by design, is exactly what is needed. */
    tmp = *v1;
    *v1 = *v2;
    *v2 = tmp;

 out:
    return;
}

void
cgul_vector__foldl(cgul_exception_t* cex,
                   cgul_vector_t v,
                   cgul_vector__fold_t f,
                   void* data)
{
    int f_rv = 0;
    size_t i = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(v);
    assert(f);

    /* Iterate over the nodes in the vector. */
    for (i = 0 ; i < v->size ; ++i) {
        /* Call the combining function. */
        f_rv = (*f)(cex, v->v[i], data);
        if (*cex || !f_rv) {
            goto out;
        }
    }

 out:
    return;
}

void
cgul_vector__foldr(cgul_exception_t* cex,
                   cgul_vector_t v,
                   cgul_vector__fold_t f,
                   void* data)
{
    int f_rv = 0;
    size_t i = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(v);
    assert(f);

    /* Iterate over the nodes in the vector. */
    for (i = v->size ; i > 0 ; --i) {
        /* Call the combining function. */
        f_rv = (*f)(cex, v->v[i - 1], data);
        if (*cex || !f_rv) {
            goto out;
        }
    }

 out:
    return;
}
