/**
 * This program generates BDF index files.  You can pass it the name
 * of multiple BDF files on the command line.  It will then append an
 * 'i' character to each BDF file name and write the index to that
 * file.
 *
 * @file
 * @author Paul Serice
 * @brief generate bdf index files
 */

#include "cgul.h"

int
main(int argc,
     char* argv[])
{
    int rv = 0;
    int i = 0;
    cgul_bdf_t bdf = NULL;
    cgul_string_t index_file_name = NULL;
    cgul_exception_t local = NULL;

    /* Verify the command line. */
    if (argc < 2) {
        fprintf(stderr,
                "Usage: main_bdf_genbdfi <font.bdf> [<font.bdf> ...]\n");
        goto out;
    }

    /* Allocate the string we use for generating the name of the index
     * files. */
    index_file_name = cgul_string__new(&local);
    if (local) {
        goto out;
    }

    /* Allocate an empty cgul_bdf object. */
    bdf = cgul_bdf__new(&local);
    if (local) {
        goto out;
    }

    /* Iterate over all the BDF fonts on the command line. */
    for (i = 1 ; i < argc ; ++i) {

        printf("Processing %s ... ", argv[i]);
        fflush(stdout);

        /* Load the BDF font.  It is important to not attempt specify
         * an index file when loading the font here.  This way all of
         * the glyphs will be loaded into memory and subsequently have
         * their offset information written to the index file.  There
         * would also a chicken-or-the-egg problem if you tried to use
         * an index file to generate an index file. */
        cgul_bdf__load_explicit(&local, bdf, argv[i], NULL);

        /* Generate the index file name by simply appending an 'i' to
         * the end of the BDF file name. */
        cgul_string__set_value(&local, index_file_name, argv[i]);
        cgul_string__append_char(&local, index_file_name, 'i');

        /* Write the index file. */
        cgul_bdf__write_index(
            &local,
            bdf,
            cgul_string__get_value(&local, index_file_name));

        /* Unload the font so we can reuse "bdf". */
        cgul_bdf__unload(&local, bdf);

        if (local) {
            goto out;
        }

        printf("Done.\n");
        fflush(stdout);

    }

 out:

    /* Catch exceptions. */
    if (local) {
        const char* msg = cgul_exception__get_message(local);
        fprintf(stderr, "*** Error: %s\n", msg);
        cgul_exception__delete(local);
        local = NULL;
        rv = 1;
    }

    /* Clean up. */
    if (bdf) {
        cgul_bdf__delete(bdf);
    }
    if (index_file_name) {
        cgul_string__delete(index_file_name);
    }

    return rv;
}
