#include <iostream>
#include <stdexcept>
#include "cgul_cxx.h"

using namespace std;

/**
 * This function prints the value stored in the node <code>n</code>.
 *
 * @param[in] l list
 * @param[in] n node
 * @param[in] data client data
 */
static int
print_value(cgul_list_cxx* l,
            cgul_list_node_cxx* n,
            void* data)
{
    printf("%d\n", *(int*)n->get_value());
    return 1;
}

/**
 * This function calls <code>free()</code> on the value stored in the
 * node <code>n</code>.
 *
 * @param[in] l list
 * @param[in] n node
 * @param[in] data client data
 */
static int
free_value(cgul_list_cxx* l,
           cgul_list_node_cxx* n,
           void* data)
{
    free(n->get_value());
    return 1;
}

/**
 * This function returns a list of integers in the half-open range
 * <code>[0, upper)</code>.  The caller is responsible for calling
 * <code>free()</code> on each element in the list and for calling
 * <code>delete()</code> on the list itself.
 *
 * @param[in] upper exclusive upper bound on the range
 * @return list holding integers from 0 to <code>upper</code>
 */
static cgul_list_cxx*
range(int upper)
{
    int i = 0;
    int* value = NULL;
    cgul_list_cxx* rv = NULL;

    try {

        /* Allocate the return value. */
        rv = new cgul_list_cxx;

        /* Iterate from 0 to upper. */
        for (i = 0 ; i < upper ; ++i) {

            /* Allocate a new int to hold this value. */
            value = (int*)malloc(sizeof(*value));
            if (!value) {
                throw runtime_error("out of memory");
            }

            /* Assign the value. */
            *value = i;

            /* Push the value on to the stack. */
            rv->push_back(value);

            /* The list now owns value. */
            value = NULL;
        }

    } catch (...) {
        /* Clean up. */
        if (value) {
            free(value);
        }
        if (rv) {
            rv->traverse(free_value, NULL);
            delete rv;
        }
        /* Rethrow. */
        throw;
    }

    return rv;
}

int
main(int argc,
     char* argv[])
{
    int rv = 0;
    cgul_list_cxx* l = NULL;
    cgul_list_node_cxx* first = NULL;
    cgul_list_node_cxx* last = NULL;

    try {

        /* Create a new list that holds a range of values. */
        l = range(5);

        /* Print the list. */
        l->traverse(print_value, NULL);

        cout << "=====" << endl;

        /* Print a range of the list. */
        first = l->get_front();
        if (first) {
            first = first->get_next();
        }
        last = l->get_back();
        if (last) {
            last = last->get_prev();
        }
        l->traverse_range(first, last, print_value, NULL);

    } catch (const exception& e) {
        cerr << "*** Error: " << e.what() << endl;
        rv = 1;
    } catch (const cgul_exception_cxx& e) {
        cerr << "*** Error: " << e.get_message() << endl;
        rv = 1;
    }

    /* Clean up. */
    if (l) {
        l->traverse(free_value, NULL);
        delete l;
    }

    return rv;
}
