#include <assert.h>
#include "cgul.h"

/*
 * NOTE:
 * ====
 *
 * A good way to test cgul_merge_sort_pointers() (which is the basic
 * engine for cgul_merge_sort()) is with main_trie.c which can
 * generate lots of random strings of various sizes and verifies
 * sorting by comparing the result of cgul_merge_sort_pointers() with 
 * the result of inserting the strings into cgul_trie.
 */

/*
 * When sorting an array of strings (as we do in this test program),
 * normally cgul_merge_sort_pointers() and strcmp() would be used
 * directly, but this test program is for testing cgul_merge_sort()
 * which is designed to create another level of indirection by
 * creating an array of pointers into the array we want to sort
 * (because it is faster to move pointers around than large data
 * structures).  This means we cannot use strcmp() directly.  Instead,
 * the function below is used to undo the extra layer of indirection
 * before calling strcmp().
 */
static int
mystrcmp(const char** s1,
         const char** s2)
{
    return strcmp(*s1, *s2);
}

static void
print_array(cgul_exception_t* cex,
            char** a,
            size_t a_count)
{
    unsigned long int i = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    for (i = 0 ; i < a_count ; ++i) {
        puts(a[i]);
    }

    puts("=====");

 out:
    return;
}

int
main(int argc,
     char* argv[])
{
    int rv = 0;
    size_t a_count = 0;
    char** a = NULL;
    cgul_vector_t v = NULL;
    cgul_exception_t local = NULL;

    /* Allocate. */
    v = cgul_vector__new(&local);
    if (local) {
        goto out;
    }

    /* Add unsorted strings. */
    cgul_vector__push_back(&local, v, "foo");
    cgul_vector__push_back(&local, v, "bar");
    cgul_vector__push_back(&local, v, "baz");
    cgul_vector__push_back(&local, v, "bar");
    cgul_vector__push_back(&local, v, "zzz");

    /* Normally, you would just use cgul_vector__sort() for sorting
     * the vector, but for testing we need to take ownership of the
     * array. */
    a_count = cgul_vector__get_size(&local, v);
    a = (char**)cgul_vector__take_array(&local, v);

    /* Print vector of strings. */
    print_array(&local, a, a_count);

    /* Sort. */
#if 1
    cgul_merge_sort(&local,
                    a,
                    a_count,
                    sizeof(*a),
                    (cgul_merge_sort__compare_t)mystrcmp);
#elif 0
    cgul_merge_sort_iteratively(&local,
                    a,
                    a_count,
                    sizeof(*a),
                    (cgul_merge_sort__compare_t)mystrcmp);
#elif 0
    cgul_merge_sort_pointers(&local,
                    (void**)a,
                    a_count,
                    (cgul_merge_sort__compare_t)strcmp);
#elif 0
    cgul_merge_sort_pointers_iteratively(&local,
                    (void**)a,
                    a_count,
                    (cgul_merge_sort__compare_t)strcmp);
#endif
    if (local) {
        goto out;
    }

    /* Print array of strings. */
    print_array(&local, a, a_count);

 out:

    /* Catch exceptions. */
    if (local) {
        const char* msg = cgul_exception__get_message(local);
        fprintf(stderr, "*** Error: %s\n", msg);
        cgul_exception__delete(local);
        local = NULL;
        rv = 1;
    }

    /* Clean up. */
    if (a) {
        free(a);
    }
    if (v) {
        cgul_vector__delete(v);
    }

    return rv;
}
