#include <iostream>
#include <stdexcept>
#include "cgul_cxx.h"

using namespace std;

/*
 * When sorting an array of strings (as we do in this test program),
 * normally cgul_merge_sort_pointers() and strcmp() would be used
 * directly, but this test program is for testing cgul_merge_sort()
 * which is designed to create another level of indirection by
 * creating an array of pointers into the array we want to sort
 * (because it is faster to move pointers around than large data
 * structures).  This means we cannot use strcmp() directly.  Instead,
 * the function below is used to undo the extra layer of indirection
 * before calling strcmp().
 */
static int
mystrcmp(const char** s1,
         const char** s2)
{
    return strcmp(*s1, *s2);
}

static void
print_array(char** a,
            size_t a_count)
{
    unsigned long int i = 0;

    for (i = 0 ; i < a_count ; ++i) {
        puts(a[i]);
    }

    puts("=====");
}

int
main(int argc,
     char* argv[])
{
    int rv = 0;
    size_t a_count = 0;
    char** a = NULL;
    cgul_vector_cxx v;

    try {

        /* Add unsorted strings. */
        v.push_back("foo");
        v.push_back("bar");
        v.push_back("baz");
        v.push_back("bar");
        v.push_back("zzz");

        /* Normally, you would just use v.sort() for sorting the
         * vector, but for testing we need to take ownership of the
         * array. */
        a_count = v.get_size();
        a = (char**)v.take_array();

        /* Print vector of strings. */
        print_array(a, a_count);

        /* Sort. */
#if 1
        cgul_merge_sort_cxx::sort(
            a,
            a_count,
            sizeof(*a),
            (cgul_merge_sort_cxx::compare_t)mystrcmp);
#elif 0
        cgul_merge_sort_cxx::sort_iteratively(
            a,
            a_count,
            sizeof(*a),
            (cgul_merge_sort_cxx::compare_t)mystrcmp);
#elif 0
        cgul_merge_sort_cxx::sort_pointers(
            (void**)a,
            a_count,
            (cgul_merge_sort_cxx::compare_t)strcmp);
#elif 0
        cgul_merge_sort_cxx::sort_pointers_iteratively(
            (void**)a,
            a_count,
            (cgul_merge_sort_cxx::compare_t)strcmp);
#endif

        /* Print array of strings. */
        print_array(a, a_count);

    } catch (const exception& e) {
        cerr << "*** Error: " << e.what() << endl;
        rv = 1;

    } catch (const cgul_exception_cxx& e) {
        cerr << "*** Error: " << e.get_message() << endl;
        rv = 1;
    }

    /* Clean up. */
    if (a) {
        free(a);
    }

    return rv;
}
