#include "cgul.h"

/**
 * This function prints the key/value pair stored in the node <code>n</code>.
 *
 * @param[in] cex c-style exception
 * @param[in] mm multimap
 * @param[in] n node
 * @param[in] data client data
 */
static int
print_pair(cgul_exception_t* cex,
           cgul_multimap_t mm,
           cgul_multimap_node_t n,
           void* data)
{
    int rv = 0;
    char* key = NULL;
    int* value = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    key = (char*)cgul_multimap_node__get_key(cex, n);
    value = (int*)cgul_multimap_node__get_value(cex, n);
    printf("key=%s  value=%d\n", key, *value);
    rv = 1;

 out:
    return rv;
}

/**
 * This function calls <code>free()</code> on the key and on the value
 * stored in the node <code>n</code>.
 *
 * @param[in] cex c-style exception
 * @param[in] mm multimap
 * @param[in] curr current node
 * @param[in] data client data
 */
static int
free_pair(cgul_exception_t* cex,
          cgul_multimap_t mm,
          cgul_multimap_node_t curr,
          void* data)
{
    int rv = 0;
    void* key_curr = NULL;
    void* key_next = NULL;
    cgul_multimap_node_t next = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* We always need to free the value. */
    free(cgul_multimap_node__get_value(cex, curr));

    /* We only free this key if no more values are associated with it. */
    next = cgul_multimap_node__get_next(cex, curr, 0);
    if (next) {
        key_next = (void*)cgul_multimap_node__get_key(cex, next);
    }
    key_curr = (void*)cgul_multimap_node__get_key(cex, curr);
    if ((next == NULL) || (key_curr != key_next)) {
        free(key_curr);
    }

    rv = 1;

 out:
    return rv;
}

/**
 * Initialize the multimap <code>mm</code> by loading key/value pairs.
 *
 * @param[in,out] cex c-style exception
 * @param[in] t tree
 */
static void
init_multimap(cgul_exception_t* cex,
              cgul_multimap_t mm)
{
    int i = 0;
    char* key1 = NULL;
    char* key2 = NULL;
    int* value1 = NULL;
    int* value2 = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(mm);

    for (i = 0 ; i < 5 ; ++i) {

        /* Allocate. */
        key1 = cgul_libc__calloc(cex, 1, sizeof(char) * 2);
        key2 = cgul_libc__calloc(cex, 1, sizeof(char) * 2);
        value1 = cgul_libc__calloc(cex, 1, sizeof(int));
        value2 = cgul_libc__calloc(cex, 1, sizeof(int));
        if (*cex) {
            goto out;
        }

        /* Assign values. */
        *key1 = 'a' + i;
        *key2 = 'a' + i;
        *value1 = i * 2;
        *value2 = i * 2 + 1;

        /* Insert key1 and value1 into multimap. */
        cgul_multimap__insert(cex, mm, key1, value1, NULL);
        if (*cex) {
            goto out;
        }

        /* The multimap now owns key1 and value1. */
        key1 = NULL;
        value1 = NULL;

        /* Insert value2 into multimap. */
        cgul_multimap__insert(cex, mm, key2, value2, NULL);
        if (*cex) {
            goto out;
        }

        /* Free key2 because it is not owned by the multimap, but do
         * not free value2 because it is owned by the multimap. */
        free(key2);
        key2 = NULL;
        value2 = NULL;
    }

 out:

    /* Clean up. */
    if (key1) {
        free(key1);
    }
    if (key2) {
        free(key2);
    }
    if (value1) {
        free(value1);
    }
    if (value2) {
        free(value2);
    }

    return;
}

int
main(int argc,
     char* argv[])
{
    int rv = 0;
    cgul_multimap_t mm = NULL;
    cgul_multimap_node_t first = NULL;
    cgul_multimap_node_t last = NULL;
    cgul_exception_t local = NULL;

    /* Create a new multimap that holds a range of values. */
    mm = cgul_multimap__new(&local, (cgul_multimap__compare_t)strcmp);
    if (local) {
        goto out;
    }

    /* Initialize multimap. */
    init_multimap(&local, mm);
    if (local) {
        goto out;
    }

    /* Print the multimap. */
    cgul_multimap__traverse(&local, mm, 0, print_pair, NULL);

    puts("=====");

    /* Print a range of the multimap. */
    first = cgul_multimap__get_front(&local, mm, 0);
    if (first) {
        first = cgul_multimap_node__get_next(&local, first, 0);
    }
    last = cgul_multimap__get_back(&local, mm, 0);
    if (last) {
        last = cgul_multimap_node__get_prev(&local, last, 0);
    }
    cgul_multimap__traverse_range(&local, mm, first, last, 0, print_pair, NULL);

 out:

    /* Catch exceptions. */
    if (local) {
        const char* msg = cgul_exception__get_message(local);
        fprintf(stderr, "*** Error: %s\n", msg);
        cgul_exception__delete(local);
        local = NULL;
        rv = 1;
    }

    /* Clean up. */
    if (mm) {
        cgul_multimap__traverse(&local, mm, 0, free_pair, NULL);
        cgul_multimap__delete(mm);
    }

    return rv;
}
