/**
 * Test whether cgul_radix_sort() is stable.  An easy way to perform
 * an overnight test is with the following one-liner: <p>
 *
 * <pre>
 *     while ./main_radix_sort_stable ; do : ; done
 * </pre>
 */

#include "cgul.h"

/**
 * This variable controls the range of the pseudo-random number
 * generator.  The numbers generated by the pseudo-random number
 * generator control how many numbers to sort as well as the actual
 * numbers that get sorted.
 */
static unsigned long int PRNG_MAX_RANGE = 100000000;

/**
 * Whether to seed the PRNG with a constant value.
 */
#define CONST_SEED 0

/**
 * If <code>CONST_SEED</code> is not <code>0</code>, this is the seed
 * value for the PRNG.
 */
#define CONST_SEED_VALUE 0

/**
 * Whether to print the unsorted and sorted vectors.
 */
#define ENABLE_PRINT 0

/**
 * Structure used to test whether <code>cgul_radix_sort()</code> is
 * stable.  The <code>x</code> value is set to a random number, and
 * <code>y</code> is set to monotonically increasing number such that
 * points inserted early into the unsorted array always have
 * <code>y</code> values less than points sorter later.  This creates
 * a total ordering that allows for checking that the start is stable.
 */
struct point {
    unsigned long int x;
    unsigned long int y;
};

/**
 * Enumerate each point in the array.  This allows
 * <code>cgul_radix_sort()</code> to sort the points in order of the
 * points' enumerations.  Here, the enumeration is just the
 * x-coordinate.  This way multiple points can sort to the same value.
 * Then, we can check the y-coordinate which should be monotonically
 * increasing for a stable sort.
 */
static unsigned long int
main__enumerate(const void* element)
{
    struct point* point = (struct point*)element;
    return point->x;
}

#if ENABLE_PRINT
static void
print_points(cgul_exception_t* cex,
             struct point* points,
             size_t points_size)
{
    size_t i = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Iterate over each element in the array. */
    for (i = 0 ; i < points_size ; ++i) {
        printf("(%6lu, %6lu)\n", points[i].x, points[i].y);
    }

    puts("=====");

 out:
    return;
}
#endif

static void
initialize_points_array(cgul_exception_t* cex,
                        struct point* points,
                        size_t points_size,
                        cgul_random_xorwow32_t prng,
                        cgul_uint32_t rmax)
{
    unsigned long int i = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Create each point. */
    for (i = 0 ; i < points_size ; ++i) {

        /* Get the next random number. */
        points[i].x = cgul_random_xorwow32__next_in_range(cex, prng, rmax);
        if (*cex) {
            goto out;
        }

        /* Track the order in which the points are inserted so we can
         * check later to see if the sort is stable. */
        points[i].y = i;
    }

 out:
    return;
}

static int
is_sorted_and_stable(cgul_exception_t* cex,
                     struct point* points,
                     size_t points_size)
{
    int rv = 0;
    size_t i = 0;
    size_t same_x = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* If the array is empty, it is sorted and stable. */
    if (points_size == 0) {
        rv = 1;
        goto out;
    }

    /* Iterate over the points. */
    for (i = 1 ; i < points_size ; ++i) {

        /* The x-value is a randomly generated number.  It is used to
         * test whether the vector is sorted.  If the current value of x
         * is less than the previous value of x, the vector is not
         * sorted. */
        if (points[i].x < points[i-1].x) {
            goto out;
        }

        /* The y-value is the order the nodes were inserted.  It is
         * used to test for stability.  Stability can only be tested
         * when points[i-1].x and points[i].x have the same value. */
        if (points[i-1].x == points[i].x) {

            /* Keep track of the number of points that sorted the same. */
            ++same_x;

            /* Check for stability. */
            if (points[i].y < points[i-1].y) {
                goto out;
            }
        }
    }

    /* Print the number of points with the same ordering. */
    printf("    same_x = %9lu\n", (unsigned long)same_x);
    fflush(stdout);

    /* Success. */
    rv = 1;

 out:
    return rv;
}

int
main(int argc,
     char* argv[])
{
    int rv = 0;
    size_t points_size = 0;
    struct point* points = NULL;
    cgul_uint32_t rmax = PRNG_MAX_RANGE;
    cgul_random_xorwow32_t prng = NULL;
    cgul_exception_t local = NULL;

    /* Create the PRNG. */
#if CONST_SEED
    prng = cgul_random_xorwow32__new_from_seed(&local, CONST_SEED_VALUE);
#else
    prng = cgul_random_xorwow32__new(&local);
#endif
    if (local) {
        goto out;
    }

    /* Determine how many elements the array to sort can hold. */
    points_size =
        (size_t)cgul_random_xorwow32__next_in_range(&local, prng, rmax);
    if (local) {
        goto out;
    }

    /* Print the size of the array. */
    printf("array size = %9lu\n", (unsigned long)points_size);
    fflush(stdout);

    /* Allocate the array. */
    points = cgul_libc__calloc(&local, points_size, sizeof(*points));
    if (local) {
        goto out;
    }

    /* Initialize the "points" array with random points. */
    initialize_points_array(&local, points, points_size, prng, rmax);
    if (local) {
        goto out;
    }

#if ENABLE_PRINT
    /* Print the points. */
    print_points(&local, points, points_size);
#endif

    /* Sort. */
    cgul_radix_sort(&local,
                    points,
                    points_size,
                    sizeof(*points),   /* element_size */
                    1,                 /* ascending_order */
                    main__enumerate,   /* enumeration function */
                    0);                /* are_enumerations_signed */
    if (local) {
    goto out;
}

#if ENABLE_PRINT
    /* Print the points. */
    print_points(&local, points, points_size);
#endif

    /* Test whether the vector is in fact sorted and stable. */
    if (!is_sorted_and_stable(&local, points, points_size)) {
        local = cgul_exception__new("Not sorted or not stable.");
        goto out;
    }

 out:

    /* Handle exceptions. */
    if (local) {
        const char* msg = cgul_exception__get_message(local);
        fprintf(stderr, "*** Error: %s\n", msg);
        cgul_exception__delete(local);
        local = NULL;
        rv = 1;
    }

    /* Clean up. */
    if (points) {
        free(points);
    }
    if (prng) {
        cgul_random_xorwow32__delete(prng);
    }

    return rv;
}
