/*
 * Test the following methods using randomly generated keys:
 *
 *     cgul_rbtree_cxx::find_at()
 *     cgul_rbtree_cxx::find_rank()
 *     cgul_rbtree_cxx::remove_at()
 */

#include "cgul_cxx.h"
#include <iostream>
#include <stdexcept>

using namespace std;

/**
 * Minimum number of keys to generate.
 */
#define KEY_COUNT_MIN (0)

/**
 * Maximum number of keys to generate.
 */
#define KEY_COUNT_MAX (1000000)

/**
 * Return less than zero, zero, or greater than zero if
 * <code>*n1</code> is less than, equal to, or greater than
 * <code>*n2</code>.
 *
 * @param[in] n1 first number
 * @param[in] n2 second number
 * @return ordering of the first number relative to the second number
 */
static int
main__compare_uint32(cgul_uint32_t* n1,
                     cgul_uint32_t* n2)
{
    int rv = 0;

    if (*n1 < *n2) {
        rv = -1;
    } else if (*n1 > *n2) {
        rv = 1;
    }

    return rv;
}

/**
 * Insert between <code>key_count_min</code> (inclusive) and
 * <code>key_count_max</code> (inclusive) randomly generated keys into
 * the red-black tree <code>t</code> and push a copy of each key in
 * the order generated onto the <code>keys</code> vector (which can be
 * used later for verification purposes).  The key are randomly
 * generated by the PRNG <code>prng</code>.  The user is responsible
 * for calling <code>free()</code> on the keys inserted into the tree.
 * If an error occurs, an exception is thrown.
 *
 * @param[in] t <code>cgul_rbtree</code> instance
 * @param[in] prng PRNG
 * @param[in] keys keys
 * @param[in] key_count_min minimum key count
 * @param[in] key_count_max maximum key count
 */
static void
main__insert_keys(cgul_rbtree_cxx& t,
                  cgul_vector_cxx& keys,
                  cgul_random_lfg32_cxx& prng,
                  cgul_uint32_t key_count_min,
                  cgul_uint32_t key_count_max)
{
    int insert_rv = 0;
    cgul_uint32_t i = 0;
    cgul_uint32_t key_count = 0;
    cgul_uint32_t* key = NULL;

    try {

        /* Determine the number of keys to generate. */
        key_count = prng.next_in_range(key_count_max - key_count_min + 1);
        key_count += key_count_min;

#if 1
        printf("key_count = %lu\n", (unsigned long)key_count);
#endif

        /* Generate and insert the keys. */
        for (i = 0 ; i < key_count ; ++i) {

            /* Allocate space for a new key. */
            key = (cgul_uint32_t*)cgul_libc_cxx::calloc(1, sizeof(*key));

            /* Randomly generate the next key. */
            *key = prng.next();

            /* Try to insert the key which could fail if a key repeats. */
            insert_rv = t.insert(key, NULL);
            if (!insert_rv) {
                free(key);
                key = NULL;
                continue;
            }

            /* Push the key onto the verification vector. */
            keys.push_back(key);

            /* The red-black tree now owns the key. */
            key = NULL;

        }

    } catch (...) {
        /* Clean up. */
        if (key) {
            free(key);
        }
        throw;
    }

    /* Clean up. */
    if (key) {
        free(key);
    }
}

/**
 * Test <code>cgul_rbtree__find_at()</code> by verifying the that the
 * keys in the <code>keys</code> vector exist in the red-black tree
 * <code>t</code> in the same order.  If an error occurs, an exception
 * is thrown.
 *
 * @param[in] t <code>cgul_rbtree</code> instance
 * @param[in] keys keys
 */
static void
main__test_find_at(cgul_rbtree_cxx& t,
                   cgul_vector_cxx& keys)
{
    unsigned long int i = 0;
    cgul_uint32_t* key = NULL;
    cgul_uint32_t** arr = NULL;
    unsigned long int arr_size = 0;
    cgul_rbtree_node_cxx* node = NULL;

    /* Unpack the keys vector. */
    arr = (cgul_uint32_t**)keys.get_array();
    arr_size = keys.get_size();

    /* Verify that each key in "keys" exists in "t" and in the same order. */
    for (i = 0 ; i < arr_size ; ++i) {

        /* Find the node by index. */
        node = t.find_at(i);
        if (!node) {
            throw cgul_exception_cxx(
                "main__test_find_at: Missing key: %lu", i);
        }

        /* Verify the keys match. */
        key = (cgul_uint32_t*)node->get_key();
        if (*key != *arr[i]) {
            throw cgul_exception_cxx(
                "main__test_find_at: "
                "Invalid key at index %lu: (expecting, actual) = (%lu, %lu)",
                i, (unsigned long int)*arr[i], (unsigned long int)*key);
        }
    }

    /* Verify lookup misses return NULL. */
    node = t.find_at(arr_size);
    if (node) {
        key = (cgul_uint32_t*)node->get_key();
        throw cgul_exception_cxx(
            "main__test_find_at: "
            "Unexpectedly found key for out-of-bounds index: "
            "(index, key) = (%lu, %lu)",
            arr_size, (unsigned long int)*key);
    }

}

/**
 * Test <code>cgul_rbtree__find_rank()</code> by verifying the index
 * of each key in the <code>keys</code> vector equals the rank of that
 * same key in the red-black tree <code>t</code>.  If an error occurs,
 * an exception is thrown.
 *
 * @param[in] t <code>cgul_rbtree</code> instance
 * @param[in] keys keys
 */
static void
main__test_find_rank(cgul_rbtree_cxx& t,
                     cgul_vector_cxx& keys)
{
    int rank_rv = 0;
    unsigned long int i = 0;
    unsigned long int rank = 0;
    cgul_uint32_t** arr = NULL;
    unsigned long int arr_size = 0;

    /* Unpack the keys vector. */
    arr = (cgul_uint32_t**)keys.get_array();
    arr_size = keys.get_size();

    /* Verify that each key in "keys" exists in "t" and in the same order. */
    for (i = 0 ; i < arr_size ; ++i) {

        /* Find the rank of the key. */
        rank_rv = t.find_rank(arr[i], &rank);
        if (!rank_rv) {
            throw cgul_exception_cxx(
                "main__test_find_rank: Missing key: %lu", i);
        }

        /* Verify the index of the key match the rank. */
        if (rank != i) {
            throw cgul_exception_cxx(
                "main__test_find_at: Invalid rank for key: "
                "(key, expecting, actual) = (%lu, %lu, %lu)",
                (unsigned long int)*arr[i], i, rank);
        }
    }
}

/**
 * Randomly remove all keys from the rbtree <code>rbtree</code> using
 * <code>cgul_rbtree__remove_at()</code>.  The order in which keys are
 * removed is determined by the PRNG <code>prng</code>.  If an error
 * occurs, an exception is thrown.
 *
 * @param[in] rbtree <code>cgul_rbtree</code> instance
 * @param[in] PRNG prng
 */
static void
main__test_remove_at(cgul_rbtree_cxx& t,
                     cgul_random_lfg32_cxx& prng)
{
    int remove_rv = 0;
    cgul_uint32_t index = 0;
    cgul_uint32_t* key = NULL;

    while (!t.is_empty()) {

        /* Generate the index of the next key to remove. */
        index = prng.next_in_range(t.get_size());

        /* Remove this key from the rbtree. */
        remove_rv = t.remove_at(index, (void**)&key, NULL);
        if (!remove_rv) {
            throw cgul_exception_cxx(
                "main__test_remove_at: "
                "Unable to remove key at index: %lu",
                (unsigned long int)index);
        }

        /* Free the key. */
        free(key);
        key = NULL;

    }

}

int
main(int argc,
     char* argv[])
{
    int rv = 0;
    cgul_uint32_t seed = 0;
    cgul_rbtree_cxx t((cgul_rbtree_cxx::compare_t)main__compare_uint32,
                      1 /* with_indexing */);
    cgul_vector_cxx keys;
    cgul_random_lfg32_cxx prng;

    try {

#if 1
        /* Use the time as the seed. */
        seed = (cgul_uint32_t)time(NULL);
#else
        /* Use specific seed value. */
        seed = 1425126242;
        printf("===== REUSING SEED!!!\n");
#endif
        printf("===== seed: %lu\n", (unsigned long)seed);

        /* Insert randomly generated numeric keys. */
        main__insert_keys(t, keys, prng, KEY_COUNT_MIN, KEY_COUNT_MAX);

        /* Sort the keys vector. */
        keys.sort((cgul_vector__compare_t)main__compare_uint32);

        /* Test cgul_rbtree__find_at(). */
        main__test_find_at(t, keys);

        /* Test cgul_rbtree__find_rank(). */
        main__test_find_rank(t, keys);

        /* Test cgul_rbtree_cxx::remove_at().  (This removes all the
         * keys from the rbtree in random order.) */
        main__test_remove_at(t, prng);

    } catch (const exception& e) {
        cerr << "*** Error: " << e.what() << endl;
        rv = 1;
    } catch (const cgul_exception_cxx& e) {
        cerr << "*** Error: " << e.get_message() << endl;
        rv = 1;
    }

    return rv;
}
