/*
 * Test the following methods using randomly generated keys:
 *
 *     cgul_rbtree__find_floor()
 *     cgul_rbtree__find_ceiling()
 */

#include "cgul.h"

/**
 * Minimum number of tests to execute.
 */
#define TEST_COUNT_MIN (0)

/**
 * Maximum number of tests to execute.
 */
#define TEST_COUNT_MAX (100000)

/**
 * Minimum number of keys to generate.
 */
#define KEY_COUNT_MIN (0)

/**
 * Maximum number of keys to generate.
 */
#define KEY_COUNT_MAX (1000000)

/**
 * Return less than zero, zero, or greater than zero if
 * <code>*n1</code> is less than, equal to, or greater than
 * <code>*n2</code>.
 *
 * @param[in] n1 first number
 * @param[in] n2 second number
 * @return ordering of the first number relative to the second number
 */
static int
main__compare_uint32(cgul_uint32_t* n1,
                     cgul_uint32_t* n2)
{
    int rv = 0;

    if (*n1 < *n2) {
        rv = -1;
    } else if (*n1 > *n2) {
        rv = 1;
    }

    return rv;
}

/**
 * Insert between <code>key_count_min</code> (inclusive) and
 * <code>key_count_max</code> (inclusive) randomly generated keys into
 * the red-black tree <code>t</code> and push a copy of each key in
 * the order generated onto the <code>keys</code> vector (which can be
 * used later for verification purposes).  The key are randomly
 * generated by the PRNG <code>prng</code>.  The user is responsible
 * for calling <code>free()</code> on the keys inserted into the tree.
 * If an error occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] t <code>cgul_rbtree</code> instance
 * @param[in] prng PRNG
 * @param[in] keys keys
 * @param[in] key_count_min minimum key count
 * @param[in] key_count_max maximum key count
 */
static void
main__insert_keys(cgul_exception_t* cex,
                  cgul_rbtree_t t,
                  cgul_vector_t keys,
                  cgul_random_lfg32_t prng,
                  cgul_uint32_t key_count_min,
                  cgul_uint32_t key_count_max)
{
    int insert_rv = 0;
    cgul_uint32_t i = 0;
    cgul_uint32_t key_count = 0;
    cgul_uint32_t* key = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Determine the number of keys to generate. */
    key_count = cgul_random_lfg32__next_in_range(
        cex, prng, key_count_max - key_count_min + 1);
    if (*cex) {
        goto out;
    }
    key_count += key_count_min;

#if 1
    printf("key_count = %lu\n", (unsigned long)key_count);
#endif

    /* Generate and insert the keys. */
    for (i = 0 ; i < key_count ; ++i) {

        /* Allocate space for a new key. */
        key = (cgul_uint32_t*)cgul_libc__calloc(cex, 1, sizeof(*key));
        if (*cex) {
            goto out;
        }

        /* Randomly generate the next key. */
        *key = cgul_random_lfg32__next(cex, prng);
        if (*cex) {
            goto out;
        }

        /* Try to insert the key which could fail if a key repeats. */
        insert_rv = cgul_rbtree__insert(cex, t, key, NULL);
        if (*cex) {
            goto out;
        }
        if (!insert_rv) {
            free(key);
            key = NULL;
            continue;
        }

        /* Push the key onto the verification vector. */
        cgul_vector__push_back(cex, keys, key);
        if (*cex) {
            goto out;
        }

        /* The red-black tree now owns the key. */
        key = NULL;

    }

 out:

    /* Clean up. */
    if (key) {
        free(key);
    }

    return;
}

/**
 * Test the ability of <code>cgul_rbtree__find_floor()</code> to find
 * nodes when the search key is an exact match for one of the keys in
 * the red-black tree <code>t</code>.  The <code>keys</code> vector is
 * used to verify results.  If an error occurs, an exception is
 * thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] t <code>cgul_rbtree</code> instance
 * @param[in] keys keys
 */
static void
main__test_find_floor__exact(cgul_exception_t* cex,
                             cgul_rbtree_t t,
                             cgul_vector_t keys)
{
    unsigned long int i = 0;
    cgul_uint32_t* key = NULL;
    cgul_uint32_t** arr = NULL;
    unsigned long int arr_size = 0;
    cgul_rbtree_node_t node = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(t);
    assert(keys);

    /* Unpack the keys vector. */
    arr = (cgul_uint32_t**)cgul_vector__get_array(cex, keys);
    arr_size = cgul_vector__get_size(cex, keys);

    /* Verify the floor function returns all the exact matches. */
    for (i = 0 ; i < arr_size ; ++i) {

        /* Find the floor. */
        node = cgul_rbtree__find_floor(cex, t, arr[i]);
        if (!node) {
            *cex = cgul_exception__new(
                "main__test_find_floor__exact: Missing key: %lu",
                (unsigned long int)*arr[i]);
            goto out;
        }

        /* Verify the keys match. */
        key = (cgul_uint32_t*)cgul_rbtree_node__get_key(cex, node);
        if (*key != *arr[i]) {
            *cex = cgul_exception__new(
                "main__test_find_floor__exact: "
                "Unexpected floor: (expected, actual) = (%lu, %lu)",
                (unsigned long int)*arr[i], (unsigned long int)*key);
            goto out;
        }
    }

 out:
    return;
}

/**
 * Test the ability of <code>cgul_rbtree__find_floor()</code> to find
 * nodes when the search key is an exact match for one of the keys in
 * the red-black tree <code>t</code>.  The <code>keys</code> vector is
 * used to verify results.  The PRNG <code>prng</code> is used as a
 * source of randomness.  Between <code>test_count_min</code>
 * (inclusive) and <code>test_count_max</code> (inclusive) are
 * executed.  If an error occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] t <code>cgul_rbtree</code> instance
 * @param[in] keys keys
 * @param[in] prng PRNG
 * @param[in] test_count_min minimum number of tests to execute
 * @param[in] test_count_max maximum number of tests to execute
 */
static void
main__test_find_floor__random(cgul_exception_t* cex,
                              cgul_rbtree_t t,
                              cgul_vector_t keys,
                              cgul_random_lfg32_t prng,
                              cgul_uint32_t test_count_min,
                              cgul_uint32_t test_count_max)
{
    cgul_uint32_t i = 0;
    cgul_uint32_t test_count = 0;
    cgul_uint32_t search_key = 0;
    cgul_uint32_t* expected = 0;
    cgul_uint32_t* actual = 0;
    cgul_uint32_t** arr = NULL;
    unsigned long int arr_size = 0;
    cgul_rbtree_node_t node = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(t);
    assert(keys);

    /* Unpack the keys vector. */
    arr = (cgul_uint32_t**)cgul_vector__get_array(cex, keys);
    arr_size = cgul_vector__get_size(cex, keys);

    /* Determine the number of tests to execute. */
    test_count = cgul_random_lfg32__next_in_range(
        cex, prng, test_count_max - test_count_min + 1);
    if (*cex) {
        goto out;
    }
    test_count += test_count_min;

    /* Verify the floor function returns all the exact matches. */
    for (i = 0 ; i < test_count ; ++i) {

        /* Randomly generate the key. */
        search_key = cgul_random_lfg32__next(cex, prng);

        /* Find the expected floor. */
        expected = cgul_binary_search_floor_pointers(
            cex,
            (const void*)&search_key,
            (const void**)arr,
            arr_size,
            (cgul_binary_search__compare_t)main__compare_uint32);

        /* Find the actual floor. */
        node = cgul_rbtree__find_floor(cex, t, &search_key);
        if (!node || !expected) {
            if (node || expected) {
                *cex = cgul_exception__new(
                    "main__test_find_floor__random: "
                    "Unexpected floor: (expected, actual) = (%p, %p)",
                    expected, node);
                goto out;
            }
            continue;
        }

        /* Get the floor that was actually found. */
        actual = (cgul_uint32_t*)cgul_rbtree_node__get_key(cex, node);

#if 0
        printf("--> floor: "
               "(search_key, expected, actual) = (%lu, %lu, %lu)\n",
                (unsigned long int)search_key,
                (unsigned long int)*expected,
                (unsigned long int)*actual);
#endif

        /* Verify the expected and actual floors match. */
        if (*actual != *expected) {
            *cex = cgul_exception__new(
                "main__test_find_floor__random: "
                "Unexpected floor: (expected, actual) = (%lu, %lu)",
                (unsigned long int)*expected, (unsigned long int)*actual);
            goto out;
        }
    }

 out:
    return;
}

/**
 * Test <code>cgul_rbtree__find_floor()</code> using the
 * <code>keys</code> vector to verify results.  When randomly testing
 * the floor function, between <code>test_count_min</code> (inclusive)
 * and <code>test_count_max</code> (inclusive) are executed.  If an
 * error occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] t <code>cgul_rbtree</code> instance
 * @param[in] keys keys
 * @param[in] prng PRNG
 * @param[in] test_count_min minimum number of tests to execute
 * @param[in] test_count_max maximum number of tests to execute
 */
static void
main__test_find_floor(cgul_exception_t* cex,
                      cgul_rbtree_t t,
                      cgul_vector_t keys,
                      cgul_random_lfg32_t prng,
                      cgul_uint32_t test_count_min,
                      cgul_uint32_t test_count_max)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Test keys that are exact matches. */
    main__test_find_floor__exact(cex, t, keys);
    if (*cex) {
        goto out;
    }

    /* Test randomly generated keys. */
    main__test_find_floor__random(
        cex, t, keys, prng, test_count_min, test_count_max);
    if (*cex) {
        goto out;
    }

 out:
    return;
}

/**
 * Test the ability of <code>cgul_rbtree__find_ceiling()</code> to
 * find nodes when the search key is an exact match for one of the
 * keys in the red-black tree <code>t</code>.  The <code>keys</code>
 * vector is used to verify results.  If an error occurs, an exception
 * is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] t <code>cgul_rbtree</code> instance
 * @param[in] keys keys
 */
static void
main__test_find_ceiling__exact(cgul_exception_t* cex,
                               cgul_rbtree_t t,
                               cgul_vector_t keys)
{
    unsigned long int i = 0;
    cgul_uint32_t* key = NULL;
    cgul_uint32_t** arr = NULL;
    unsigned long int arr_size = 0;
    cgul_rbtree_node_t node = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(t);
    assert(keys);

    /* Unpack the keys vector. */
    arr = (cgul_uint32_t**)cgul_vector__get_array(cex, keys);
    arr_size = cgul_vector__get_size(cex, keys);

    /* Verify the ceiling function returns all the exact matches. */
    for (i = 0 ; i < arr_size ; ++i) {

        /* Find the ceiling. */
        node = cgul_rbtree__find_ceiling(cex, t, arr[i]);
        if (!node) {
            *cex = cgul_exception__new(
                "main__test_find_ceiling__exact: Missing key: %lu",
                (unsigned long int)*arr[i]);
            goto out;
        }

        /* Verify the keys match. */
        key = (cgul_uint32_t*)cgul_rbtree_node__get_key(cex, node);
        if (*key != *arr[i]) {
            *cex = cgul_exception__new(
                "main__test_find_ceiling__exact: "
                "Unexpected ceiling: (expected, actual) = (%lu, %lu)",
                (unsigned long int)*arr[i], (unsigned long int)*key);
            goto out;
        }
    }

 out:
    return;
}

/**
 * Test the ability of <code>cgul_rbtree__find_ceiling()</code> to
 * find nodes when the search key is an exact match for one of the
 * keys in the red-black tree <code>t</code>.  The <code>keys</code>
 * vector is used to verify results.  The PRNG <code>prng</code> is
 * used as a source of randomness.  Between
 * <code>test_count_min</code> (inclusive) and
 * <code>test_count_max</code> (inclusive) are executed.  If an error
 * occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] t <code>cgul_rbtree</code> instance
 * @param[in] keys keys
 * @param[in] prng PRNG
 * @param[in] test_count_min minimum number of tests to execute
 * @param[in] test_count_max maximum number of tests to execute
 */
static void
main__test_find_ceiling__random(cgul_exception_t* cex,
                                cgul_rbtree_t t,
                                cgul_vector_t keys,
                                cgul_random_lfg32_t prng,
                                cgul_uint32_t test_count_min,
                                cgul_uint32_t test_count_max)
{
    cgul_uint32_t i = 0;
    cgul_uint32_t test_count = 0;
    cgul_uint32_t search_key = 0;
    cgul_uint32_t* expected = 0;
    cgul_uint32_t* actual = 0;
    cgul_uint32_t** arr = NULL;
    unsigned long int arr_size = 0;
    cgul_rbtree_node_t node = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(t);
    assert(keys);

    /* Unpack the keys vector. */
    arr = (cgul_uint32_t**)cgul_vector__get_array(cex, keys);
    arr_size = cgul_vector__get_size(cex, keys);

    /* Determine the number of tests to execute. */
    test_count = cgul_random_lfg32__next_in_range(
        cex, prng, test_count_max - test_count_min + 1);
    if (*cex) {
        goto out;
    }
    test_count += test_count_min;

    /* Verify the ceiling function returns all the exact matches. */
    for (i = 0 ; i < test_count ; ++i) {

        /* Randomly generate the key. */
        search_key = cgul_random_lfg32__next(cex, prng);

        /* Find the expected ceiling. */
        expected = cgul_binary_search_ceiling_pointers(
            cex,
            (const void*)&search_key,
            (const void**)arr,
            arr_size,
            (cgul_binary_search__compare_t)main__compare_uint32);

        /* Find the actual ceiling. */
        node = cgul_rbtree__find_ceiling(cex, t, &search_key);
        if (!node || !expected) {
            if (node || expected) {
                *cex = cgul_exception__new(
                    "main__test_find_ceiling__random: "
                    "Unexpected ceiling: (expected, actual) = (%p, %p)",
                    expected, node);
                goto out;
            }
            continue;
        }

        /* Get the ceiling that was actually found. */
        actual = (cgul_uint32_t*)cgul_rbtree_node__get_key(cex, node);

#if 0
        printf("--> ceiling: "
               "(search_key, expected, actual) = (%lu, %lu, %lu)\n",
                (unsigned long int)search_key,
                (unsigned long int)*expected,
                (unsigned long int)*actual);
#endif

        /* Verify the expected and actual ceilings match. */
        if (*actual != *expected) {
            *cex = cgul_exception__new(
                "main__test_find_ceiling__random: Unexpected ceiling: "
                "(search_key, expected, actual) = (%lu, %lu, %lu)",
                (unsigned long int)search_key,
                (unsigned long int)*expected,
                (unsigned long int)*actual);
            goto out;
        }
    }

 out:
    return;
}

/**
 * Test <code>cgul_rbtree__find_ceiling()</code> using the
 * <code>keys</code> vector to verify results.  When randomly testing
 * the ceiling function, between <code>test_count_min</code>
 * (inclusive) and <code>test_count_max</code> (inclusive) are
 * executed.  If an error occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] t <code>cgul_rbtree</code> instance
 * @param[in] keys keys
 * @param[in] prng PRNG
 * @param[in] test_count_min minimum number of tests to execute
 * @param[in] test_count_max maximum number of tests to execute
 */
static void
main__test_find_ceiling(cgul_exception_t* cex,
                        cgul_rbtree_t t,
                        cgul_vector_t keys,
                        cgul_random_lfg32_t prng,
                        cgul_uint32_t test_count_min,
                        cgul_uint32_t test_count_max)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Test keys that are exact matches. */
    main__test_find_ceiling__exact(cex, t, keys);
    if (*cex) {
        goto out;
    }

    /* Test randomly generated keys. */
    main__test_find_ceiling__random(
        cex, t, keys, prng, test_count_min, test_count_max);
    if (*cex) {
        goto out;
    }

 out:
    return;
}

int
main(int argc,
     char* argv[])
{
    int rv = 0;
    cgul_uint32_t seed = 0;
    cgul_rbtree_t t = NULL;
    cgul_vector_t keys = NULL;
    cgul_random_lfg32_t prng = NULL;
    cgul_exception_t local = NULL;

#if 1
    /* Use the time as the seed. */
    seed = (cgul_uint32_t)time(NULL);
#else
    /* Use specific seed value. */
    seed = 1425126242;
    printf("===== REUSING SEED!!!\n");
#endif
    printf("===== seed: %lu\n", (unsigned long)seed);

    /* Allocate. */
    keys = cgul_vector__new(&local);
    prng = cgul_random_lfg32__new_from_seed(&local, seed);
    t = cgul_rbtree__new_with_indexing(
            &local,
            (cgul_rbtree__compare_t)main__compare_uint32);
    if (local) {
        goto out;
    }

    /* Insert randomly generated numeric keys. */
    main__insert_keys(&local, t, keys, prng, KEY_COUNT_MIN, KEY_COUNT_MAX);
    if (local) {
        goto out;
    }

    /* Sort the keys vector. */
    cgul_vector__sort(
        &local, keys, (cgul_vector__compare_t)main__compare_uint32);

    /* Test cgul_rbtree__find_floor(). */
    main__test_find_floor(
        &local, t, keys, prng, TEST_COUNT_MIN, TEST_COUNT_MAX);
    if (local) {
        goto out;
    }

    /* Test cgul_rbtree__find_ceiling(). */
    main__test_find_ceiling(
        &local, t, keys, prng, TEST_COUNT_MIN, TEST_COUNT_MAX);
    if (local) {
        goto out;
    }

 out:

    /* Handle exceptions. */
    if (local) {
        const char* msg = cgul_exception__get_message(local);
        fprintf(stderr, "*** Error: %s\n", msg);
        fprintf(stderr, "***  seed: %lu\n", (unsigned long)seed);
        cgul_exception__delete(local);
        local = NULL;
        rv = 1;
    }

    /* Clean up. */
    if (t) {
        cgul_rbtree__free_keys(t);
        cgul_rbtree__delete(t);
    }
    if (prng) {
        cgul_random_lfg32__delete(prng);
    }
    if (keys) {
        cgul_vector__delete(keys);
    }

    return rv;
}
