#include "cgul.h"

/**
 * This function prints the key/value pair stored in the node <code>n</code>.
 *
 * @param[in] cex c-style exception
 * @param[in] t tree
 * @param[in] n node
 * @param[in] data client data
 */
static int
print_pair(cgul_exception_t* cex,
           cgul_rbtree_t t,
           cgul_rbtree_node_t n,
           void* data)
{
    int rv = 0;
    char* key = NULL;
    int* value = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    key = (char*)cgul_rbtree_node__get_key(cex, n);
    value = (int*)cgul_rbtree_node__get_value(cex, n);
    printf("key=%s  value=%d\n", key, *value);
    rv = 1;

 out:
    return rv;
}

/**
 * This function calls <code>free()</code> on the key and on the value
 * stored in the node <code>n</code>.
 *
 * @param[in] cex c-style exception
 * @param[in] t tree
 * @param[in] n node
 * @param[in] data client data
 */
static int
free_pair(cgul_exception_t* cex,
          cgul_rbtree_t t,
          cgul_rbtree_node_t n,
          void* data)
{
    int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    free((void*)cgul_rbtree_node__get_key(cex, n));
    free(cgul_rbtree_node__get_value(cex, n));
    rv = 1;

 out:
    return rv;
}

/**
 * Initialize the rbtree <code>t</code> by loading key/value pairs.
 *
 * @param[in,out] cex c-style exception
 * @param[in] t tree
 */
static void
init_rbtree(cgul_exception_t* cex,
            cgul_rbtree_t t)
{
    int i = 0;
    int insert_rv = 0;
    char* key = NULL;
    int* value = NULL;
    cgul_rbtree_node_t n = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(t);

    for (i = 0 ; i < 5 ; ++i) {

        /* Allocate. */
        key = cgul_libc__calloc(cex, 1, sizeof(char) * 2);
        value = cgul_libc__calloc(cex, 1, sizeof(int));
        if (*cex) {
            goto out;
        }

        /* Assign values. */
        *key = 'a' + i;
        *value = i;

        /* Insert into rbtree. */
        insert_rv = cgul_rbtree__insert(cex, t, key, &n);
        if (*cex) {
            goto out;
        }
        if (!insert_rv) {
            *cex = cgul_exception__new("key already exists: %s", key);
            goto out;
        }
        cgul_rbtree_node__set_value(cex, n, value);

        /* The rbtree now owns the key and value. */
        key = NULL;
        value = NULL;
    }

 out:

    /* Clean up. */
    if (key) {
        free(key);
    }
    if (value) {
        free(value);
    }

    return;
}

int
main(int argc,
     char* argv[])
{
    int rv = 0;
    cgul_rbtree_t t = NULL;
    cgul_rbtree_node_t first = NULL;
    cgul_rbtree_node_t last = NULL;
    cgul_exception_t local = NULL;

    /* Create a new rbtree that holds a range of values. */
    t = cgul_rbtree__new(&local, (cgul_rbtree__compare_t)strcmp);
    if (local) {
        goto out;
    }

    /* Initialize rbtree. */
    init_rbtree(&local, t);
    if (local) {
        goto out;
    }

    /* Print the rbtree. */
    cgul_rbtree__traverse(&local, t, print_pair, NULL);

    puts("=====");

    /* Print a range of the rbtree. */
    first = cgul_rbtree__get_front(&local, t);
    if (first) {
        first = cgul_rbtree_node__get_next(&local, first);
    }
    last = cgul_rbtree__get_back(&local, t);
    if (last) {
        last = cgul_rbtree_node__get_prev(&local, last);
    }
    cgul_rbtree__traverse_range(&local, t, first, last, print_pair, NULL);

 out:

    /* Catch exceptions. */
    if (local) {
        const char* msg = cgul_exception__get_message(local);
        fprintf(stderr, "*** Error: %s\n", msg);
        cgul_exception__delete(local);
        local = NULL;
        rv = 1;
    }

    /* Clean up. */
    if (t) {
        cgul_rbtree__traverse(&local, t, free_pair, NULL);
        cgul_rbtree__delete(t);
    }

    return rv;
}
