#include <assert.h>
#include <string.h>
#include "cgul_libc.h"
#include "cgul_stanza.h"

/**
 * This method prints the key/value pair for <code>key</code>.
 *
 * @param[in] cex c-style exception
 * @param[in] stanza stanza
 * @param[in] key key
 */
static void
print_key_value_pair(cgul_exception_t* cex,
                     cgul_stanza_t stanza,
                     const char* key)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

#if 0
    /* You can either do a normal lookup and let cgul_stanza own the
     * value. */
    {
        const char* value = NULL;

        value = cgul_stanza__lookup(cex, stanza, key);
        if (value == NULL) {
            value = "<invalid key>";
        }
        cgul_libc__printf(cex, "%s ==> %s\n", key, value);
    }

#else

    /* Or, you can take the value for long-term use.  Just remember to
     * free it. */
    {
        char* value = NULL;
        value = cgul_stanza__take_value(cex, stanza, key);
        if (value == NULL) {
            cgul_libc__printf(cex, "%s ==> %s\n", key, "<invalid key>");
        } else {
            cgul_libc__printf(cex, "%s ==> %s\n", key, value);
            /* Remember to free the value. */
            free(value);
        }
    }

#endif

 out:
    return;
}

int main(int argc, char* argv[])
{
    int rv = 0;
    cgul_stanza_t stanza = NULL;
    cgul_rbtree_t allowed_keys = NULL;
    cgul_rbtree_t mandatory_keys = NULL;
    cgul_exception_t local = NULL;

    /*
     * It is not necessary, but we want to test to make sure the
     * allowed_keys and the mandatory_keys code is working.
     *
     * NOTE: Below we have two different cgul_rbtree objects for the
     * allowed_keys list and the mandatory_keys list.  Typically, you
     * will pass in the same list for both.
     */

    /* The allowed_keys list is designed to warn when unexpected keys
     * are in a stanza. */
    allowed_keys = cgul_rbtree__new(&local, (cgul_rbtree__compare_t)strcmp);
    cgul_rbtree__insert(&local, allowed_keys, "foo", NULL);
    cgul_rbtree__insert(&local, allowed_keys, "bar", NULL);
    cgul_rbtree__insert(&local, allowed_keys, "baz", NULL);
    cgul_rbtree__insert(&local, allowed_keys, "hmm", NULL);

    /* The mandatory_keys list is designed to warn when an expected
     * key is missing from a stanza. */
    mandatory_keys = cgul_rbtree__new(&local, (cgul_rbtree__compare_t)strcmp);
    cgul_rbtree__insert(&local, mandatory_keys, "foo", NULL);
    cgul_rbtree__insert(&local, mandatory_keys, "bar", NULL);
    cgul_rbtree__insert(&local, mandatory_keys, "baz", NULL);

    /*
     * Create a new stanza object.
     */
    stanza = cgul_stanza__new_from_fname(&local,
                                         "main_stanza_input.txt",
                                         1 /* is_block_buffered */);
    if (local) {
        goto out;
    }
#if 0
    cgul_stanza__all_one_stanza(&local, stanza, 1);
    cgul_stanza__allow_duplicate_keys(&local, stanza, 1);
#endif
    cgul_stanza__set_allowed_keys(&local, stanza, allowed_keys);
    cgul_stanza__set_mandatory_keys(&local, stanza, mandatory_keys);

    /* Iterate over each stanza. */
    while (cgul_stanza__load_next(&local, stanza)) {
        print_key_value_pair(&local, stanza, "foo");
        print_key_value_pair(&local, stanza, "bar");
        print_key_value_pair(&local, stanza, "baz");
        print_key_value_pair(&local, stanza, "hmm");
        cgul_libc__printf(&local, "====\n");
        if (local) {
            goto out;
        }
    }

 out:

    if (local) {
        const char* msg = cgul_exception__get_message(local);
        fprintf(stderr, "***\n");
        fprintf(stderr, "*** main_stanza: Error: %s\n", msg);
        fprintf(stderr, "***\n");
        cgul_exception__delete(local);
        local = NULL;
        rv = 1;
    }

    if (stanza) {
        cgul_stanza__delete(stanza);
    }

    if (mandatory_keys) {
        cgul_rbtree__delete(mandatory_keys);
    }

    if (allowed_keys) {
        cgul_rbtree__delete(allowed_keys);
    }

    return rv;
}
