/**
 * This program uses <code>cgul_stanza</code> to convert a stanza file
 * to CSV format.  It uses the keys in the first stanza to generate
 * the column headers that populate the first row of the CSV file.
 * The column headers will be in the same order as the keys that
 * appear in the first stanza.  All the stanzas that follow must
 * contain the exact same set of keys but not necessarily in the exact
 * same order.  It then produces one CSV data row per stanza where the
 * values in the stanza map to the data in the CSV file. <p>
 *
 * Use the <code>main_csv_to_stanza</code> program to convert from CSV
 * format to <code>cgul_stanza</code> format. <p>
 *
 * @file
 * @author Paul Serice
 * @brief stanza to csv converter
 */

#include <assert.h>
#include "cgul.h"

/**
 * This function prints the CSV row of <code>headers</code> to
 * <code>f</code>.  The total number of headers is held in
 * <code>headers_count</code>.  The <code>row</code> is just a
 * temporary string that is used by the various output functions to
 * generate each row.  <code>row</code> is passed in to avoid having
 * to create and destroy it for each row that gets printed.
 *
 * @param[in,out] cex c-style exception
 * @param[in] f output file
 * @param[in] headers headers
 * @param[in] headers_count count of headers
 * @param[in] row row
 */
static void
print_csv_headers_row(cgul_exception_t* cex,
                      FILE* f,
                      char** headers,
                      size_t headers_count,
                      cgul_string_t row)
{
    size_t i = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(f);
    assert(row);

    /* Clear the "row" string. */
    cgul_string__set_value(cex, row, "");

    /* Iterate over the headers. */
    for (i = 0 ; i < headers_count ; ++i) {

        /* Append the comma. */
        if (i > 0) {
            cgul_string__append_char(cex, row, ',');
            if (*cex) {
                goto out;
            }
        }

        /* Append the properly quoted header. */
        cgul_csv__quote(cex, headers[i], row);
        if (*cex) {
            goto out;
        }

    }

    /* Print the result to f. */
    cgul_libc__fprintf(cex,
                       f,
                       "%s\n",
                       cgul_string__get_value(cex, row));

 out:
    return;
}

/**
 * This function prints the CSV row of data to <code>f</code> by using
 * <code>stanza</code> to look up all of the <code>headers</code> in
 * the current stanza.  The total number of headers is held in
 * <code>headers_count</code>.  The <code>row</code> is just a
 * temporary string that is used by the various output functions to
 * generate each row.  <code>row</code> is passed in to avoid having
 * to create and destroy it for each row that gets printed.
 *
 * @param[in,out] cex c-style exception
 * @param[in] f output file
 * @param[in] headers headers
 * @param[in] headers_count count of headers
 * @param[in] stanza stanza
 * @param[in] row row
 */
static void
print_csv_data_row(cgul_exception_t* cex,
                   FILE* f,
                   char** headers,
                   size_t headers_count,
                   cgul_stanza_t stanza,
                   cgul_string_t row)
{
    size_t i = 0;
    const char* value = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(f);
    assert(stanza);
    assert(row);

    /* Clear the "row" string. */
    cgul_string__set_value(cex, row, "");

    /* Iterate over the headers. */
    for (i = 0 ; i < headers_count ; ++i) {

        /* Append the comma. */
        if (i > 0) {
            cgul_string__append_char(cex, row, ',');
            if (*cex) {
                goto out;
            }
        }

        /* Lookup the value associated with this header. */
        value = cgul_stanza__lookup(cex, stanza, headers[i]);
        if (value == NULL) {
            *cex =
                cgul_exception__new(
                    "Unable to find data for header \"%s\" for the stanza "
                    "starting on line %ld",
                    headers[i],
                    cgul_stanza__get_line_number(cex, stanza));
            goto out;
        }

        /* Append the properly quoted header. */
        cgul_csv__quote(cex, value, row);
        if (*cex) {
            goto out;
        }

    }

    /* Print the result to f. */
    cgul_libc__fprintf(cex,
                       f,
                       "%s\n",
                       cgul_string__get_value(cex, row));

 out:
    return;
}

/**
 * This function delete the vector of headers held in
 * <code>headers</code>.  This includes calling
 * <code>cgul_vector__delete()</code> on the vector itself.  Thus, the
 * client must not use <code>headers</code> after calling this
 * function.
 *
 * @param[in] cex c-style exception
 * @param[in] keys vector of keys
 */
static void
delete_vector_of_keys(cgul_exception_t* cex,
                      cgul_vector_t keys)
{
    char* key = NULL;
    cgul_exception_t local = NULL;

    /* Delete each key on the vector. */
    while ((key = (char*)cgul_vector__pop_back(&local, keys)) != NULL) {
        free(key);
    }

    /* Delete the vector itself. */
    cgul_vector__delete(keys);
}

int
main(int argc,
     char* argv[])
{
    int rv = 0;
    size_t headers_count = 0;
    char** headers = NULL;
    cgul_string_t row = NULL;
    cgul_vector_t keys = NULL;
    cgul_stanza_t stanza = NULL;
    cgul_exception_t local = NULL;

    /* Verify command line. */
    if (argc != 2) {
        fprintf(stderr, "Usage: main_stanza_to_csv <file.txt>\n");
        goto out;
    }

    /* Allocate "row". */
    row = cgul_string__new(&local);
    if (local) {
        goto out;
    }

    /*
     * Open the stanza file.
     */
    stanza = cgul_stanza__new_from_fname(&local,
                                         argv[1],
                                         1 /* is_block_buffered */);
    if (local) {
        goto out;
    }

    /*
     * Load the first stanza.
     */
    if (!cgul_stanza__load_next(&local, stanza)) {
        goto out;
    }

    /* The headers for the CSV columns correspond to the keys in the
     * first stanza.  We use cgul_stanza__copy_keys() to maintain the
     * order of the keys as they appear in the cgul_stanza file. */
    keys = cgul_stanza__copy_keys(&local, stanza);
    if (local) {
        goto out;
    }

    /* Get the headers. */
    headers = (char**)cgul_vector__get_array(&local, keys);

    /* Get the count of headers. */
    headers_count = cgul_vector__get_size(&local, keys);

    /*
     * Print the CSV headers and data rows for the first stanza.
     */
    print_csv_headers_row(&local, stdout, headers, headers_count, row);
    print_csv_data_row(&local, stdout, headers, headers_count, stanza, row);

    /*
     * Print the CSV data for the remaining stanzas.
     */
    while (cgul_stanza__load_next(&local, stanza)) {
        print_csv_data_row(&local, stdout, headers, headers_count, stanza, row);
    }

 out:

    /* Catch exceptions. */
    if (local) {
        const char* msg = cgul_exception__get_message(local);
        fprintf(stderr, "*** Error: %s\n", msg);
        cgul_exception__delete(local);
        local = NULL;
        rv = 1;
    }

    /*
     * Clean up.
     */

    /* stanza */
    if (stanza) {
        cgul_stanza__delete(stanza);
    }

    /* keys */
    if (keys) {
        delete_vector_of_keys(&local, keys);
    }

    /* row */
    if (row) {
        cgul_string__delete(row);
    }

    return rv;
}
