/**
 * This program uses <code>cgul_stanza_cxx</code> to convert a stanza file
 * to CSV format.  It uses the keys in the first stanza to generate
 * the column headers that populate the first row of the CSV file.
 * The column headers will be in the same order as the keys that
 * appear in the first stanza.  All the stanzas that follow must
 * contain the exact same set of keys but not necessarily in the exact
 * same order.  It then produces one CSV data row per stanza where the
 * values in the stanza map to the data in the CSV file. <p>
 *
 * Use the <code>main_csv_to_stanza_cxx</code> program to convert from CSV
 * format to <code>cgul_stanza</code> format. <p>
 *
 * @file
 * @author Paul Serice
 * @brief stanza to csv converter
 */

#include <iostream>
#include <sstream>
#include <stdexcept>
#include "cgul_cxx.h"

using namespace std;

/**
 * This function prints the CSV row of <code>headers</code> to
 * <code>cout</code>.  The total number of headers is held in
 * <code>headers_count</code>.  The <code>row</code> is just a
 * temporary string that is used by the various output functions to
 * generate each row.  <code>row</code> is passed in to avoid having
 * to create and destroy it for each row that gets printed.
 *
 * @param[in] headers headers
 * @param[in] headers_count count of headers
 * @param[in] row row
 */
static void
print_csv_headers_row(char** headers,
                      size_t headers_count,
                      cgul_string_cxx& row)
{
    size_t i = 0;

    /* Clear the "row" string. */
    row.set_value("");

    /* Iterate over the headers. */
    for (i = 0 ; i < headers_count ; ++i) {

        /* Append the comma. */
        if (i > 0) {
            row.append_char(',');
        }

        /* Append the properly quoted header. */
        cgul_csv_cxx::quote(headers[i], row);

    }

    /* Print the result. */
    cout << row.get_value() << endl;
}

/**
 * This function prints the CSV row of data to <code>cout</code> by using
 * <code>stanza</code> to look up all of the <code>headers</code> in
 * the current stanza.  The total number of headers is held in
 * <code>headers_count</code>.  The <code>row</code> is just a
 * temporary string that is used by the various output functions to
 * generate each row.  <code>row</code> is passed in to avoid having
 * to create and destroy it for each row that gets printed.
 *
 * @param[in] headers headers
 * @param[in] headers_count count of headers
 * @param[in] stanza stanza
 * @param[in] row row
 */
static void
print_csv_data_row(char** headers,
                   size_t headers_count,
                   cgul_stanza_cxx& stanza,
                   cgul_string_cxx& row)
{
    size_t i = 0;
    const char* value = NULL;

    /* Clear the "row" string. */
    row.set_value("");

    /* Iterate over the headers. */
    for (i = 0 ; i < headers_count ; ++i) {

        /* Append the comma. */
        if (i > 0) {
            row.append_char(',');
        }

        /* Lookup the value associated with this header. */
        value = stanza.lookup(headers[i]);
        if (value == NULL) {
            ostringstream msg;
            msg << "Unable to find data for header "
                << "\"" << headers[i] << "\""
                << "for the stanza starting on line "
                << stanza.get_line_number();
            throw runtime_error(msg.str());
        }

        /* Append the properly quoted header. */
        cgul_csv_cxx::quote(value, row);

    }

    /* Print the result to f. */
    cout << row.get_value() << endl;
}

/**
 * This function delete the vector of headers held in
 * <code>headers</code>.
 *
 * @param[in] keys vector of keys
 */
static void
delete_vector_of_keys(cgul_vector_cxx* keys)
{
    char* key = NULL;

    if (keys) {
        /* Delete each key on the vector. */
        while ((key = (char*)keys->pop_back()) != NULL) {
            free(key);
        }

        delete keys;
    }
}

int
main(int argc,
     char* argv[])
{
    int rv = 0;
    size_t headers_count = 0;
    char** headers = NULL;
    cgul_string_cxx row;
    cgul_vector_cxx* keys = NULL;
    cgul_stanza_cxx stanza;

    try {

        /* Verify command line. */
        if (argc != 2) {
            fprintf(stderr, "Usage: main_stanza_to_csv <file.txt>\n");
            return 1;
        }

        /*
         * Open the stanza file.
         */
        stanza.open_fname(argv[1]);

        /*
         * Load the first stanza.
         */
        if (stanza.load_next()) {

            /* The headers for the CSV columns correspond to the keys
             * in the first stanza.  We use cgul_stanza__copy_keys()
             * to maintain the order of the keys as they appear in the
             * cgul_stanza file. */
            keys = stanza.copy_keys();

            /* Get the headers. */
            headers = (char**)keys->get_array();

            /* Get the count of headers. */
            headers_count = keys->get_size();

            /*
             * Print the CSV headers and data rows for the first stanza.
             */
            print_csv_headers_row(headers, headers_count, row);
            print_csv_data_row(headers, headers_count, stanza, row);

            /*
             * Print the CSV data for the remaining stanzas.
             */
            while (stanza.load_next()) {
                print_csv_data_row(headers, headers_count, stanza, row);
            }

        }

    } catch (const exception& e) {
        cerr << "*** Error: " << e.what() << endl;
        rv = 1;
    } catch (const cgul_exception_cxx& e) {
        cerr << "*** Error: " << e.get_message() << endl;
        rv = 1;
    }

    /*
     * Clean up.
     */

    /* keys */
    if (keys) {
        delete_vector_of_keys(keys);
    }

    return rv;
}
