/**
 * Convert the EOL sequences for UTF-8 text files.  This mainly used
 * to exercise the UTF-8 decoder, but it also works with the ASCII
 * decoder if you se the USE_ASCII_DECODER macro.  For better
 * performance, use the main_stream_reader_crlf program instead.  You
 * can diff the two source files to see the minor differences.
 *
 * @file
 */

#include "cgul.h"

#ifdef WIN32
#include <fcntl.h>
#include <io.h>
#endif

/**
 * Whether the use the ASCII decoder.  If not defined, the UTF-8
 * decoder is used instead.  For mostly ASCII files, the ASCII decoder
 * is slightly more efficient, but for extended character sets, the
 * UTF-8 decoder is more efficient.
 */
#define USE_ASCII_DECODER 0

/*
 * Pick the correct cgul_string__append_*() function depending on
 * whether we are using the ASCII or UTF-8 decoder.
 */
#if USE_ASCII_DECODER
#  define MAIN__APPEND_CHAR cgul_string__append_char
#else
#  define MAIN__APPEND_CHAR cgul_string__append_wchar
#endif

/**
 * Type definition for the EOL sequence to use.
 */
typedef enum {

    /** unknown */
    EOL_UNKNOWN = 0,

    /** '\r' */
    EOL_CR,

    /** '\n' */
    EOL_LF,

    /** "\r\n" */
    EOL_CRLF,

} eol_sequence_t;

/**
 * Print the usage message to the output file <code>fout</code>.  The
 * base name of the executable <code>bname</code> is used as part of
 * the usage message.
 *
 * @param[in] cex c-style exception
 * @param[in] bname base name of executable
 * @param[in] fout output file
 */
static void
main__print_usage(cgul_exception_t* cex,
                  const char* bname,
                  FILE* fout)
{
    static const char* usage = "\
\n\
Usage: %s [options] [<file> . . .]\n\
\n\
       General Options\n\
        -h or --help    : Show this help.\n\
        -v or --verbose : Report files on stdout.\n\
\n\
       Conversion Options\n\
        -d or --crlf    : Convert to DOS format.\n\
        -m or --cr      : Convert to Mac format.\n\
        -u or --lf      : Convert to Unix format.\n\
\n\
       Instructions\n\
        The default conversion depends on the name used to\n\
        start the program.  If you invoke it as \"crlf\", DOS\n\
        conversion is the default.  If you invoke it as \"r\",\n\
        Mac conversion is the default.  If you invoke it as\n\
        \"lf\", Unix conversion is the default.\n\
\n\
        If you don't specify a file or specify the \"-\" file,\n\
        standard input and standard output is used.  For all\n\
        other files, the conversion is safely performed\n\
        in-place.\n\
\n\
        This program does not overwrite the original file.  Instead,\n\
        it generates a random file name based on the name of the file\n\
        being converted.  Ideally, this temporary file would be \n\
        atomically renamed to match the name of the original file, but\n\
        there is no portable way to do this.  Instead, the temporary\n\
        file is just left on the file system for you to move however\n\
        your operating system allows.\n\
\n\
        Note that there is a race condition between testing if the\n\
        temporary file already exists and then writing to it that could\n\
        be exploited.\n\
\n\
        DO NOT USE THIS PROGRAM UNLESS YOU FULLY UNDERSTAND THE RISK!!!\n\
\n\
";

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(bname);
    assert(fout);

    fprintf(fout, usage, bname);

 out:
    return;
}

/**
 * Force the standard input <code>stdin</code> and standard output
 * <code>stdout</code> to be in binary mode (instead of text mode).
 * If an error occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 */
static void
main__force_binary_mode(cgul_exception_t* cex)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

#ifdef WIN32
    /* Convert stdin and stdout to binary mode. */
    if (_setmode(_fileno(stdin), _O_BINARY) < 0) {
        *cex = cgul_exception__new("_setmode(stdin): %m");
        goto out;
    }
    if (_setmode(_fileno(stdout), _O_BINARY) < 0) {
        *cex = cgul_exception__new("_setmode(stdout): %m");
        goto out;
    }
#endif

 out:
    return;
}

/**
 * Add the file name <code>fname</code> to the vector of file names
 * <code>fnames</code>.  If an error occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[out] fnames file names
 * @param[in] fname file name
 */
static void
main__add_fname(cgul_exception_t* cex,
                cgul_vector_t fnames,
                const char* fname)
{
    char* tmp = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(fnames);

    /* Sanity check. */
    if (!fname) {
        goto out;
    }

    /* Allocate. */
    tmp = cgul_libc__strdup(cex, fname);
    if (*cex) {
        goto out;
    }

    /* Add the file name. */
    cgul_vector__push_back(cex, fnames, tmp);
    if (*cex) {
        goto out;
    }

    /* fnames now owns tmp. */
    tmp = NULL;

 out:

    /* Clean up. */
    if (tmp) {
        free(tmp);
    }

    return;
}

/**
 * Process the command line returning the file names that are present
 * on the command line in <code>fnames</code> and the explicitly set
 * EOL sequence in <code>eol_sequence</code>.  If the user wants help,
 * <code>user_wants_help</code> will be set; otherwise, it will be
 * cleared.  If the user wants verbosity,
 * <code>user_wants_verbosity</code> will be set; otherwise, it will
 * be cleared.  If an error occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] argc command-line argument count
 * @param[in] argv command-line argument vector
 * @param[out] fnames file names
 * @param[in,out] eol_sequence EOL sequence
 * @param[out] user_wants_help whether the user user wants help
 * @param[out] user_wants_verbose whethe the user wants verbosity
 */
static void
main__process_cmdline(cgul_exception_t* cex,
                      int argc,
                      char** argv,
                      cgul_vector_t fnames,
                      eol_sequence_t* eol_sequence,
                      int* user_wants_help,
                      int* user_wants_verbosity)
{
    int i = 0;
    int only_file_names_remain = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(fnames);

    /* Clear help and verbosity. */
    if (user_wants_help) {
        *user_wants_help = 0;
    }
    if (user_wants_verbosity) {
        *user_wants_verbosity = 0;
    }

    /* Iterate over the command-line arguments. */
    for (i = 1 ; i < argc ; ++i) {

        /* Unconditionally treat this command-line argument as a file
         * if "--" has been seen or if it does not start with a
         * dash. */
        if (only_file_names_remain ||
            ((argv[i][0] != '\0') && (argv[i][0] != '-')))
        {
            main__add_fname(cex, fnames, argv[i]);
            if (*cex) {
                goto out;
            }
            continue;
        }

        /* -- */
        if (strcmp(argv[i], "--") == 0) {
            only_file_names_remain = 1;
            continue;
        }

        /* --help */
        if ((strcmp(argv[i], "-h") == 0) ||
            (strcmp(argv[i], "-help") == 0) ||
            (strcmp(argv[i], "--help") == 0))
        {
            *user_wants_help = 1;
            continue;
        }

        /* --verbose */
        if ((strcmp(argv[i], "-v") == 0) ||
            (strcmp(argv[i], "-verbose") == 0) ||
            (strcmp(argv[i], "--verbose") == 0))
        {
            *user_wants_verbosity = 1;
            continue;
        }

        /* --crlf */
        if ((strcmp(argv[i], "-d") == 0) ||
            (strcmp(argv[i], "-crlf") == 0) ||
            (strcmp(argv[i], "--crlf") == 0))
        {
            *eol_sequence = EOL_CRLF;
            continue;
        }

        /* --cr */
        if ((strcmp(argv[i], "-m") == 0) ||
            (strcmp(argv[i], "-cr") == 0) ||
            (strcmp(argv[i], "--cr") == 0))
        {
            *eol_sequence = EOL_CR;
            continue;
        }

        /* --lf */
        if ((strcmp(argv[i], "-u") == 0) ||
            (strcmp(argv[i], "-lf") == 0) ||
            (strcmp(argv[i], "--lf") == 0))
        {
            *eol_sequence = EOL_LF;
            continue;
        }

        /* Error */
        *cex = cgul_exception__new(
            "Invalid command line argument: %s", argv[i]);
        goto out;

    }

 out:
    return;
}

/**
 * Use the pseudo-random number generator <code>prng</code> to
 * generate the name of the output file from <code>fname</code> which
 * is the name of the file being converted.  First we try to open the
 * file for reading.  If this fails, we know the file does not exist
 * at that time so that is the file name we will use.  Then, we will
 * open the same file name for writing and return it.  If an error
 * occurs, <code>NULL</code> is returned, and an exception is
 * thrown. <p>
 *
 * @note There is an obvious race condition here that could be
 * exploited, but <code>mkstemp()</code> is not widely available at
 * the time of writing.
 *
 * @param[in,out] cex c-style exception
 * @param[in] fname name of file to convert
 * @param[in] prng pseudo-random number generator
 * @return output file
 */
static FILE*
main__open_output_file(cgul_exception_t* cex,
                       const char* fname,
                       cgul_random_cmwc32_t prng)
{
    FILE* rv = NULL;
    cgul_uint32_t uint32 = 0;
    cgul_string_t tmp = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(fname);
    assert(prng);

    /* Allocate. */
    tmp = cgul_string__new(cex);
    if (*cex) {
        goto out;
    }

    /* Iterate until we can open the output file with a randomly
     * generated name.  NOTE: There is a race condition here, but
     * mkstemp() is not widely available at the time of writing!!! */
    for (;;) {

        /* Clear the temporary file name. */
        cgul_string__set_length(cex, tmp, 0);
        if (*cex) {
            goto out;
        }

        /* Generate a 32-bit random number and append it to fname.
         * This would be more difficult to exploit if there was a
         * secure, portable way to seed the PRNG. */
        uint32 = cgul_random_cmwc32__next(cex, prng);
        cgul_string__append_sprintf(
            cex, tmp, "%s.%08lX", fname, (unsigned long)uint32);
        if (*cex) {
            goto out;
        }

        /* Try to open the file for *reading* which will fail if the
         * file does not exist (which is what we want). */
        rv = fopen(cgul_string__get_value(cex, tmp), "rb");
        if (rv) {
            /* We were able to open the file so it already exists.
             * Thus, this is *not* a file we should be overwriting.
             * Close it and try again. */
            fclose(rv);
            rv = NULL;
            continue;
        }

        /*
         * Race Condition: The fopen() above and the
         * cgul_libc__fopen() below should be atomic but aren't, and
         * there is no good way to do this portably.
         */

        /* Now try to open the file for writing. */
        rv = cgul_libc__fopen(cex, cgul_string__get_value(cex, tmp), "wb");
        if (*cex) {
            goto out;
        }

        /* Success. */
        break;
    }

 out:

    /* Clean up. */
    if (*cex && rv) {
        fclose(rv);
        rv = NULL;
    }
    if (tmp) {
        cgul_string__delete(tmp);
    }

    return rv;
}

/**
 * Process the file <code>fname</code> by converting all of its EOL
 * sequences to the type specified by <code>eol_sequence</code>
 * writing the result to the output file that will have its name based
 * on <code>fname</code> but with a random extension generated using
 * the pseudo-random number generator <code>prng</code>.  If
 * <code>fname</code> is <code>"-"</code>, <code>stdin</code> will be
 * converted and written to <code>stdout</code>.  If an error occurs,
 * an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] fname name of file to convert
 * @param[in] eol_sequence EOL sequence type
 * @param[in] prng pseudo-random number generator
 */
static void
main__process_file(cgul_exception_t* cex,
                   const char* fname,
                   eol_sequence_t eol_sequence,
                   cgul_random_cmwc32_t prng)
{
    int we_own_files = 0;
    FILE* fin = NULL;
    FILE* fout = NULL;
    cgul_wchar_t next_char = CGUL_WCHAR__NUL;
    cgul_string_t line = NULL;
    cgul_block_reader_file_t br = NULL;
    cgul_stream_reader_t sr = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Sanity check. */
    if (!fname) {
        goto out;
    }

    /* Allocate. */
    line = cgul_string__new(cex);
    br = cgul_block_reader_file__new(cex);
    sr = cgul_stream_reader__new(cex);
    if (*cex) {
        goto out;
    }

#if USE_ASCII_DECODER
    /* Set the stream reader to use the ASCII decoder. */
    cgul_stream_reader__set_decoder(
        cex, sr, cgul_stream_reader__decode_ascii);
#endif

    /* We do not need to track line and column numbers. */
    cgul_stream_reader__set_line_and_column_tracking(cex, sr, 0);

    /* Use stdin and stdout if "-" is specified as the file name. */
    if (strcmp(fname, "-") == 0) {
        fin = stdin;
        fout = stdout;
    } else {
        /* Remember that we need to close any file that we open. */
        we_own_files = 1;
        /* Open the input file. */
        fin = cgul_libc__fopen(cex, fname, "rb");
        if (*cex) {
            goto out;
        }
        /* Open the output file. */
        fout = main__open_output_file(cex, fname, prng);
        if (*cex) {
            goto out;
        }
    }

    /* Configure the stream reader to use "fin".  Remember that "sr"
     * assumes the input file is UTF-8, and it normalizes EOL
     * sequences to exactly one '\n' character. */
    cgul_block_reader_file__open_file(cex, br, fin);
    cgul_stream_reader__add_block_reader(
        cex, sr, cgul_block_reader_file__as_block_reader(cex, br));

    /* Iterate over each character in the file.  (This is slow.) */
    for (;;) {

        /* Read the next character. */
        if (!cgul_stream_reader__read(cex, sr, &next_char)) {
            if (*cex) {
                goto out;
            }
            /* EOF */
            break;
        }

        /* If this is not EOL, just append the character to "line" and
         * continue with the next character. */
        if (next_char != (cgul_wchar_t)'\n') {
            MAIN__APPEND_CHAR(cex, line, next_char);
            if (*cex) {
                goto out;
            }
            continue;
        }

        /* Otherwise, replace with the EOL sequence requested by the user. */
        switch (eol_sequence) {
        case EOL_CRLF:
            MAIN__APPEND_CHAR(cex, line, 0xD);
            MAIN__APPEND_CHAR(cex, line, 0xA);
            break;
        case EOL_CR:
            MAIN__APPEND_CHAR(cex, line, 0xD);
            break;
        case EOL_LF:
            MAIN__APPEND_CHAR(cex, line, 0xA);
            break;
        default:
            *cex = cgul_exception__new(
                "internal error: invalid EOL sequence");
            goto out;
        }

        /* Proagate exceptions. */
        if (*cex) {
            goto out;
        }

        /* Write the line. */
        fprintf(fout, "%s", cgul_string__get_value(cex, line));
        if (*cex) {
            goto out;
        }

        /* Clear the line. */
        cgul_string__set_length(cex, line, 0);
    }

    /* Print the remainder. */
    if (cgul_string__get_length(cex, line) > 0) {
        fprintf(fout, "%s", cgul_string__get_value(cex, line));
    }

 out:

    /* Clean up. */
    if (we_own_files) {
        if (fout) {
            fclose(fout);
        }
        if (fin) {
            fclose(fin);
        }
    }
    if (sr) {
        cgul_stream_reader__delete(sr);
    }
    if (br) {
        cgul_block_reader_file__delete(br);
    }
    if (line) {
        cgul_string__delete(line);
    }

    return;
}

int
main(int argc,
     char* argv[])
{
    int rv = 0;
    int user_wants_help = 0;
    int user_wants_verbosity = 0;
    size_t i = 0;
    size_t i_max = 0;
    char* rname = NULL;
    char* bname = NULL;
    eol_sequence_t eol_sequence = EOL_UNKNOWN;
    cgul_vector_t fnames = NULL;
    cgul_random_cmwc32_t prng = NULL;
    cgul_exception_t local = NULL;

    /* Sanity check. */
    if (argc <= 0) {
        local = cgul_exception__new("operating system fault!");
        goto out;
    }

    /* Allocate. */
    fnames = cgul_vector__new(&local);
    prng = cgul_random_cmwc32__new(&local);
    if (local) {
        goto out;
    }

    /* Get the base name for the executable without the file extension. */
    rname = cgul_libc__rootname(&local, argv[0]);
    bname = cgul_libc__basename(&local, rname);
    if (local) {
        goto out;
    }

    /* Multi-call binary picks the default EOL sequence. */
    if (strcmp(bname, "crlf") == 0) {
        eol_sequence = EOL_CRLF;
    } else if (strcmp(bname, "cr") == 0) {
        eol_sequence = EOL_CR;
    } else if (strcmp(bname, "lf") == 0) {
        eol_sequence = EOL_LF;
    }

    /* Process the command line arguments. */
    main__process_cmdline(&local, argc, argv, fnames, &eol_sequence,
                          &user_wants_help, &user_wants_verbosity);
    if (local) {
        goto out;
    }

    /* Print help if requested. */
    if (user_wants_help) {
        main__print_usage(&local, bname, stdout);
        goto out;
    }

    /* If we still do not have an EOL sequence, print the usage and exit. */
    if (eol_sequence == EOL_UNKNOWN) {
        main__print_usage(&local, bname, stderr);
        goto out;
    }

    /* Wait until after there is no longer a chance that we will need
     * to print a usage message before switching stdin and stdout to
     * binary mode. */
    main__force_binary_mode(&local);
    if (local) {
        goto out;
    }

    /* Determine the number of files that were on the command line. */
    i_max = cgul_vector__get_size(&local, fnames);

    if (i_max == 0) {

        /* Use stdin and stdout if no files were specified. */
        main__process_file(&local, "-", eol_sequence, prng);
        if (local) {
            goto out;
        }

    } else {

        /* Process each file specified on the command line. */
        for (i = 0 ; i < i_max ; ++i) {

            /* Get the file name. */
            const char* fname =
                (const char*)cgul_vector__at(&local, fnames, i);

            /* Verbose output. */
            if (user_wants_verbosity) {
                printf("Converting %s ... ", fname);
                fflush(stdout);
            }

            /* Process the file. */
            main__process_file(&local, fname, eol_sequence, prng);
            if (local) {
                goto out;
            }

            /* Verbose output. */
            if (user_wants_verbosity) {
                printf("Done.\n");
                fflush(stdout);
            }
        }
    }

 out:

    /* Handle exceptions. */
    if (local) {
        const char* msg = cgul_exception__get_message(local);
        fprintf(stderr, "*** Error: %s\n", msg);
        cgul_exception__delete(local);
        local = NULL;
        rv = 1;
    }

    /* Clean up. */
    if (bname) {
        free(bname);
    }
    if (rname) {
        free(rname);
    }
    if (prng) {
        cgul_random_cmwc32__delete(prng);
    }
    if (fnames) {
        cgul_vector__free_values(fnames);
        cgul_vector__delete(fnames);
    }

    return rv;
}
